Add LoadPath "./metatheory".

Set Implicit Arguments.

Require Import Metatheory.
Require Import lang.
Require Import langExtra.
Require Import infrastructure.
Require Import Coq.Logic.Classical_Prop.

(*****************************************************************)

Axiom lc_isEq : forall D e1 e2, 
  isEq D e1 e2 -> 
  lc_ectx D /\ lc_exp e1 /\ lc_exp e2.

Axiom ref_isEq : forall D e, 
  lc_ectx D -> lc_exp e ->
  isEq D e e.
Axiom sym_isEq : forall D e1 e2, 
  isEq D e1 e2 -> isEq D e2 e1.
Axiom trans_isEq : forall D e1 e2 e3, 
  isEq D e1 e2 -> isEq D e2 e3 -> isEq D e1 e3.
Axiom beta_isEq : forall e1 e2, 
  reduce e1 e2 -> isEq nil e1 e2.
Axiom empty_isEq : forall C1 C2 e1 e2, 
  pure e1 -> pure e2 -> 
  C1 <> C2 -> ~(isEq nil (e_Con C1 e1)(e_Con C2 e2)).
Axiom ctx_conv_isEq : 
  forall D1 D2 e1 e2,
    ctxereq D1 D2 -> isEq D1 e1 e2 -> isEq D2 e1 e2.
Axiom cut_isEq : 
  forall D1 D2 u1 u2 e1 e2, 
    isEq (D2 ++ (u1, u2) :: D1) e1 e2 -> 
    isEq D1 u1 u2 -> 
    isEq (D2 ++ D1) e1 e2.
Axiom weakening_isEq : 
  forall D1 D2 D3 e1 e2, 
    lc_ectx D2 ->
    isEq (D1 ++ D3) e1 e2 ->
    isEq (D1 ++ D2 ++ D3) e1 e2.
Axiom subst_isEq :
  forall D e1 e2 w x,
    pure w ->
    isEq D e1 e2 -> 
    isEq (esubst_ectx w x D)
         (esubst_exp w x e1)
         (esubst_exp w x e2).

Axiom injection2_isEq : forall D C w1 w2,
  pure w1 ->
  pure w2 ->
  isEq D (e_Con C w1) (e_Con C w2) ->
  isEq D w1 w2.

Lemma injection1_isEq : forall D C1 w1 C2 w2,
  ~ incon D ->
  pure w1 ->
  pure w2 ->
  isEq D (e_Con C1 w1) (e_Con C2 w2) ->
  C1 = C2.
Proof.
  intros D C1 w1 C2 w2 Hincon Hw1 Hw2 HisEq.
  destruct (C1 == C2); subst; auto.
    assert (False) as J.
      apply Hincon.
      eapply incon_intro with (e1:=w1) (e2:=w2); eauto.
    inversion J.
Qed.

Lemma lc_pure : forall w,
  pure w ->
  lc_exp w.
Proof.
  intros w Pw.
  induction Pw; auto.
Qed.

Lemma subst_v_w : forall w v x,
  is_v_of_exp v ->
  pure w ->
  is_v_of_exp (esubst_exp w x v).
Proof.
  intros w v x Hv Hw.
  induction v; simpl; auto.
    destruct (e == x); subst; auto.
      inversion Hv.
    simpl in Hv. destruct Hv as [Hv1 Hv2]; auto.
Qed.

Lemma subst_pure : forall w w' x,
  pure w ->
  pure w' ->
  pure (esubst_exp w' x w).
Proof.
  intros w w' x Pw Pw'.
  induction Pw; simpl; auto.
    destruct (x0 == x); auto.
    apply pure_lam; auto.
      apply esubst_exp_lc_exp with (e2:=w') (x1:=x) in H; auto.
        apply lc_pure; auto.
    apply pure_fun; auto.
      apply subst_v_w; auto.
      apply esubst_exp_lc_exp with (e2:=w') (x1:=x) in H0; auto.
        apply lc_pure; auto.
Qed.

Lemma subst_incon :
  forall D x w,
  pure w -> 
  incon D -> incon (esubst_ectx w x D).
Proof.
  intros D x w Pw HD.
  inversion HD; subst.
  apply incon_intro with (C1:=C1) (C2:=C2) (e1:=esubst_exp w x e1) (e2:=esubst_exp w x e2); auto.
    apply subst_pure; auto.
    apply subst_pure; auto.
    apply subst_isEq with (x:=x) (w:=w) in H1; auto. 
Qed.

Lemma empty_incon : 
  ~( incon nil ).
Proof.
  unfold not. intro H. inversion H.
  pose (L := @empty_isEq C1 C2 e1 e2).
  unfold not in L.
  apply L; auto.
Qed.

Lemma cut_incon : 
  forall D D' e e', 
    incon (D ++ (e, e') :: D') -> 
    isEq D' e e' -> 
    incon (D ++ D').
Proof.
  intros D D' e e' H1 H2.
  inversion H1; subst.
  apply incon_intro with (C1:=C1) (C2:=C2) (e1:=e1) (e2:=e2); auto.
    eapply cut_isEq; eauto. 
Qed.

Lemma ctx_conv_incon :
  forall D D', 
    incon D -> ctxereq D D' -> incon D'.
Proof.
  intros D D' Hincon Hctxeq.
  inversion Hincon; subst.
  apply incon_intro with (C1:=C1) (C2:=C2) (e1:=e1) (e2:=e2); auto.
    eapply ctx_conv_isEq; eauto. 
Qed.
  
Lemma weakening_incon :
  forall D1 D2 D3, 
    incon (D1 ++ D2) -> 
    lc_ectx D3 ->
    incon (D1 ++ D3 ++ D2).
Proof.
  intros. 
  inversion H; subst.
  eapply incon_intro with (e1 := e1) (e2:=e2) (C1 := C1) (C2:= C2) ; eauto. 
    eapply weakening_isEq; eauto.
Qed.

Lemma incon_decidable: forall D,
  (incon D) \/ ~(incon D).
Proof. 
  intros.
  apply classic.
Qed.
