Add LoadPath "./metatheory".

Require Import Metatheory.
Require Import lang.
Require Import langExtra.
Require Import infrastructure. 
Require Import reductions.

Definition closed (e : exp) : Prop :=
  efv_exp e [=] empty.

Inductive msubst : Set :=
| msubst_nil : msubst
| msubst_cons : forall (w : exp) (x : expvar) (d : msubst),
  pure w -> msubst.

Fixpoint msubst_exp (d : msubst) (e : exp) {struct d} : exp :=
  match d with
  | msubst_nil => e
  | msubst_cons w x d' _ => msubst_exp d' (esubst_exp w x e)
  end.

Fixpoint size_msubst (d : msubst) :=
  match d with
  | msubst_nil => 0
  | msubst_cons _ _ d' _ => S (size_msubst d')
  end.

Lemma msubst_front_back : forall d w x,
  pure w ->
  exists d', exists w', exists x',
    pure w' /\
    size_msubst d = size_msubst d' /\
    forall e, (msubst_exp d (esubst_exp w x e)) = (esubst_exp w' x' (msubst_exp d' e)).
Proof.
intro d. 
 induction d. 
 Case "nil".
 intros;  exists msubst_nil; exists w; exists x; split; simpl; eauto.
 Case "cons".
 intros. simpl. destruct (IHd w x) as [d' [w' [x' (Jp, (Jsize, Jeq))]]]; auto.
  exists (msubst_cons w0 x0 d' H); exists w'; exists x'; repeat split; simpl; auto. 
Qed.

Definition terminates (e : exp) : Prop :=
  exists u,
    mwreduce e u
    /\  forall u', ~wreduce u u'.

Definition isEqC e1 e2 : Prop :=
  forall E d,
    closed (plug E (msubst_exp d e1)) ->
    closed (plug E (msubst_exp d e2)) ->
    lc_evctx E ->
    is_cbv_evctx E ->
    (terminates (plug E (msubst_exp d e1)) <-> terminates (plug E (msubst_exp d e2))).

Lemma lc_wreduce2 : forall e1 e2,
  wreduce e1 e2 -> lc_exp e2.
Proof. intros. destruct (lc_wreduce _ _  H). assumption. Qed.

Hint Resolve lc_wreduce2.
Hint Constructors mwreduce.

Lemma terminates_wbeta : forall e e',
  wreduce e e' ->
  (terminates e <-> terminates e').
Proof.
  intros e e' Hstep.
  split.
    (* terminates e -> terminates e': *)
    unfold terminates.
    intros [u [Hr Hstuck]].
    exists u; split; [idtac  | assumption].
    assert (exists u', mwreduce u u' /\ mwreduce e' u') as H.
      apply mwreduce_diamond with (e:=e); eauto.
    destruct H as [u' [H1 H2]].
    inversion H1; subst; [ assumption | contradict H; auto].

    (* terminates e' -> terminates e *)
    unfold terminates.
    intros [u [Hr Hstuck]].
    exists u; eauto.
Qed.

Lemma terminates_mwbeta : forall e e',
  mwreduce e e' ->
  (terminates e <-> terminates e').
Proof.
  induction 1.
    tauto.
    rewrite (terminates_wbeta e e'); auto.
Qed.

Lemma reduce_msubst : forall d e1 e2,
  reduce e1 e2 ->
  reduce (msubst_exp d e1) (msubst_exp d e2).
Proof.
  induction d.
    simpl; auto.
    intros; simpl; apply IHd; apply reduce_subst; auto.
Qed.

Lemma msubst_extend : forall d w (H : pure w) x e,
  msubst_exp d  (esubst_exp w x e) = msubst_exp (msubst_cons w x d H) e.
Proof.
  intros; reflexivity.
Qed.

Lemma pure_msubst : forall d w,
  pure w ->
  pure (msubst_exp d w).
Proof.
  induction d; simpl; auto using pure_subst.
Qed.

Lemma isEqC_subst : forall w x e1 e2,
  pure w ->
  isEqC e1 e2 ->
  isEqC (esubst_exp w x e1) (esubst_exp w x e2).
Proof.
  unfold isEqC; intros.
  repeat rewrite msubst_extend with (H:=H).
  auto.
Qed.

Definition isEqC_msubst_prop (n : nat) :=
  forall e1 e2 d,
    n = size_msubst d ->
    isEqC e1 e2 ->
    isEqC (msubst_exp d e1) (msubst_exp d e2).

Lemma isEqC_msubst' : forall n,
  isEqC_msubst_prop n.
Proof.
  intros n. apply lt_wf_ind. clear n.
  unfold isEqC_msubst_prop.
  intros n IH e1 e2 d Heq Hiseq. subst.
  destruct d.
    simpl; tauto.

    repeat rewrite <- msubst_extend.
    destruct (msubst_front_back d w x p) as [d' [w' [x' [p' [Hsize Hfrontback]]]]].
    assert (size_msubst d' < size_msubst (msubst_cons w x d p)) by (simpl;omega).
    assert (isEqC (msubst_exp d' e1) (msubst_exp d' e2)).
      apply IH with (m:=size_msubst d'); auto.
    rewrite (Hfrontback e1).
    rewrite (Hfrontback e2).
    apply isEqC_subst; auto.
Qed.

Lemma isEqC_msubst : forall e1 e2 d,
  isEqC e1 e2 ->
  isEqC (msubst_exp d e1) (msubst_exp d e2).
Proof.
  intros.
  pose (isEqC_msubst' (size_msubst d)).
  unfold isEqC_msubst_prop in *.
  auto.
Qed.

Fixpoint respects d D {struct D} : Prop :=
  match D with
  | nil => True
  | (e1,e2)::D' => isEqC (msubst_exp d e1) (msubst_exp d e2) /\ respects d D'
  end.

Definition isEqCA D e1 e2 : Prop :=
  forall E d,
    closed (plug E (msubst_exp d e1)) ->
    closed (plug E (msubst_exp d e2)) ->
    respects d D ->
    lc_evctx E ->
    is_cbv_evctx E ->
    (terminates (plug E (msubst_exp d e1)) <-> terminates (plug E (msubst_exp d e2))).

Definition isEqCA_refl : forall D e,
   isEqCA D e e.
Proof. firstorder. Qed.

Definition isEqCA_symm : forall D e1 e2,
  isEqCA D e1 e2 -> isEqCA D e2 e1.
Proof. firstorder. Qed.

Fixpoint msubst_pe (d : msubst) (pe1 : pe) {struct d} := 
  match d with
  | msubst_nil => pe1
  | msubst_cons w x d _ => msubst_pe d (esubst_pe w x pe1)
  end.

Lemma lc_msubst_exp : forall d e,
  lc_exp e ->
  lc_exp (msubst_exp d e).
Proof.
  induction d; simpl; auto using lc_esubst_exp.
Qed.

Lemma lc_msubst_pe : forall d pe1,
  lc_pe pe1 ->
  lc_pe (msubst_pe d pe1).
Proof.
  induction d; simpl; auto with lngen.
Qed.

Lemma is_v_of_exp_msubst_exp : forall d v,
  is_v_of_exp v ->
  is_v_of_exp (msubst_exp d v).
Proof.
  induction d; simpl; auto using is_v_of_exp_esubst_exp.
Qed.


Fixpoint msubst_evctx (d: msubst) (E: evctx) {struct E} : evctx :=
  match E with
    | Ctxhole => Ctxhole
    | Ctxapp1  E e2     => Ctxapp1 (msubst_evctx d E) (msubst_exp d e2) 
    | Ctxapp2  e1 E    => Ctxapp2 (msubst_exp d e1) (msubst_evctx d E) 
    | Ctxpair1 E e2     => Ctxpair1 (msubst_evctx d E) (msubst_exp d e2) 
    | Ctxpair2 e1 E  => Ctxpair2 (msubst_exp d e1) (msubst_evctx d E) 
    | Ctxproj1 E         => Ctxproj1 (msubst_evctx d E)
    | Ctxproj2 E         => Ctxproj2 (msubst_evctx d E)
    | Ctxcase1 E pe1    => Ctxcase1 (msubst_evctx d E) (msubst_pe d pe1)
    | CtxCon C E         => CtxCon C (msubst_evctx d E)
  end.

Lemma msubst_plug_dist_app : forall d e1 e2,
  msubst_exp d (e_App e1 e2) = e_App (msubst_exp d e1) (msubst_exp d e2).
Proof. induction d; simpl; auto. Qed.
Lemma msubst_plug_dist_sie : forall d e1 e2,
  msubst_exp d (e_Sie e1 e2) = e_Sie (msubst_exp d e1) (msubst_exp d e2).
Proof. induction d; simpl; auto. Qed.
Lemma msubst_plug_dist_proj1 : forall d e1,
  msubst_exp d (e_Proj1 e1) = e_Proj1 (msubst_exp d e1).
Proof. induction d; simpl; auto. Qed.
Lemma msubst_plug_dist_proj2 : forall d e1,
  msubst_exp d (e_Proj2 e1) = e_Proj2 (msubst_exp d e1).
Proof. induction d; simpl; auto. Qed.
Lemma msubst_plug_dist_case1 : forall d e1 pe1,
   msubst_exp d (e_Case e1 pe1) = e_Case (msubst_exp d e1) (msubst_pe d pe1).
Proof. induction d; simpl; auto. Qed.
Lemma msubst_plug_dist_con : forall d C e1,
   msubst_exp d (e_Con C e1) = e_Con C (msubst_exp d e1).
Proof. induction d; simpl; auto. Qed.

Hint Rewrite msubst_plug_dist_app msubst_plug_dist_app msubst_plug_dist_sie  
             msubst_plug_dist_sie msubst_plug_dist_proj1 msubst_plug_dist_proj2
             msubst_plug_dist_case1 msubst_plug_dist_con
             : msubst_plug_dist.

Lemma msubst_plug_dist : forall d E e,
  msubst_exp d (plug E e) = plug (msubst_evctx d E) (msubst_exp d e).
Proof.
  induction E; simpl; intros;
   auto; autorewrite with msubst_plug_dist; congruence.
Qed.

Definition closed_pe (pe' : pe) := 
  efv_pe pe' [=] empty.

Fixpoint closed_evctx (E : evctx) {struct E} : Prop :=
  match E  with
  | Ctxhole  => True
  | Ctxapp1  E' e   => closed_evctx E' /\ closed e
  | Ctxapp2  v E'  => closed v /\ closed_evctx E'
  | Ctxpair1 E' e    => closed_evctx E' /\ closed e
  | Ctxpair2 v E' => closed v /\ closed_evctx E'
  | Ctxproj1 E'       => closed_evctx E'
  | Ctxproj2 E'       => closed_evctx E'
  | Ctxcase1 E' pe1 => closed_evctx E'/\ closed_pe pe1
  | CtxCon   C E'     => closed_evctx E'
  end.

Lemma closed_plug_dist_app : forall e1 e2, closed (e_App e1 e2) <-> (closed e1 /\ closed e2).
Proof.
unfold closed; simpl in *.
split; [ intros; split; fsetdec | fsetdec ].
Qed.
Lemma closed_plug_dist_sie : forall e1 e2, closed (e_Sie e1 e2) <-> (closed e1 /\ closed e2).
Proof.
unfold closed; simpl in *.
split; [ intros; split; fsetdec | fsetdec ].
Qed.
Lemma closed_plug_dist_proj1 : forall e, closed (e_Proj1 e) <-> closed e.
Proof.
unfold closed; simpl in *.
tauto.
Qed.
Lemma closed_plug_dist_proj2 : forall e, closed (e_Proj2 e) <-> closed e.
Proof.
unfold closed; simpl in *.
tauto.
Qed.
Lemma closed_plug_dist_case1 : forall e1 pe1, closed (e_Case e1 pe1) <-> (closed e1 /\ closed_pe pe1).
Proof.
unfold closed; unfold closed_pe; simpl in *.
split; [ intros; split; fsetdec | fsetdec ].
Qed.
Lemma closed_plug_dist_con : forall C e, closed (e_Con C e) <-> closed e.
Proof.
unfold closed; simpl in *.
tauto.
Qed.

Lemma closed_plug_dist : forall E e,
  closed (plug E e) <->
  (closed_evctx E) /\ (closed e).
Proof.
  induction E; simpl.
    tauto.
    intros; rewrite closed_plug_dist_app; pose (IHE e0); tauto.
    intros; rewrite closed_plug_dist_app. pose (IHE e0); tauto. 
    intros; rewrite closed_plug_dist_sie; pose (IHE e0); tauto.
    intros; rewrite closed_plug_dist_sie; pose (IHE e0); tauto.
    intros; rewrite closed_plug_dist_proj1; pose (IHE e); tauto.
    intros; rewrite closed_plug_dist_proj2; pose (IHE e); tauto.
    intros; rewrite closed_plug_dist_case1; pose (IHE e); tauto.
    intros; rewrite closed_plug_dist_con; pose (IHE e); tauto.
Qed.

Lemma msubst_exp_closed : forall d e,
  closed e ->
  msubst_exp d e = e.
Proof.
  unfold closed.
  induction d; simpl; intros.
    tauto.
    rewrite esubst_exp_fresh_eq.
      auto.
      fsetdec.
Qed.

Lemma msubst_pe_closed : forall d pe,
  closed_pe pe ->
  msubst_pe d pe = pe.
Proof.
  unfold closed_pe.
  induction d; simpl; intros.
    tauto.
    rewrite esubst_pe_fresh_eq.
      auto.
      fsetdec.
Qed.

Lemma msubst_evctx_closed : forall d E,
  closed_evctx E ->
  msubst_evctx d E = E.
Proof.
  induction E; simpl; intros; 
  try solve [
    auto | 
    destruct H as [H1 H2]; rewrite IHE; auto; rewrite msubst_exp_closed; auto |
    rewrite IHE; auto |
    destruct H as [H1 H2]; rewrite IHE; auto; rewrite msubst_pe_closed; auto
    ].
Qed.

Fixpoint unit_substitution d :=
  match d with
  | msubst_nil => True
  | msubst_cons w x d _ => w = e_unit /\ unit_substitution d
  end.

Fixpoint msubst_domain d :=
  match d with
  | msubst_nil => {}
  | msubst_cons w x d _ => {{ x }} `union` (msubst_domain d)
  end.

Notation diff := AtomSetImpl.diff.

Lemma efv_exp_unit_msubst : forall d e,
  unit_substitution d ->
  efv_exp (msubst_exp d e)  [<=] diff (efv_exp e) (msubst_domain d).
Proof.
  induction d; simpl.
    fsetdec.

    intros e [ Heqw Hclosing]; subst.
    assert (efv_exp (msubst_exp d (esubst_exp e_unit x e))[<=] diff (efv_exp (esubst_exp e_unit x e)) (msubst_domain d)).
      apply IHd; auto.
    assert (efv_exp (esubst_exp e_unit x e)[<=] union (efv_exp e_unit) (remove x (efv_exp e))).
      exact (efv_exp_esubst_exp_upper e e_unit x).
    simpl in *.
    remember (efv_exp (msubst_exp d (esubst_exp e_unit x e))) as F1.
    remember (efv_exp (esubst_exp e_unit x e)) as F2.
    remember (msubst_domain d) as F3.
    remember (efv_exp e)  as F4.
    fsetdec.
Qed.

Lemma unit_pure : pure e_unit.
auto.
Qed.

Fixpoint closing_substitution_for' (xs : list expvar) : msubst :=
  match xs with
  | nil => msubst_nil
  | x::xs => msubst_cons e_unit x (closing_substitution_for' xs) unit_pure
  end.

Lemma closing_substitution_for'_subset : forall (A : atoms),
  A [<=] msubst_domain (closing_substitution_for' (AtomSetImpl.elements A)).
Proof.
  unfold AtomSetImpl.Subset.
  intros A a HIn.
  induction (AtomSetImpl.elements_1 HIn); subst; simpl.
    fsetdec.

    remember (msubst_domain (closing_substitution_for' l)) as F.
    fsetdec.
Qed.

Lemma closing_substitution_for'_unit : forall xs,
  unit_substitution (closing_substitution_for' xs).
Proof.
  induction xs; simpl; auto.
Qed.

Definition closing_substitution_for (xs : atoms) :=
  closing_substitution_for' (AtomSetImpl.elements xs).

Lemma closing_substitution_for_subset : forall (xs : atoms),
  xs [<=] msubst_domain (closing_substitution_for xs).
Proof. 
  unfold closing_substitution_for.
  auto using closing_substitution_for'_subset.
Qed.

Lemma closing_substitution_for_unit : forall xs,
  unit_substitution (closing_substitution_for xs).
Proof.
  unfold closing_substitution_for.
  auto using closing_substitution_for'_unit.
Qed.

Lemma exists_closing : forall e,
  exists d,
   closed (msubst_exp d e).
Proof.
  unfold closed.
  intros e.
  exists (closing_substitution_for (efv_exp e)).
  assert ((efv_exp e) [<=] msubst_domain (closing_substitution_for (efv_exp e))).
    apply closing_substitution_for_subset.
  assert ( efv_exp (msubst_exp (closing_substitution_for (efv_exp e)) e)
          [<=] diff (efv_exp e) (msubst_domain (closing_substitution_for (efv_exp e)))).
    apply efv_exp_unit_msubst. apply closing_substitution_for_unit.
  remember (msubst_exp (closing_substitution_for (efv_exp e)) e) as F1.
  remember (efv_exp e) as F2.
  remember (msubst_domain (closing_substitution_for F2)) as F3.
  fsetdec.
Qed.

Lemma exists_closing2 : forall e1 e2,
  exists d,
   closed (msubst_exp d e1) /\ closed (msubst_exp d e2).
Proof.
  unfold closed.
  intros e1 e2.
  exists (closing_substitution_for ((efv_exp e1) `union` (efv_exp e2))).
  assert (((efv_exp e1) `union` (efv_exp e2)) [<=] msubst_domain (closing_substitution_for ((efv_exp e1) `union` (efv_exp e2)))).
    apply closing_substitution_for_subset.
  assert (efv_exp (msubst_exp (closing_substitution_for((efv_exp e1) `union` (efv_exp e2))) e1)
          [<=] diff (efv_exp e1) (msubst_domain (closing_substitution_for ((efv_exp e1) `union` (efv_exp e2))))).
    apply efv_exp_unit_msubst. apply closing_substitution_for_unit.
  assert (efv_exp (msubst_exp (closing_substitution_for((efv_exp e1) `union` (efv_exp e2))) e2)
          [<=] diff (efv_exp e2) (msubst_domain (closing_substitution_for ((efv_exp e1) `union` (efv_exp e2))))).
    apply efv_exp_unit_msubst. apply closing_substitution_for_unit.
  remember (efv_exp e1) as F1.
  remember (efv_exp e2) as F2.
  remember (msubst_domain (closing_substitution_for (union F1 F2))) as F3.
  remember (efv_exp (msubst_exp (closing_substitution_for (union F1 F2)) e1)) as F4.
  remember (efv_exp (msubst_exp (closing_substitution_for (union F1 F2)) e2)) as F5.
  split; fsetdec.
Qed.

Fixpoint msubst_comp (d1 : msubst) (d2 : msubst) {struct d2} : msubst :=
  match d2 with 
  | msubst_nil => d1
  | msubst_cons w x d H => msubst_cons w x (msubst_comp d1 d) H
  end.

Lemma msubst_comp_spec : forall d1 d2 e,
  (msubst_exp d1 (msubst_exp d2 e)) = (msubst_exp (msubst_comp d1 d2) e).
Proof.
  induction d2.
    simpl. tauto.

    intros e.
    simpl. 
    apply IHd2.
Qed.

Lemma respects_msubst_comp : forall d1 d2 D,
  respects d2 D -> respects (msubst_comp d1 d2) D.
Proof.
  induction D.
    tauto.

    destruct a as (u1,u2).
    simpl.
    intros [Hiseq Hrespects].
    split; [idtac | auto].
      repeat rewrite <- msubst_comp_spec.
      apply isEqC_msubst; auto.
Qed.

Lemma closed_plug_comp : forall E e d' d,
 closed (plug E (msubst_exp d e)) ->
 closed (plug E (msubst_exp (msubst_comp d' d) e)).
Proof.
  intros.
    assert (closed_evctx E) 
      by (rewrite closed_plug_dist in H; destruct H; auto).
    rewrite <- msubst_comp_spec.
    rewrite <- (msubst_evctx_closed d' E); [idtac|auto].
    rewrite <- msubst_plug_dist.
    rewrite msubst_exp_closed; auto.
Qed.

Definition isEqCA_trans : forall D e1 e2 e3,
  isEqCA D e1 e2 -> isEqCA D e2 e3 -> isEqCA D e1 e3.
Proof. 
  unfold isEqCA. intros.
  assert (closed_evctx E)
    by (rewrite closed_plug_dist in H1; destruct H1; auto).
  assert (closed (msubst_exp d e1))
    by (rewrite closed_plug_dist in H1; destruct H1; auto).
  assert (closed (msubst_exp d e3))
    by (rewrite closed_plug_dist in H2; destruct H2; auto).
  destruct (exists_closing (plug E (msubst_exp d e2))) as [d' Hd'closing].
  assert ((plug E (msubst_exp d e1)) = (plug E (msubst_exp (msubst_comp d' d) e1)))
    by (rewrite <- msubst_comp_spec;  rewrite msubst_exp_closed with (d:=d'); auto). 
  assert ((plug E (msubst_exp d e3)) = (plug E (msubst_exp (msubst_comp d' d) e3)))
    by (rewrite <- msubst_comp_spec;  rewrite msubst_exp_closed with (d:=d'); auto). 
  assert (respects (msubst_comp d' d) D)
    by auto using respects_msubst_comp.
  rewrite msubst_plug_dist in Hd'closing. rewrite msubst_evctx_closed in Hd'closing; [idtac|auto]. rewrite msubst_comp_spec in Hd'closing.
  rewrite H9 in *. rewrite H10 in *.
  pose (H E  (msubst_comp d' d)).
  pose (H0 E (msubst_comp d' d)).
  tauto.
Qed.


Lemma respects_weakening : forall d D1 D2 D3, 
  respects d (D1 ++ D2 ++ D3) ->
  respects d (D1 ++ D3).
Proof.
  induction D1.
    induction D2; simpl in *.
      tauto.
      destruct a as (e1,e2); firstorder.

    destruct a as (e1,e2). simpl. firstorder.
Qed.

Lemma isEqCA_weakening  : forall D1 D2 D3 e1 e2,
  isEqCA (D1 ++ D3) e1 e2 ->
  isEqCA (D1 ++ D2 ++ D3) e1 e2.
Proof.
  unfold isEqCA.
  intros D1 D2 D3 e1 e2 Hiseq E d Hclosed1 Hclosed2 Hrespects.
  apply Hiseq;
    auto;
    apply respects_weakening with (D2:=D2); assumption.
Qed. 

Lemma isEqCA_cut' : forall D u1 u2 e1 e2,
  isEqCA D u1 u2 ->
  isEqCA ((u1,u2)::D) e1 e2 ->
  isEqCA D e1 e2.
Proof.
  unfold isEqCA.
  intros D u1 u2 e1 e2 Hiseq_u Hiseq E d Hclosed1 Hclosed2 Hrespects.
  apply Hiseq.
    assumption.
    assumption.
    simpl. split.
      unfold isEqC. intros E' d' Hclosed1' Hclosed2'.
      repeat rewrite msubst_comp_spec.
      apply Hiseq_u.
        rewrite <- msubst_comp_spec; auto.
        rewrite <- msubst_comp_spec; auto.
        apply respects_msubst_comp; assumption.

      assumption.
Qed.

Lemma respects_permute : forall u1 u2 d D1 D2,
  respects d (D2 ++ (u1,u2)::D1) <-> respects d ((u1,u2)::D2++D1).
Proof.
  induction D2.
    simpl. tauto.

    destruct a as (u1',u2').
    simpl in *.  tauto.
Qed.

Lemma isEqCA_permute : forall u1 u2 D1 D2 e1 e2,
  isEqCA(D2 ++ (u1,u2)::D1) e1 e2 <-> isEqCA ((u1,u2)::D2++D1) e1 e2.
Proof.
  unfold isEqCA. split; intros;
    [rewrite <- respects_permute in * | rewrite  respects_permute in *]; firstorder.
Qed.

Lemma isEqCA_cut : forall D D' u1 u2 e1 e2,
  isEqCA D u1 u2 ->
  isEqCA (D' ++ (u1,u2)::D) e1 e2 ->
  isEqCA (D'++ D) e1 e2.
Proof.
  intros.
  apply isEqCA_cut' with (u1:=u1) (u2:=u2).
    apply isEqCA_weakening with (D1:=nil) (D2:=D') (D3:=D); auto.
    rewrite <- isEqCA_permute; auto.
Qed.
    
Lemma respects_invsubst :  forall w x (H : pure w) d D,
  respects d (esubst_ectx w x D) ->
  respects (msubst_cons w x d H) D.
Proof.
  induction D.
    tauto.

    destruct a as (e1, e2).
    simpl. rewrite msubst_extend with (H:=H). tauto.
Qed.

Lemma isEqCA_subst : forall D w x e1 e2,
  pure w ->
  isEqCA D e1 e2 -> 
  isEqCA (esubst_ectx w x D) (esubst_exp w x e1) (esubst_exp w x e2).
Proof.
  unfold isEqCA.
  intros D w x e1 e2 Hpure HisEq E d Hclosed1 Hclose2 Hrespects.
    repeat rewrite msubst_extend with (H:=Hpure).
    apply HisEq. 
      simpl; assumption.
      simpl; assumption.
      apply respects_invsubst; assumption.
Qed.

Lemma isEqCA_beta : forall D e1 e2,
  reduce e1 e2 ->
  isEqCA D e1 e2.
Proof. 
  unfold isEqCA.
  intros.
  apply terminates_wbeta.
    apply wreduce_beta.
      apply reduce_plug; auto.
        apply reduce_msubst; auto.
Qed.

Lemma pure_wreduce : forall w,
  pure w ->
  forall w',
  wreduce w w' ->
  pure w'.
Proof.
  induction 1; inversion 1; subst; auto.
Qed.

Hint Constructors wreduce mwreduce.
Lemma mwreduce_ctr_cong : forall C e e',
  mwreduce e e' ->
  mwreduce (e_Con C e) (e_Con C e').
Proof.  induction 1; eauto. Qed.

Lemma mwreduce_proj_cong1 : forall e e',
  mwreduce e e' ->
  mwreduce (e_Proj1 e) (e_Proj1 e').
Proof. induction 1; eauto. Qed.

Lemma mwreduce_proj_cong2 : forall e e',
  mwreduce e e' ->
  mwreduce (e_Proj2 e) (e_Proj2 e').
Proof. induction 1; eauto. Qed.

Lemma wreduce_pure_size : forall w w',
 wreduce w w' ->
 pure w ->
 size_exp w' < size_exp w.
Proof.
 induction 1; inversion 1; subst;
   try match goal with [H: pure ?u -> size_exp ?e1 < size_exp ?e2 |- _ ] =>
                      assert (size_exp e1 < size_exp e2) by auto
               end;
   simpl; omega.
Qed.

Require Import Classical.

Definition nwsteps_implies_wsteps_prop (n:nat) :=
  forall e e',
    n = size_exp e ->
    nwreduce e e' ->
    exists u, wreduce e u.

Definition lc_nwreduce_prop e e' (WF : nwreduce e e') :=
  lc_exp e /\ lc_exp e'.
Definition lc_nwsteps_prop e (WF : nwsteps e) :=
  lc_exp e.
Lemma lc_nwreduce_mut:
  (forall e e' WF, @lc_nwreduce_prop e e' WF) /\
  (forall e WF, @lc_nwsteps_prop e WF).
Proof.
  apply nwreduce_mutind with
    (P  := lc_nwreduce_prop)
    (P0 := lc_nwsteps_prop);
  unfold lc_nwreduce_prop, lc_nwsteps_prop; intros; subst; simpl; intuition.
    apply lc_find_pe in f.
    auto with lngen.

    apply lc_body_find_pe in f. auto with lngen.
Qed.

Lemma lc_nwreduce_1 : forall e e',
  nwreduce e e' ->
  lc_exp e.
Proof.
  intros.
  pose (J := lc_nwreduce_mut).
  unfold lc_nwreduce_prop in J.
  decompose [and] J; auto.
    apply H0 in H. destruct H; auto.
Qed.

Lemma lc_nwreduce_2 : forall e e',
  nwreduce e e' ->
  lc_exp e'.
Proof.
  intros.
  pose (J := lc_nwreduce_mut).
  unfold lc_nwreduce_prop in J.
  decompose [and] J; auto.
    apply H0 in H. destruct H; auto.
Qed.

Lemma lc_nwsteps : forall e,
  nwsteps e ->
  lc_exp e.
Proof.
  intros.
  pose (J := lc_nwreduce_mut).
  unfold lc_nwsteps_prop in J.
  decompose [and] J; auto.
Qed.

Hint Extern 1 (~nwsteps ?e) =>
  unfold not; inversion 1 as [n n' Hn]; inversion Hn; subst;
   match goal with [H: ~ nwsteps _ |- _ ] => 
                       contradict H; econstructor; eassumption end.

Hint Resolve lc_nwreduce_1 lc_find_pe.

Lemma nwsteps_implies_wsteps' : forall n,
  nwsteps_implies_wsteps_prop n.
Proof.
  intros n. apply lt_wf_ind. clear n.
  unfold nwsteps_implies_wsteps_prop.
  intros n IH e e' Heqn Hnwred; subst.
  inversion Hnwred; subst.
    Case "nwreduce_app_beta".
    destruct (classic (nwsteps w2)) as [F1 | F1]; [inversion F1; subst | idtac].
      assert (size_exp w2 < size_exp (e_App (e_Lam e1) w2)) as Hlt by (simpl; omega).
      destruct (IH _ Hlt _ _ (refl_equal _) H1) as [u Hu].
      exists (e_App (e_Lam e1) u); auto.

      exists (open_exp_wrt_exp e1 w2); auto.
    Case "nwreduce_app_fun_beta".
      destruct (classic (nwsteps w2)) as [F1 | F1]; [inversion F1; subst | idtac].
      assert (size_exp w2 < size_exp (e_App (e_Fix v1) w2)) as Hlt by (simpl; omega).
      destruct (IH _ Hlt _ _ (refl_equal _) H2) as [u Hu].
      exists (e_App (e_Fix v1) u); auto.

      exists (e_App (open_exp_wrt_exp v1 (e_Fix v1))  w2); auto.
    Case "nwreduce_app_cong1".
      assert (size_exp e1 < size_exp (e_App e1 e2)) as Hlt by (simpl;omega).
      destruct (IH _ Hlt _ _ (refl_equal _) H0) as [u Hu].
      exists (e_App u e2); auto.
    Case "nwreduce_app_cong2".
      destruct (classic (nwsteps e1)) as [F1 | F1]; [inversion F1; subst | idtac].
      assert (size_exp e1 < size_exp (e_App e1 e2)) as Hlt by (simpl; omega).
      destruct (IH _ Hlt _ _ (refl_equal _) H1) as [u Hu].
      exists (e_App u e2); apply wreduce_app_cong1. eapply lc_nwreduce_1.  eauto. auto.

      assert (size_exp e2 < size_exp (e_App e1 e2)) as Hlt by (simpl; omega).
      destruct (IH _ Hlt _ _ (refl_equal _) H0) as [u Hu].
      exists (e_App e1 u); auto.
    Case "nwreduce_ctr_cong". 
      inversion Hnwred; simpl.
      assert (size_exp e1 < size_exp (e_Con C e1)) as Hlt by (simpl; omega).
      destruct (IH _ Hlt _ _ (refl_equal _) H1) as [u Hu].
      exists (e_Con C u); auto.
    Case "nwreduce_pair_cong1".
      assert (size_exp e1 < size_exp (e_Sie e1 e2)) as Hlt by (simpl;omega).
      destruct (IH _ Hlt _ _ (refl_equal _) H0) as [u Hu].
      exists (e_Sie u e2); auto.
    Case "nwreduce_pair_cong2".
      destruct (classic (nwsteps e1)) as [F1 | F1]; [inversion F1; subst | idtac].
      assert (size_exp e1 < size_exp (e_Sie e1 e2)) as Hlt by (simpl; omega).
      destruct (IH _ Hlt _ _ (refl_equal _) H1) as [u Hu].
      exists (e_Sie u e2). 
          apply wreduce_pair_cong1. apply lc_nwreduce_1  with (e' := e2'). auto. auto.

      assert (size_exp e2 < size_exp (e_Sie e1 e2)) as Hlt by (simpl; omega).
      destruct (IH _ Hlt _ _ (refl_equal _) H0) as [u Hu].
      exists (e_Sie e1 u); auto.
    Case "nwreduce_proj_beta1".
      destruct (classic (nwsteps e')) as [F1 | F1]; [inversion F1; subst | idtac].
      assert (size_exp e' < size_exp (e_Proj1 (e_Sie e' w2))) as Hlt by (simpl; omega).
      destruct (IH _ Hlt _ _ (refl_equal _) H1) as [u Hu].
      exists (e_Proj1 (e_Sie u w2)); auto.
      
      destruct (classic (nwsteps w2)) as [F2 | F2]; [inversion F2; subst | idtac].
        assert (size_exp w2 < size_exp (e_Proj1 (e_Sie e' w2))) as Hlt by (simpl; omega).
        destruct (IH _ Hlt _ _ (refl_equal _) H1) as [u Hu].
        exists (e_Proj1 (e_Sie e' u)); auto. 
      
        exists e'; auto.
    Case "nwreduce_proj_beta2".
      destruct (classic (nwsteps w1)) as [F1 | F1]; [inversion F1; subst | idtac].
      assert (size_exp w1 < size_exp (e_Proj1 (e_Sie w1 e'))) as Hlt by (simpl; omega).
      destruct (IH _ Hlt _ _ (refl_equal _) H1) as [u Hu].
      exists (e_Proj2 (e_Sie u e')); auto.
      
      destruct (classic (nwsteps e')) as [F2 | F2]; [inversion F2; subst | idtac].
        assert (size_exp e' < size_exp (e_Proj1 (e_Sie w1 e'))) as Hlt by (simpl; omega).
        destruct (IH _ Hlt _ _ (refl_equal _) H1) as [u Hu].
        exists (e_Proj2 (e_Sie w1 u)); auto.
      
        exists e'; auto.
    Case "nwreduce_proj_cong1".
      assert (size_exp e0 < size_exp (e_Proj1 e0)) as Hlt by (simpl; omega).
      destruct (IH _ Hlt _ _ (refl_equal _) H) as [u Hu].
      exists (e_Proj1 u); auto.
    Case "nwreduce_proj_cong2".
      assert (size_exp e0 < size_exp (e_Proj2 e0)) as Hlt by (simpl; omega).
      destruct (IH _ Hlt _ _ (refl_equal _) H) as [u Hu].
      exists (e_Proj2 u); auto.
    Case "nwreduce_case_cong".
      assert (size_exp e0 < size_exp (e_Case e0 pe5)) as Hlt by (simpl; omega).
      destruct (IH _ Hlt _ _ (refl_equal _) H0) as [u Hu].
      exists (e_Case u pe5); auto.
    Case "nwreduce_case_beta".
      destruct (classic (nwsteps w)) as [F1 | F1]; [inversion F1; subst | idtac].
      assert (size_exp w < size_exp (e_Case (e_Con Cj w) pe5)) as Hlt by (simpl; omega).
      destruct (IH _ Hlt _ _ (refl_equal _) H1) as [u Hu].
      exists (e_Case (e_Con Cj u) pe5); apply wreduce_case_cong; eauto.

      exists (open_exp_wrt_exp ej w); auto.
Qed.

Lemma nwsteps_implies_wsteps : forall e e',
    nwreduce e e' ->
    exists u, wreduce e u.
Proof.
  intros.
  pose (H1 := nwsteps_implies_wsteps' (size_exp e)).
  unfold nwsteps_implies_wsteps_prop in *. 
  eauto.
Qed.
  
Definition pure_wtermination_prop (n:nat) :=
  forall w,
    n = size_exp w ->
    pure w ->
    exists u, mwreduce w u /\ pure u /\ ~nwsteps u.

Lemma pure_wtermination' : forall n,
  pure_wtermination_prop n.
Proof.
  intros n. apply lt_wf_ind. clear n.
  unfold pure_wtermination_prop.
  intros n IH w Heqn Hpure; subst.
  destruct (classic (exists w', wreduce w w')) as [[w' Hsteps] | Hstuck].
    assert (size_exp w' < size_exp w) as Hlt by auto using wreduce_pure_size.
    destruct (IH _ Hlt w') as [u [Hmwred_u [Hpure_u Hstuck_u]]]. 
      auto. 
      eapply pure_wreduce; eauto.
    exists u; eauto.

    exists w. split. auto. split. auto.
      unfold not; inversion 1; subst.
      contradict Hstuck. eauto using nwsteps_implies_wsteps.
Qed.

Lemma pure_wtermination : forall w,
  pure w ->
  exists u, mwreduce w u /\ pure u /\ ~nwsteps u.
Proof.
  intros.
  pose (pure_wtermination' (size_exp w)).
  unfold pure_wtermination_prop in *.
  auto.  
Qed.

Definition omega : exp :=
  e_App (e_Fix (e_Lam (e_App (e_Var_b 1) (e_Var_b 0)))) e_unit.

Hint Extern 2 (lc_exp (open_exp_wrt_exp _ _)) =>
  compute; auto with lngen : lngen.

Lemma wreduce_mwreduce : forall e1 e2,
  wreduce e1 e2 ->
  mwreduce e1 e2.
Proof.
  intros.
  apply mwreduce_step with (e':=e2); auto.
    apply mwreduce_refl; auto.
      apply lc_wreduce2 in H; auto.
Qed.

Definition omega' : exp :=
  (e_App
     (e_Lam
        (e_App (e_Fix (e_Lam (e_App (e_Var_b 1) (e_Var_b 0)))) (e_Var_b 0)))
      e_unit).

Lemma omega_wreduce_to_omega' :
  wreduce omega omega'.
Proof.
    assert (wreduce omega (e_App 
                                (open_exp_wrt_exp 
                                   (e_Lam (e_App (e_Var_b 1) (e_Var_b 0))) 
                                   (e_Fix (e_Lam (e_App (e_Var_b 1) (e_Var_b 0))))
                                )
                                e_unit) = 
            wreduce omega omega') as EQ.
      unfold open_exp_wrt_exp. simpl. auto.
    
    rewrite <- EQ.
    apply wreduce_app_fun_beta; simpl; auto with lngen.
Qed.

Lemma omega'_wreduce_to_omega :
  wreduce omega' omega.
Proof.
    assert (wreduce omega' omega =
            wreduce (e_App
                       (e_Lam
                          (e_App (e_Fix (e_Lam (e_App (e_Var_b 1) (e_Var_b 0)))) (e_Var_b 0)))
                        e_unit) 
                    (open_exp_wrt_exp 
                       (e_App (e_Fix (e_Lam (e_App (e_Var_b 1) (e_Var_b 0)))) (e_Var_b 0))
                       e_unit) 
                    ) as EQ.
      unfold open_exp_wrt_exp. unfold omega. simpl. auto.
    rewrite EQ.
    apply wreduce_app_beta; simpl; auto with lngen.
Qed.

Lemma omega_mwreduce_to_omega :
  mwreduce omega omega.
Proof.
  apply mwreduce_step with (e':=omega').
    apply omega_wreduce_to_omega'.
    apply wreduce_mwreduce.
      apply omega'_wreduce_to_omega.
Qed.
    
Lemma omega'_mwreduce_to_omega' :
  mwreduce omega' omega'.
Proof.
  apply mwreduce_step with (e':=omega).
    apply omega'_wreduce_to_omega.
    apply wreduce_mwreduce.
      apply omega_wreduce_to_omega'.
Qed.

Lemma omega_only_reduce_to_omega' : forall e,
  wreduce omega e -> e = omega'.
Proof.
  intros e H.
  assert (J := omega_wreduce_to_omega').
  apply wreduce_deterministic with (e:=omega); auto.
Qed.

Lemma omega'_only_reduce_to_omega : forall e,
  wreduce omega' e -> e = omega.
Proof.
  intros e H.
  assert (J := omega'_wreduce_to_omega).
  apply wreduce_deterministic with (e:=omega'); auto.
Qed.

Lemma omega_always_mreduce_to_omega_aux : forall e e',
  mwreduce e e' -> 
  (e = omega \/ e = omega') ->
  (e' = omega \/ e' = omega').
Proof.
  intros e e' MReduce H.
  induction MReduce; auto.
    destruct H as [H | H]; subst.
      apply omega_only_reduce_to_omega' in H0; subst.
      apply IHMReduce; auto.

      apply omega'_only_reduce_to_omega in H0; subst.
      apply IHMReduce; auto.
Qed.

Lemma omega_always_mreduce_to_o : forall e,
  mwreduce omega e -> 
  (e = omega \/ e = omega').
Proof.
  intros.
  apply omega_always_mreduce_to_omega_aux with (e:=omega); auto.
Qed.

Lemma omega'_always_mreduce_to_o : forall e,
  mwreduce omega' e -> 
  (e = omega \/ e = omega').
Proof.
  intros.
  apply omega_always_mreduce_to_omega_aux with (e:=omega'); auto.
Qed.

Lemma omega_diverges : ~terminates omega.
Proof.
  intros J.
  inversion J.
  destruct H as [H1 H2].
  apply omega_always_mreduce_to_o in H1.
  destruct H1; subst.
    apply (@H2 omega').
      apply omega_wreduce_to_omega'.
    apply (@H2 omega).
      apply omega'_wreduce_to_omega.
Qed.

Lemma unit_terminates : terminates e_unit.
Proof.
unfold terminates. exists e_unit. split. 
  auto.
  unfold not. inversion 1.
Qed.

Lemma lc_plug : forall E e,
  lc_exp e ->
  lc_evctx E ->
  lc_exp (plug E e).
Proof.
 induction E; intros; inversion H0; simpl; auto.
Qed.


Hint Resolve lc_plug.
Hint Constructors mwreduce.
Lemma mwreduce_plug : forall E e e',
  lc_evctx E->
  is_cbv_evctx E ->
  mwreduce e e' ->
  mwreduce (plug E e) (plug E e').
Proof.
induction 3; eauto using wreduce_plug. 
Qed.

Definition ctx1 (C1 C2 : contmvar) : evctx.
intros.
refine (Ctxcase1 Ctxhole (pe_Cons (peb_Branch C1 omega) 
                         (pe_Cons (peb_Branch C2 e_unit) pe_Empty))
                         ).
Defined.

Hint Constructors lc_evctx is_cbv_evctx.

Lemma lc_evctx_ctx1 : forall C1 C2, 
 lc_evctx (ctx1 C1 C2).
Proof.
 intros; unfold ctx1; eauto.
 apply lc_Ctxcase1; auto. apply lc_pe_Cons ; auto. apply lc_peb_Branch; auto.
 intros. auto with lngen.
Qed.

Lemma is_cbv_evctx_ctx1: forall C1 C2,
 is_cbv_evctx (ctx1 C1 C2).
Proof. 
 intros; unfold ctx1; eauto.
Qed.

Lemma ctx1_steps_1 : forall (C1 C2 : contmvar) w,
  pure w ->
  ~ (nwsteps w) ->
  wreduce (plug (ctx1 C1 C2) (e_Con C1 w)) omega.
Proof.
  intros C1 C2 w Pw Stuckw.
  simpl.
  assert (wreduce (e_Case (e_Con C1 w) (pe_Cons (peb_Branch C1 omega) (pe_Cons (peb_Branch C2 e_unit) pe_Empty))) omega
        = wreduce (e_Case (e_Con C1 w) (pe_Cons (peb_Branch C1 omega) (pe_Cons (peb_Branch C2 e_unit) pe_Empty))) (open_exp_wrt_exp omega w)) as EQ. auto.
  rewrite EQ.
  apply wreduce_case_beta; auto.
    apply pe_here; auto with lngen.
Qed.

Lemma ctx1_steps_2 : forall (C1 C2 : contmvar) w,
  pure w ->
  ~ (nwsteps w) ->
  C1 <> C2 -> 
  wreduce (plug (ctx1 C1 C2) (e_Con C2 w)) e_unit.
Proof.
  intros C1 C2 w Pw Stuckw NotEqC.
  simpl.
  assert (wreduce (e_Case (e_Con C2 w) (pe_Cons (peb_Branch C1 omega) (pe_Cons (peb_Branch C2 e_unit) pe_Empty))) e_unit
        = wreduce (e_Case (e_Con C2 w) (pe_Cons (peb_Branch C1 omega) (pe_Cons (peb_Branch C2 e_unit) pe_Empty))) (open_exp_wrt_exp e_unit w)) as EQ. auto.
  rewrite EQ.
  apply wreduce_case_beta; auto.
    apply pe_there. 
      apply pe_here; auto with lngen.
      apply notmatching_peb_intro; auto with lngen.
Qed.

Lemma ctx1_closed : forall C1 C2 e,
  closed e ->
  closed (plug (ctx1 C1 C2) e).
Proof.
  intros C1 C2 e H.
  eapply closed_plug_dist.
    split; simpl; auto.
      split; auto.
        unfold closed_pe. simpl. fsetdec.
Qed.

Lemma msubst_exp_ctr_comm : forall d C e,
  msubst_exp d (e_Con C e) = e_Con C (msubst_exp d e).
Proof.
  intros d.
  induction d; intros; simpl; auto.
Qed.    


Lemma isEqCA_empty : forall C1 C2 w1 w2, 
  pure w1 -> pure w2 -> 
  C1 <> C2 -> 
  ~(isEqCA nil (e_Con C1 w1)(e_Con C2 w2)).
Proof.
  unfold not; intros C1 C2 w1 w2 Hpure_w1 Hpure_w2 Hneq Hiseq.
  destruct (exists_closing2 (e_Con C1 w1) (e_Con C2 w2)) as [d [Hclosing1 Hclosing2]].
  pose (H:= Hiseq (ctx1 C1 C2) d (ctx1_closed C1 C2 _ Hclosing1) (ctx1_closed C1 C2 _ Hclosing2) I
             (lc_evctx_ctx1 C1 C2) (is_cbv_evctx_ctx1 C1 C2)).
  repeat rewrite msubst_exp_ctr_comm in H.
  destruct (pure_wtermination (msubst_exp d w1) (pure_msubst d _ Hpure_w1)) as [u1 [Hreduce_u1 [Hpure_u1 Hstuck_u1]]].
  destruct (pure_wtermination (msubst_exp d w2) (pure_msubst d _ Hpure_w2)) as [u2 [Hreduce_u2 [Hpure_u2 Hstuck_u2]]].
  rewrite (terminates_mwbeta _ _  (mwreduce_plug (ctx1 C1 C2) _ _ (lc_evctx_ctx1 C1 C2) (is_cbv_evctx_ctx1 C1 C2) (mwreduce_ctr_cong C1 _ _ Hreduce_u1))) in H.
  rewrite (terminates_mwbeta _ _  (mwreduce_plug (ctx1 C1 C2) _ _ (lc_evctx_ctx1 C1 C2) (is_cbv_evctx_ctx1 C1 C2) (mwreduce_ctr_cong C2 _ _ Hreduce_u2))) in H.
  rewrite (terminates_wbeta _ _ (ctx1_steps_1 C1 C2  _ Hpure_u1 Hstuck_u1)) in H.
  rewrite (terminates_wbeta _ _ (ctx1_steps_2 C1 C2  _ Hpure_u2 Hstuck_u2 Hneq)) in H.
  pose omega_diverges; pose unit_terminates; intuition.
Qed.

Fixpoint evctx_com (E1 : evctx) (E2 : evctx) {struct E1} : evctx :=
  match E1  with
  | Ctxhole  => E2
  | Ctxapp1  E1' e   => Ctxapp1 (evctx_com E1' E2) e
  | Ctxapp2  v E1' => Ctxapp2 v (evctx_com E1' E2) 
  | Ctxpair1 E1' e    => Ctxpair1 (evctx_com E1' E2) e 
  | Ctxpair2 v E1' => Ctxpair2 v (evctx_com E1' E2) 
  | Ctxproj1 E1'       => Ctxproj1 (evctx_com E1' E2)
  | Ctxproj2 E1'       => Ctxproj2 (evctx_com E1' E2)
  | Ctxcase1 E1' pe1  => Ctxcase1 (evctx_com E1' E2) pe1
  | CtxCon C E1'       => CtxCon C (evctx_com E1' E2)
  end.

Definition ctx2 (C : contmvar) : evctx.
intros.
refine (Ctxcase1 Ctxhole (pe_Cons (peb_Branch C (e_Var_b 0)) pe_Empty)
                         ).
Defined.

Lemma lc_evctx_ctx2 : forall C, 
 lc_evctx (ctx2 C).
Proof.
 intros; unfold ctx2; eauto.
 apply lc_Ctxcase1; auto. apply lc_pe_Cons ; auto. apply lc_peb_Branch; auto.
 intros. auto with lngen.
Qed.

Lemma is_cbv_evctx_ctx2: forall C,
 is_cbv_evctx (ctx2 C).
Proof. 
 intros; unfold ctx2; eauto.
Qed.

Lemma case2_steps : forall (C : contmvar) w,
  pure w ->
  ~ (nwsteps w) ->
  wreduce (plug (ctx2 C) (e_Con C w)) w.
Proof.
  intros C w Pw Stuckw. simpl.
  assert (wreduce (e_Case (e_Con C w) (pe_Cons (peb_Branch C (e_Var_b 0)) pe_Empty)) w
        = wreduce (e_Case (e_Con C w) (pe_Cons (peb_Branch C (e_Var_b 0)) pe_Empty)) (open_exp_wrt_exp (e_Var_b 0) w)) as EQ. auto.
  rewrite EQ.
  apply wreduce_case_beta; auto.
    apply pe_here; auto with lngen.
Qed.

Lemma wreduce_evctx_com : forall (E E' : evctx) (e e' : exp),
  wreduce (plug E' e) e' ->
  lc_evctx E ->
  is_cbv_evctx E ->
  wreduce (plug (evctx_com E E') e) (plug E e').
Proof. 
  induction E; intros E' e1 e1' H Hlc Hcbv; inversion Hlc; inversion Hcbv; subst; simpl; auto.
Qed.

Lemma ctx2_steps : forall (E : evctx) (C : contmvar) w,
  pure w ->
  ~ (nwsteps w) ->
  lc_evctx E ->
  is_cbv_evctx E ->
  wreduce (plug (evctx_com E (ctx2 C)) (e_Con C w)) (plug E w).
Proof.
  intros E C w Pw Stuckw Hlc Hcbv.
  apply wreduce_evctx_com; auto.
    apply case2_steps; auto.
Qed.

Lemma mwreduce_trans : forall e1 e2 e3,
  mwreduce e1 e2 ->
  mwreduce e2 e3 ->
  mwreduce e1 e3.
Proof.
  induction 1; [ tauto | eauto].
Qed.

Lemma closed_evctx_com : forall E E',
  closed_evctx (evctx_com E E') <-> closed_evctx E /\ closed_evctx E'.
Proof.
  induction E; intros E'; simpl; split; intro J.
    split; auto.
    destruct J; auto.
    destruct J as [J1 J2]. apply IHE in J1. destruct J1 as [J11 J12]. split; auto.
    destruct J as [[J11 J12] J2]. split; auto. eapply IHE; eauto.
    destruct J as [J1 J2]. apply IHE in J2. destruct J2 as [J21 J22]. split; auto.
    destruct J as [[J11 J12] J2]. split; auto. eapply IHE; eauto.
    destruct J as [J1 J2]. apply IHE in J1. destruct J1 as [J11 J12]. split; auto.
    destruct J as [[J11 J12] J2]. split; auto. eapply IHE; eauto.
    destruct J as [J1 J2]. apply IHE in J2. destruct J2 as [J21 J22]. split; auto.
    destruct J as [[J11 J12] J2]. split; auto. eapply IHE; eauto.
    apply IHE in J. destruct J as [J1 J2]. split; auto.
    apply IHE in J. auto. 
    apply IHE in J. auto.
    apply IHE in J. auto.
    destruct J as [J1 J2]. apply IHE in J1. destruct J1 as [J11 J12]. split; auto.
    destruct J as [[J11 J12] J2]. split; auto. eapply IHE; eauto.
    apply IHE in J. auto.
    apply IHE in J. auto. 
Qed.


Lemma lc_evctx_com : forall E E',
  lc_evctx E ->
  lc_evctx E' ->
  lc_evctx (evctx_com E E').
Proof.
  intros E E' lcE lcE'.
  generalize dependent E'.
  induction lcE; intros; simpl; auto.
Qed.    

Lemma is_cbv_evctx_com : forall E E',
  is_cbv_evctx E ->
  is_cbv_evctx E' ->
  is_cbv_evctx (evctx_com E E').
Proof.
  intros E E' HE HE'.
  generalize dependent E'.
  induction HE; intros; simpl; auto.
Qed.    

Lemma injection_isEqCA : forall D C w1 w2,
    pure w1 ->
    pure w2 ->
    isEqCA D (e_Con C w1) (e_Con C w2) ->
    isEqCA D w1 w2.
Proof. 
  intros D C w1 w2 Pw1 Pw2 IsEq.
  intros E d Closed1 Closed2 Respect Lc Cbv.
  assert (closed (plug (evctx_com E (ctx2 C)) (msubst_exp d (e_Con C w1)))) as Closed1'.
    eapply closed_plug_dist.
      apply closed_plug_dist in Closed1. 
      destruct Closed1 as [J1 J2].
      split.
         eapply closed_evctx_com. split; auto. 
           simpl; split; auto.
           unfold closed_pe. simpl. fsetdec. 
         rewrite msubst_exp_ctr_comm.         
         unfold closed. simpl. auto.
  assert (closed (plug (evctx_com E (ctx2 C)) (msubst_exp d (e_Con C w2)))) as Closed2'.
    eapply closed_plug_dist.
      apply closed_plug_dist in Closed2. 
      destruct Closed2 as [J1 J2].
      split.
         eapply closed_evctx_com. split; auto. 
           simpl; split; auto.
           unfold closed_pe. simpl. fsetdec. 
         rewrite msubst_exp_ctr_comm.         
         unfold closed. simpl. auto.
  assert (J := @IsEq (evctx_com E (ctx2 C)) d Closed1' Closed2' Respect).
  assert ((msubst_exp d (e_Con C w1)) = (e_Con C (msubst_exp d w1))) as EQ1.
    auto using msubst_exp_ctr_comm.
  assert ((msubst_exp d (e_Con C w2)) = (e_Con C (msubst_exp d w2))) as EQ2.
    auto using msubst_exp_ctr_comm.
  rewrite EQ1 in J. rewrite EQ2 in J.
  assert (Psw1 := @pure_msubst d w1 Pw1).
  assert (Psw2 := @pure_msubst d w2 Pw2).
  destruct (@pure_wtermination (msubst_exp d w1) Psw1) as [u1 [Reduce1 [Pu1 Stuck1]]].
  destruct (@pure_wtermination (msubst_exp d w2) Psw2) as [u2 [Reduce2 [Pu2 Stuck2]]].
  assert (mwreduce (plug (evctx_com E (ctx2 C)) (e_Con C (msubst_exp d w1))) (plug E u1)) as MReduce1.
    apply mwreduce_trans with (e2:=(plug (evctx_com E (ctx2 C)) (e_Con C u1))).
      apply mwreduce_plug; auto.
        apply lc_evctx_com; auto.
          apply lc_evctx_ctx2.
        apply is_cbv_evctx_com; auto.
          apply is_cbv_evctx_ctx2.
        apply mwreduce_ctr_cong; auto.
      apply wreduce_mwreduce; auto.
        apply ctx2_steps; auto.
  assert (mwreduce (plug (evctx_com E (ctx2 C)) (e_Con C (msubst_exp d w2))) (plug E u2)) as MReduce2.
    apply mwreduce_trans with (e2:=(plug (evctx_com E (ctx2 C)) (e_Con C u2))).
      apply mwreduce_plug; auto.
        apply lc_evctx_com; auto.
          apply lc_evctx_ctx2.
        apply is_cbv_evctx_com; auto.
          apply is_cbv_evctx_ctx2.
        apply mwreduce_ctr_cong; auto.
      apply wreduce_mwreduce; auto.
        apply ctx2_steps; auto.
  assert (mwreduce (plug E (msubst_exp d w1)) (plug E u1)) as MReduce1'.
    apply mwreduce_plug; auto.
  assert (mwreduce (plug E (msubst_exp d w2)) (plug E u2)) as MReduce2'.
    apply mwreduce_plug; auto.
  assert (T1 := @terminates_mwbeta (plug (evctx_com E (ctx2 C)) (e_Con C (msubst_exp d w1))) 
                                  (plug E u1)
                                  MReduce1).
  assert (T2 := @terminates_mwbeta (plug (evctx_com E (ctx2 C)) (e_Con C (msubst_exp d w2))) 
                                  (plug E u2)
                                  MReduce2).
  assert (T1' := @terminates_mwbeta (plug E (msubst_exp d w1)) 
                                  (plug E u1)
                                  MReduce1').
  assert (T2' := @terminates_mwbeta (plug E (msubst_exp d w2)) 
                                  (plug E u2)
                                  MReduce2').
  split; intro H.
    apply T2'. apply T2. apply J.
        apply lc_evctx_com; auto.
          apply lc_evctx_ctx2.
        apply is_cbv_evctx_com; auto.
          apply is_cbv_evctx_ctx2.
    apply T1. apply T1'; auto.
    apply T1'. apply T1. apply J. 
        apply lc_evctx_com; auto.
          apply lc_evctx_ctx2.
        apply is_cbv_evctx_com; auto.
          apply is_cbv_evctx_ctx2.
    apply T2. apply T2'; auto.
Qed.

Inductive ctxereq : ectx -> ectx -> Prop :=    (* defn ctxereq *)
  | CEQ_Empty : 
    ctxereq  nil   nil 
  | CEQ_Eq : forall (D:ectx) (e1 e2:exp) (D':ectx) (e1' e2':exp),
    ctxereq D D' ->
    isEqCA  D   e1   e1'  ->
    isEqCA  D   e2   e2'  ->
    ctxereq   ( (  e1  ,  e2  ) ::  D  )     ( (  e1'  ,  e2'  ) ::  D'  ).  

Lemma ctx_conv_respects: forall d D1 D2,
 respects d D1 -> ctxereq D2 D1 -> respects d D2.
Proof. 
 intros d D1 D2 Hresp Heq. generalize dependent d.
 induction Heq; intros; auto.
 simpl in Hresp. destruct Hresp.
 simpl. split; eauto. 
 unfold isEqCA in H; unfold isEqCA in H0.
 unfold isEqC in H1; unfold isEqC; intros.
 assert (JcE: closed_evctx E). eapply closed_plug_dist; eauto.
 assert (Jce1': closed (msubst_exp d0 (msubst_exp d e1))). eapply closed_plug_dist; eauto.
 assert (Jce2': closed (msubst_exp d0 (msubst_exp d e2))). eapply closed_plug_dist; eauto.

 destruct (exists_closing  (msubst_exp d0 (msubst_exp d e1'))) as [d1 Jc1].
 destruct (exists_closing  (msubst_exp d1 (msubst_exp d0 (msubst_exp d e2')))) as [d1' Jc2].
 rewrite <- (@msubst_exp_closed d1' (msubst_exp d1 (msubst_exp d0 (msubst_exp d e1')))) in Jc1; eauto.
 repeat rewrite msubst_comp_spec in Jc1.
 repeat rewrite msubst_comp_spec in Jc2.
 destruct (H1 E (msubst_comp (msubst_comp d1' d1) d0)); eauto.
 repeat rewrite msubst_comp_spec; auto. 
 eapply closed_plug_dist; split; eauto.
 repeat rewrite msubst_comp_spec; auto.
 eapply closed_plug_dist; split; eauto. 

 repeat rewrite msubst_comp_spec in H7. repeat rewrite msubst_comp_spec in H8.
 destruct (H E (msubst_comp (msubst_comp (msubst_comp d1' d1) d0) d)); eauto.
 eapply closed_plug_dist; split; eauto.
 repeat rewrite <- msubst_comp_spec; auto.  
 rewrite msubst_exp_closed; auto.  rewrite msubst_exp_closed; auto.
 rewrite msubst_exp_closed; auto.
 eapply closed_plug_dist; split; eauto. 
 
 apply  respects_msubst_comp; auto.

 destruct (H0 E (msubst_comp (msubst_comp (msubst_comp d1' d1) d0) d)); eauto.
 eapply closed_plug_dist; split; eauto.
 repeat rewrite <- msubst_comp_spec; auto. 
 rewrite msubst_exp_closed; auto.  rewrite msubst_exp_closed; auto. 
 rewrite msubst_exp_closed; auto. 
 eapply closed_plug_dist; split; eauto. 

 apply  respects_msubst_comp; auto.
 rewrite <- (@msubst_exp_closed d1 (msubst_exp d0 (msubst_exp d e1))); auto.
 rewrite <- (@msubst_exp_closed d1' (msubst_exp d1 (msubst_exp d0 (msubst_exp d e1)))); auto.
 rewrite <- (@msubst_exp_closed d1 (msubst_exp d0 (msubst_exp d e2))); auto.
 rewrite <- (@msubst_exp_closed d1' (msubst_exp d1 (msubst_exp d0 (msubst_exp d e2)))); auto.
 repeat rewrite msubst_comp_spec; eauto.
 split; intros; eauto. 
 rewrite msubst_exp_closed ; auto.  
 rewrite msubst_exp_closed ; auto. 
Qed. 

Lemma ctx_conv_isEqCA : 
 forall D1 D2 e1 e2,
  ctxereq D1 D2 -> isEqCA D1 e1 e2 -> isEqCA D2 e1 e2.
Proof.
 intros. 
 unfold isEqCA; intros.
 unfold isEqCA in H0.
 destruct (H0 E d); eauto.
 eapply ctx_conv_respects; eauto.
 split; auto.
Qed.

Lemma inversion_respect: forall D1 d D2 e1 e2,
 respects d (D1 ++ (e1, e2) ::D2) ->
 isEqC (msubst_exp d e1) (msubst_exp d e2).
Proof.
 intro D1. 
 induction D1; intros; auto.
 
 simpl in H. destruct H; auto.
 destruct a. 
 replace  (((e, e0) :: D1) ++ (e1, e2) :: D2) with
    ((e,e0):: (D1 ++ (e1, e2)::D2)) in H; auto.
 simpl in H. destruct H; eauto.
Qed.
 
Lemma assumption_isEqCA: forall D1 D2 e1 e2,
 isEqCA (D1 ++ (e1, e2)::D2) e1 e2.
Proof.
 intros. unfold isEqCA. intros.
 assert (J: isEqC (msubst_exp d e1) (msubst_exp d e2)).
 eapply inversion_respect; eauto.
 unfold isEqC in J. 
 destruct (J E msubst_nil); auto.
 simpl in H4; simpl in H5; intuition.
Qed.
