Add LoadPath "./metatheory".

Require Import Metatheory.
Require Import lang.
Require Import langExtra.

Module Type ISEQ.

  Parameter isEq  : ectx -> exp -> exp -> Prop.
 
  Inductive ctxereq : ectx -> ectx -> Prop :=    (* defn ctxereq *)
  | CEQ_Empty : 
    ctxereq  nil   nil 
  | CEQ_Eq : forall (D:ectx) (e1 e2:exp) (D':ectx) (e1' e2':exp),
    ctxereq D D' ->
    isEq  D   e1   e1'  ->
    isEq  D   e2   e2'  ->
    ctxereq   ( (  e1  ,  e2  ) ::  D  )     ( (  e1'  ,  e2'  ) ::  D'  ).  

  Axiom lc_isEq : forall D e1 e2, 
    isEq D e1 e2 -> 
    lc_ectx D /\ lc_exp e1 /\ lc_exp e2.
  
  Axiom ref_isEq : forall D e, 
    lc_ectx D -> lc_exp e ->
    isEq D e e.
  Axiom sym_isEq : forall D e1 e2, 
    isEq D e1 e2 -> isEq D e2 e1.
  Axiom trans_isEq : forall D e1 e2 e3, 
    isEq D e1 e2 -> isEq D e2 e3 -> isEq D e1 e3.
  Axiom beta_isEq : forall e1 e2, 
    reduce e1 e2 -> isEq nil e1 e2.
  Axiom empty_isEq : forall C1 C2 e1 e2, 
    pure e1 -> pure e2 -> 
    C1 <> C2 -> ~(isEq nil (e_Con C1 e1)(e_Con C2 e2)).
  Axiom injection2_isEq : forall D C w1 w2,
    pure w1 ->
    pure w2 ->
    isEq D (e_Con C w1) (e_Con C w2) ->
    isEq D w1 w2.

  Axiom ctx_conv_isEq : 
    forall D1 D2 e1 e2,
      ctxereq D1 D2 -> isEq D1 e1 e2 -> isEq D2 e1 e2.
  Axiom cut_isEq : 
    forall D1 D2 u1 u2 e1 e2, 
      isEq (D2 ++ (u1, u2) :: D1) e1 e2 -> 
      isEq D1 u1 u2 -> 
      isEq (D2 ++ D1) e1 e2.
  Axiom weakening_isEq : 
    forall D1 D2 D3 e1 e2, 
      lc_ectx D2 ->
      isEq (D1 ++ D3) e1 e2 ->
      isEq (D1 ++ D2 ++ D3) e1 e2.
  Axiom subst_isEq :
    forall D e1 e2 w x,
      pure w ->
      isEq D e1 e2 -> 
      isEq (esubst_ectx w x D)
           (esubst_exp w x e1)
           (esubst_exp w x e2).
  
End ISEQ.
