(** Extra definitions / reasoning about the 
   syntax of the language, not 
   generated by OTT, but perhaps should be. **)

Add LoadPath "./metatheory".

Set Implicit Arguments.

Require Import Metatheory.
Require Import lang.

Fixpoint efv_ectx (D : ectx) : atoms :=
  match D with 
    | nil => empty
    | (e1 , e2) :: D' => efv_exp e1 \u efv_exp e2 \u efv_ectx D'
  end.

Ltac gather_atoms :=
  let A := ltac_map (fun x : vars => x) in
  let B := ltac_map (fun x : var => {{x}}) in
  let C := ltac_map (fun x : ctx => dom x) in
  let D := ltac_map (fun x => efv_exp x) in
  let E := ltac_map (fun x => efv_peb x) in
  let F := ltac_map (fun x => efv_pe x) in
  let G := ltac_map (fun x => efv_type x) in
  let H := ltac_map (fun x => efv_ptt x) in
  let I := ltac_map (fun x => efv_pt x) in
  let J := ltac_map (fun x => efv_kind x) in
  let K := ltac_map (fun x => efv_bind x) in
  let L := ltac_map (fun x : ectx => efv_ectx x) in
  simplify_list_of_atom_sets (A ++ B ++ C ++ D ++ E ++ F ++ G ++ H ++ I ++ J ++ K ++ L).

Tactic Notation "pick" "fresh" ident(x) :=
  let L := gather_atoms in (pick fresh x for L).

Tactic Notation
      "pick" "fresh" ident(atom_name) "and" "apply" constr(lemma) :=
  let L := gather_atoms in
  pick fresh atom_name excluding L and apply lemma.

(* Extra lemma about environments. *)

Lemma erase_app : forall A B (G1 G2 : list (AssumeList.asn A B)), erase (G1 ++ G2) = erase G1 ++ erase G2.
intros. induction G1. simpl. auto.
simpl. destruct a. auto. rewrite IHG1. simpl. auto.
Qed.

Hint Rewrite erase_app : rewr_list.

(*****)

Fixpoint lc_ectx (D : ectx) : Prop :=
  match D with 
    | nil => True 
    | (e1 , e2) :: D' => lc_exp e1 /\ lc_exp e2 /\ lc_ectx D' 
  end.

Fixpoint lc_ctx (G : ctx): Prop :=
  match G with 
  | nil => True
  | VarAsn x t :: G' => lc_type t /\ lc_ctx G'
  | AltAsn (e1 , e2) :: G' => lc_exp e1 /\ lc_exp e2 /\ lc_ctx G'
  end.

Lemma lc_erase : forall G, lc_ctx G -> 
  lc_ectx (erase G).
Proof.
 induction G. auto. destruct a. simpl. intuition.  
 simpl. destruct p. intuition.
Qed.  

Definition esubst_exp2 : exp -> atom -> exp * exp -> exp * exp :=
  fun e' x p => 
    let (e1, e2) := p in 
      (esubst_exp e' x e1, esubst_exp e' x e2).

Definition esubst_G : exp -> atom -> ctx -> ctx :=
  fun e' x G =>
    map (esubst_type e' x) (esubst_exp2 e' x) G.

Definition esubst_ectx : exp -> atom -> ectx -> ectx :=
  fun e' x ectx => 
  List.map (esubst_exp2 e' x) ectx.

Fixpoint efv_ctx (G1 : ctx) : atoms :=
  match G1 with 
    | nil => empty
    | VarAsn x t1 :: G2       => efv_type t1 \u efv_ctx G2
    | AltAsn (e1 , e2) :: G2 => efv_exp e1 \u efv_exp e2 \u efv_ctx G2
  end.

Lemma notin_efv_ctx__notin_efv_ectx : forall x G,
  x `notin` (efv_ctx G) ->
  x `notin` (efv_ectx (erase G)).
intros x G xnotinG.
  induction G; simpl in *; auto.
  destruct a; auto.
  destruct p; auto.
Qed.

(*********************************************************************)
(* INDUCTION SCHEMES *)

Scheme wfk_ind := Induction for wfk Sort Prop
  with wft_ind := Induction for wft Sort Prop
  with wfpt_ind := Induction for wfpt Sort Prop
  with wfe_ind := Induction for wfe Sort Prop
  with wfpe_ind := Induction for wfpe Sort Prop
  with wfc_ind := Induction for wfc Sort Prop.

Combined Scheme wf_mutind from wfk_ind, wft_ind, 
  wfpt_ind, wfe_ind, wfpe_ind, wfc_ind.


Tactic Notation "wf_mutind_cases" tactic(first) tactic(c) :=
  first;
  [ c "WFK_Type" | c "WFK_Pi" | c "K_InCon" |
    c "K_Unit" | c "K_Const" | c "K_Pi" | c "K_Sigma" | 
    c "K_App" | c "K_Case" | c "K_KConv" |
    c "PK_Empty" |  c "PK_Cons" | 
    c "T_InCon" | c "T_Unit" | c "T_Var" |  c "T_Ctr" | 
    c "T_Fun" | c "T_App" |  c "T_Fix" | 
    c "T_Sigma" | c "T_Proj1" |  c "T_Proj2" | 
    c "T_Case" | c "T_TConv" | 
    c "P_Done" | c "AP_Cons" | c "C_Empty" | 
    c "C_Term" | c "C_Eq" ].

Scheme typeeq_ind := Induction for typeeq Sort Prop
  with pateq_ind := Induction for pateq Sort Prop
  with patteq_ind := Induction for patteq Sort Prop.

Combined Scheme eq_mutind from typeeq_ind, pateq_ind, patteq_ind.

Tactic Notation "eq_mutind_cases" tactic(first) tactic(c) :=
  first;
  [ c "TQ_InCon" | c "TQ_TRefl" | c "TQ_URefl" | c "TQ_Pi" |
    c "TQ_Sigma" | c "TQ_App" | c "TQ_Red1" |
    c "TQ_Red2" | c "TQ_Case" | c "TPQ_Empty" |
    c "TPQ_Cons" | c "TTPQ_Intro" ].

Scheme lc_type_ind := Induction for lc_type Sort Prop
  with lc_pt_ind := Induction for lc_pt Sort Prop  
  with lc_ptt_ind := Induction for lc_ptt Sort Prop.

Combined Scheme lc_tp_mutind from lc_type_ind, lc_pt_ind, lc_ptt_ind.

Tactic Notation "lc_tp_mutind_cases" tactic(first) tactic(c) :=
  first;
  [ c "lc_t_Con" | c "lc_t_Unit" | c "lc_t_Pi" | c "lc_t_Sigma" |
    c "lc_t_App" | c "lc_t_Case" | c "lc_pt_empty" |
    c "lc_pt_cons" | c "lc_ptt_pattern" ].

Scheme lc_exp_ind := Induction for lc_exp Sort Prop
  with lc_pe_ind := Induction for lc_pe Sort Prop.

Combined Scheme lc_epe_mutind from lc_exp_ind, lc_pe_ind.

Scheme wfe_ind1 := Induction for wfe Sort Prop
  with wfpe_ind1 := Induction for wfpe Sort Prop.

Combined Scheme wf_epe_mutind from wfe_ind1, wfpe_ind1.

Scheme wft_ind1 := Induction for wft Sort Prop
  with wfpt_ind1 := Induction for wfpt Sort Prop.

Combined Scheme wf_tpt_mutind from wft_ind1, wfpt_ind1.

Scheme exp_ind := Induction for exp Sort Prop
  with pe_ind := Induction for pe Sort Prop
  with peb_ind := Induction for peb Sort Prop.

Combined Scheme epe_mutind from exp_ind, peb_ind, pe_ind.

Scheme  wft_ind2 := Induction for wft Sort Prop
  with wfpt_ind2 := Induction for wfpt Sort Prop
  with wfe_ind2 := Induction for wfe Sort Prop
  with wfpe_ind2 := Induction for wfpe Sort Prop
  with wfc_ind2 := Induction for wfc Sort Prop.

Combined Scheme wftec_mutind from wft_ind2, 
  wfpt_ind2, wfe_ind2, wfpe_ind2, wfc_ind2.

Scheme nwreduce_ind := Induction for nwreduce Sort Prop
  with nwsteps_ind := Induction for nwsteps Sort Prop.

Combined Scheme nwreduce_mutind from nwreduce_ind, nwsteps_ind.

Tactic Notation "wfe_ind_cases" tactic(first) tactic(c) :=
  first;
  [ c "T_InCon" | c "T_Unit" | c "T_Var" |  c "T_Ctr" | 
    c "T_Fun" | c "T_App" |  c "T_Fix" | 
    c "T_Sigma" | c "T_Proj1" |  c "T_Proj2" | 
    c "T_Case" | c "T_TConv" ].

Lemma notin_substG: forall G w x y,
  x \notin (efv_exp w) ->
  x \notin (dom G) ->
  x \notin dom(esubst_G w y G).
Proof.
  intros.
  induction G; auto.
  destruct a.
  simpl_asnlist in *. 
  assert (J: x `notin` dom (esubst_G w y G)). auto. 
  simpl. solve_notin.

  assert (J: x `notin` dom (esubst_G w y G)). auto. 
  simpl. solve_notin.
Qed.


