Add LoadPath "./metatheory".

Set Implicit Arguments.

Require Import Metatheory.
Require Import lang.
Require Import langExtra.
Require Import isEq.
Require Import infrastructure. 
Require Import thy. 

Tactic Notation "wfe_ind_cases" tactic(first) tactic(c) :=
  first;
  [ c "T_InCon" | c "T_Unit" | c "T_Var" |  c "T_Ctr" | 
    c "T_Fun" | c "T_App" |  c "T_Fix" | 
    c "T_Sigma" | c "T_Proj1" |  c "T_Proj2" | 
    c "T_Case" | c "T_TConv" ].

Lemma open_rec_value_inv: forall x e k,
  is_v_of_exp (open_exp_wrt_exp_rec k (e_Var_f x) e) ->
  is_v_of_exp e.
Proof.
  intros x e k H.
  generalize dependent k.
  induction e; simpl in *; intros; eauto.
    destruct (lt_eq_lt_dec n k); auto.
      destruct s; auto.
    destruct H; eauto.
Qed.

Lemma open_value_inv: forall x e,
  is_v_of_exp (open_exp_wrt_exp e (e_Var_f x)) ->
  is_v_of_exp e.
Proof.
  intros x e H.
  unfold open_exp_wrt_exp in H.
  apply open_rec_value_inv in H; auto.
Qed.

Lemma open_rec_is_v_of_exp : forall e f k,
  is_v_of_exp e ->
  is_v_of_exp (open_exp_wrt_exp_rec k (e_Var_f f) e).
Proof.
  intros e f k H.
  generalize dependent k.
  generalize dependent f.
  induction e; intros; simpl in *; auto.
    inversion H.
    split; intuition auto.
Qed.  

Lemma open_is_v_of_exp : forall e f,
  is_v_of_exp e ->
  is_v_of_exp (open_exp_wrt_exp e (e_Var_f f)).
Proof.
  unfold open_exp_wrt_exp.
  intros e f H.
  apply open_rec_is_v_of_exp; auto.
Qed.

Lemma inversion_typeeq_app_pi: forall D t1 w2 s1 s2,
  ~ (incon D) ->
  typeeq D (t_App t1 w2) (t_Pi s1 s2) ->
  False .
Proof.
  intros.
  remember (t_App t1 w2) as t.
  remember (t_Pi s1 s2) as s.
  generalize dependent t1.
  generalize dependent w2.
  generalize dependent s1.
  generalize dependent s2.
  induction H0; intros; subst; try solve [inversion Heqs | inversion Heqt]; eauto.
Qed.

Lemma inversion_typeeq_app_sigma: forall D t1 w2 s1 s2,
  ~ (incon D) ->
  typeeq D (t_App t1 w2) (t_Sigma s1 s2) ->
  False .
Proof.
  intros.
  remember (t_App t1 w2) as t.
  remember (t_Sigma s1 s2) as s.
  generalize dependent t1.
  generalize dependent w2.
  generalize dependent s1.
  generalize dependent s2.
  induction H0; intros; subst; try solve [inversion Heqs | inversion Heqt]; eauto.
Qed.

(** ** Canonical forms *)
Lemma canonical_form_abs : forall v t s1 s2,
  is_v_of_exp v ->
  wfe nil v t ->
  typeeq nil t (t_Pi s1 s2) ->
  exists e1, v = e_Lam e1 \/ v = e_Fix e1.
Proof.
  intros v t s1 s2 Val Wfe Teq.
  remember (nil:ctx) as G.
  generalize dependent s1.
  generalize dependent s2.
  assert (incon_false := empty_incon).
  (wfe_ind_cases (induction Wfe) Case); 
    intros; subst; simpl; simpl in Val;
    try solve [inversion Teq; contradict incon_false; auto| 
               inversion Val|
               auto].
  Case "T_Ctr".
    pick fresh x.
    assert (x `notin` L) as xnotinL. auto.
    apply H0 in xnotinL. simpl in xnotinL.
    assert (typeeq ((e_Var_f x, e)::nil) 
                   (t_App (t_Con T) (open_exp_wrt_exp u (e_Var_f x)))  
                   (t_Pi s1 s2)) as J.
      apply trans_typeeq with (t2:=t); auto.
        rewrite_env (nil ++ one(e_Var_f x, e) ++ nil).
        apply weakening_typeeq; auto.
          simpl. split; auto.
    assert (pure e) as Pe.
      apply pure_is_v_of_exp; auto.
        apply lc_wfe in Wfe. decompose [and] Wfe; auto.
    apply subst_typeeq with (x:=x) (w:=e) in J; auto.
      simpl in J.
      destruct (@eq_dec expvar _ x x); subst; try solve [tauto].
    rewrite esubst_exp_fresh_eq in J; auto.
    rewrite_env (nil ++ (e, e) :: nil) in J.
    apply cut_typeeq in J; auto using ref_isEq.
    
    apply inversion_typeeq_app_pi in J; auto using empty_incon.
      inversion J.
  Case "T_Fun".
    exists e. auto.
  Case "T_Fix".
    exists e. auto.
  Case "T_TConv".
    apply IHWfe with (s1:=s1)(s2:=s2); auto.
      apply trans_typeeq with (t2:=t'); auto.
(*  Case "T_CConv".
    inversion H; subst.
    apply IHWfe with (s1:=s1)(s2:=s2); auto.
*)
Qed.

Lemma canonical_form_sie : forall v t s1 s2,
  is_v_of_exp v ->
  wfe nil v t ->
  typeeq nil t (t_Sigma s1 s2) ->
  exists v1, exists v2, 
    v = (e_Sie v1 v2) /\ is_v_of_exp v1 /\ is_v_of_exp v2. 
Proof.
  intros v t s1 s2 Val Wfe Teq.
  remember (nil:ctx) as G.
  generalize dependent s1.
  generalize dependent s2.
  assert (incon_false := empty_incon).
  (wfe_ind_cases (induction Wfe) Case); 
    intros; subst; simpl; simpl in Val;
    try solve [inversion Teq; contradict incon_false; auto| 
               inversion Val|
               auto].
  Case "T_Ctr".
    pick fresh x.
    assert (x `notin` L) as xnotinL. auto.
    apply H0 in xnotinL. simpl in xnotinL.
    assert (typeeq ((e_Var_f x, e)::nil) 
                   (t_App (t_Con T) (open_exp_wrt_exp u (e_Var_f x)))  
                   (t_Sigma s1 s2)) as J.
      apply trans_typeeq with (t2:=t); auto.
        rewrite_env (nil ++ one(e_Var_f x, e) ++ nil).
        apply weakening_typeeq; auto.
          simpl. split; auto. 
    assert (pure e) as Pe.
      apply pure_is_v_of_exp; auto.
        apply lc_wfe in Wfe. decompose [and] Wfe; auto.
    apply subst_typeeq with (x:=x) (w:=e) in J; auto.
      simpl in J.
      destruct (@eq_dec expvar _ x x); subst; try solve [tauto].
    rewrite esubst_exp_fresh_eq in J; auto.
    rewrite_env (nil ++ (e, e) :: nil) in J.
    apply cut_typeeq in J; auto using ref_isEq.
    
    apply inversion_typeeq_app_sigma in J; auto using empty_incon.
      inversion J.
  Case "T_Sigma".
    exists e1. exists e2. split; auto.
  Case "T_TConv".
    apply IHWfe with (s1:=s1)(s2:=s2); auto.
      apply trans_typeeq with (t2:=t'); auto.
(*  Case "T_CConv".
    inversion H; subst.
    apply IHWfe with (s1:=s1)(s2:=s2); auto.
*)
Qed.

Lemma canonical_form_con : forall v t T u,
  is_v_of_exp v ->
  wfe nil v t ->
  typeeq nil t (t_App (t_Con T) u) ->
  exists v1, exists Ci, exists ui, exists si,
    v = (e_Con Ci v1) /\ is_v_of_exp v1 /\
    binds Ci (bind_t (t_Pi si (t_App (t_Con T) ui))) Signature.
Proof.
  intros v t T w Val Wfe Teq.
  remember (nil:ctx) as G.
  generalize dependent w.
  generalize dependent T.
  assert (incon_false := empty_incon).
  (wfe_ind_cases (induction Wfe) Case); 
    intros; subst; simpl; simpl in Val;
    try solve [inversion Teq; contradict incon_false; auto| 
               inversion Val|
               auto].
  Case "T_Ctr".
    pick fresh x.
    assert (x `notin` L) as xnotinL. auto.
    apply H0 in xnotinL. simpl in xnotinL.
    assert (typeeq ((e_Var_f x, e)::nil) 
                   (t_App (t_Con T) (open_exp_wrt_exp u (e_Var_f x)))  
                   (t_App (t_Con T0) w)) as J.
      apply trans_typeeq with (t2:=t); auto.
        rewrite_env (nil ++ one(e_Var_f x, e) ++ nil).
        apply weakening_typeeq; auto.
          simpl. split; auto. 
    assert (pure e) as Pe.
      apply pure_is_v_of_exp; auto.
        apply lc_wfe in Wfe. decompose [and] Wfe; auto.
    apply subst_typeeq with (x:=x) (w:=e) in J; auto.
      simpl in J.
      destruct (@eq_dec expvar _ x x); subst; try solve [tauto].
    rewrite esubst_exp_fresh_eq in J; auto.
    rewrite_env (nil ++ (e, e) :: nil) in J.
    apply cut_typeeq in J; auto using ref_isEq.

    exists e. exists C. exists u. exists s.
    repeat(split; auto).
      apply inversion_typeeq_app1 with (t1:=t_Con T) (t2:=t_Con T0) 
            (w1:=esubst_exp e x (open_exp_wrt_exp u (e_Var_f x))) 
            (w2:=esubst_exp e x w) in J; auto.
      apply inversion_typeeq_T with (T1:=T)(T2:=T0)(t:=t_Con T)(s:=t_Con T0) in J; auto.
      subst; auto.
  Case "T_TConv".
    eapply IHWfe with (w:=w); eauto.
      apply trans_typeeq with (t2:=t'); auto.
(*  Case "T_CConv".
    inversion H; subst.
    eapply IHWfe; eauto.*)
Qed.

Lemma Progress : forall e t,
  wfe nil e t ->
  is_v_of_exp e \/ exists e', reduce e e'.
Proof.
  intros e t Hwfe.
  remember (nil:ctx) as G.
  (wfe_ind_cases (induction Hwfe) Case); subst; simpl; auto.
  Case "T_InCon".
    simpl in H0.
    contradict H0; auto using empty_incon.
  Case "T_Ctr".
    destruct IHHwfe as [J | [e' J]]; auto.
      right. exists (e_Con C e'). auto.
  Case "T_App".
    destruct IHHwfe1 as [J1 | [e1' J1]]; auto.
      destruct IHHwfe2 as [J2 | [e2' J2]]; auto.
        assert (typeeq nil (t_Pi t1 t2) (t_Pi t1 t2)) as EQ.
          apply ref_typeeq; simpl; auto.
        assert (J:=Hwfe1).
        apply canonical_form_abs with (s1:=t1)(s2:=t2) in J; auto.
          destruct J as [e12 [J | J]]; subst.
            right. exists (open_exp_wrt_exp e12 e2). apply app_beta; auto.
            right. exists (e_App (open_exp_wrt_exp e12 (e_Fix e12)) e2). 
              eapply app_fun_beta; eauto.
        right. exists (e_App e1 e2'). 
          apply app_cong2; auto.
      right. exists (e_App e1' e2). 
        apply app_cong1; auto.
  Case "T_Fix".
    pick fresh f.
    assert (f `notin` L) as fnotinL. auto.
    apply H1 in fnotinL.
    apply open_value_inv in fnotinL; auto.
  Case "T_Sigma".
    destruct IHHwfe1 as [J1 | [e1' J1]]; auto.
      destruct IHHwfe2 as [J2 | [e2' J2]]; auto.
        right. exists (e_Sie e1 e2'). 
          apply pair_cong2; auto.
      right. exists (e_Sie e1' e2). 
        apply pair_cong1; auto.
  Case "T_Proj1".
    destruct IHHwfe as [J | [e' J]]; auto.
      assert (typeeq nil (t_Sigma t1 t2) (t_Sigma t1 t2)) as EQ.
        apply ref_typeeq; simpl; auto.
      assert (JJ:=Hwfe).
      apply canonical_form_sie with (s1:=t1)(s2:=t2) in JJ; auto.
        destruct JJ as [v1 [v2 [JJ1 [JJ2 JJ3]]]]; subst.
        right. exists v1. apply proj_beta1; auto.
          apply lc_wfe in Hwfe. decompose [and] Hwfe. inversion H1; auto.
          apply lc_wfe in Hwfe. decompose [and] Hwfe. inversion H1; auto.
      right. exists (e_Proj1 e'). 
        apply proj_cong1; auto.
  Case "T_Proj2".
    destruct IHHwfe as [J | [e' J]]; auto.
      assert (typeeq nil (t_Sigma t1 t2) (t_Sigma t1 t2)) as EQ.
        apply ref_typeeq; simpl; auto.
      assert (JJ:=Hwfe).
      apply canonical_form_sie with (s1:=t1)(s2:=t2) in JJ; auto.
        destruct JJ as [v1 [v2 [JJ1 [JJ2 JJ3]]]]; subst.
        right. exists v2. apply proj_beta2; auto.
          apply lc_wfe in Hwfe. decompose [and] Hwfe. inversion H3; auto.
          apply lc_wfe in Hwfe. decompose [and] Hwfe. inversion H3; auto.
      right. exists (e_Proj2 e'). 
        apply proj_cong2; auto.
  Case "T_Case".
    destruct IHHwfe as [J | [e' J]]; auto.
      assert (typeeq nil (t_App (t_Con T) u) (t_App (t_Con T) u)) as EQ.
        apply ref_typeeq; simpl; auto.
      assert (JJ:=Hwfe).
      apply canonical_form_con with (T:=T)(u:=u) in JJ; auto.
        destruct JJ as [v1 [Ci [ui [si [JJ1 [JJ2 JJ3]]]]]]; subst.
        apply H0 in JJ3.
        destruct JJ3 as [ei JJ3].
        right. exists (open_exp_wrt_exp ei v1). 
        apply case_beta; auto.
          apply lc_wfe in Hwfe. decompose [and] Hwfe. inversion H4; auto.
      right. exists (e_Case e' pe5). 
        apply case_cong; auto.
(*  Case "T_CConv".
    inversion H; subst.
    destruct IHHwfe as [J | [e' J]]; auto.
      right. exists e'. auto.*)
Qed.
