Add LoadPath "./metatheory".
Require Import List.
Require Import Metatheory.
Require Import lang.
Require Import langExtra.
Require Import infrastructure.
Require Import isEq.
Require Import thy.
Require Import progress.

(**** Lemmas about "reduce" *****)

Lemma is_v_of_exp_esubst_exp : forall u x v,
 is_v_of_exp v ->
 is_v_of_exp (esubst_exp u x v).
Proof.
 induction v; intros Isv; simpl in *; try contradiction; auto.
   Case "e_Sie"; destruct Isv; auto.
Qed.

Hint Resolve is_v_of_exp_esubst_exp.
Hint Resolve esubst_exp_lc_exp esubst_pe_lc_pe esubst_peb_lc_peb.
Hint Constructors reduce.

Lemma esubst_exp_lc_exp_lambdabody : forall u x e,
  lc_exp u ->
  lc_exp (e_Lam e) ->
  lc_exp (e_Lam (esubst_exp u x e)).
Proof.
  intros; assert (lc_exp (esubst_exp u x (e_Lam e))); auto. 
Qed.

Lemma esubst_exp_lc_exp_fixbody : forall u x e,
  lc_exp u ->
  lc_exp (e_Fix e) ->
  lc_exp (e_Fix (esubst_exp u x e)).
Proof.
  intros; assert (lc_exp (esubst_exp u x (e_Fix e))); auto. 
Qed.

Lemma lc_esubst_exp : forall u x e,
  lc_exp u ->
  lc_exp e ->
  lc_exp (esubst_exp u x e).
Proof.
  auto with lngen.
Qed.

Lemma lc_esubst_peb : forall u x peb1,
  lc_exp u ->
  lc_peb peb1 ->
  lc_peb (esubst_peb u x peb1).
Proof.
  auto with lngen.
Qed.
Hint Resolve lc_esubst_exp lc_esubst_peb.

Lemma find_pe_subst : forall u x peb1 pe1,
  lc_exp u ->
  find_pe peb1 pe1 ->
  find_pe (esubst_peb u x peb1) (esubst_pe u x pe1).
Proof.
  induction 2.
    apply pe_here; auto.
    
    simpl; apply pe_there with (peb5 := (esubst_peb u x peb5)).
      auto.
      inversion H1; subst.
      apply notmatching_peb_intro with (Ci:=Ci) (ei:=(esubst_exp u x ei)) (Cj:=Cj) (ej:=(esubst_exp u x ej));
        change (peb_Branch Ci (esubst_exp u x ei))
          with (esubst_peb u x (peb_Branch Ci ei));
        change (peb_Branch Cj (esubst_exp u x ej))
          with (esubst_peb u x (peb_Branch Cj ej));
        auto.
Qed.

Lemma find_pe_subst' : forall u x Cj ej pe1,
  lc_exp u ->
  find_pe (peb_Branch Cj ej) pe1 ->
  find_pe (peb_Branch Cj (esubst_exp u x ej)) (esubst_pe u x pe1).
Proof.
  intros;
  assert (find_pe (esubst_peb u x (peb_Branch Cj ej)) (esubst_pe u x pe1));
  auto using find_pe_subst.
Qed.
   
   
Lemma reduce_subst : forall u x e e',
  lc_exp u -> 
  reduce e e' ->
  reduce (esubst_exp u x e) (esubst_exp u x e').
Proof.
  induction 2; try (simpl ; auto ; fail).
    Case "app_beta";  simpl; rewrite esubst_exp_open_exp_wrt_exp; auto using esubst_exp_lc_exp_lambdabody.
    Case "app_fun_beta".
      change (esubst_exp u x (e_App (e_Fix v1) v2))
        with (e_App (esubst_exp u x (e_Fix v1)) (esubst_exp u x v2));
      change (esubst_exp u x (e_App (open_exp_wrt_exp v1 (e_Fix v1)) v2))
        with (e_App (esubst_exp u x (open_exp_wrt_exp v1 (e_Fix v1))) (esubst_exp u x v2));
      rewrite esubst_exp_open_exp_wrt_exp.
        apply app_fun_beta; auto using esubst_exp_lc_exp_fixbody.
      auto.
   Case "case_beta".
      change (esubst_exp u x (e_Case (e_Con Cj v5) pe5))
        with (e_Case (e_Con Cj  (esubst_exp u x v5)) (esubst_pe u x pe5)).
      rewrite esubst_exp_open_exp_wrt_exp;
        auto using find_pe_subst'.
Qed.

Lemma mreduce_subst : forall u x e e',
  lc_exp u -> 
  mreduce e e' ->
  mreduce (esubst_exp u x e) (esubst_exp u x e').
Proof.
  induction 2.
    Case "refl"; constructor; auto.
    Case "step".
      assert (reduce (esubst_exp u x e) (esubst_exp u x e')) as Hsteps by (auto using reduce_subst).
      apply (step _ _ _ Hsteps); assumption.
Qed.
   
Lemma reduce_preserves_constructor : forall C e0 e',
  reduce (e_Con C e0) e' ->
  exists e0', e' = (e_Con C e0') /\ reduce e0 e0'.
Proof.
  inversion 1; subst; eauto.
Qed.

Lemma lc_exp_e_Con_body : forall C e,
  lc_exp (e_Con C e) -> lc_exp e.
Proof. inversion 1; assumption. Qed.
Hint Resolve lc_exp_e_Con_body.

Hint Constructors mreduce.

Lemma mreduce_preserves_constructor_helper : forall C e e',
  mreduce e e' -> forall e0, e = (e_Con C e0) -> exists e0', e' = (e_Con C e0') /\ mreduce e0 e0'.
Proof.
  induction 1; intros; subst.
    Case "refl"; exists e0; split; eauto. 
    Case "step".
      destruct (reduce_preserves_constructor _ _ _ H) as [e0' [Heq' HR]].
      destruct (IHmreduce e0' Heq') as [e0'' [Heq'' HR']].
      eauto.
Qed.

Lemma mreduce_preserves_constructor : forall C e0 e',
  mreduce (e_Con C e0) e' ->
  exists e0', e' = (e_Con C e0') /\ mreduce e0 e0'.
Proof.
  intros C e0 e' H.
  apply mreduce_preserves_constructor_helper with (e := e_Con C e0) (e0 := e0); auto.
Qed.

Lemma value_not_reduce : forall v e,
  is_v_of_exp v ->
  not (reduce v e).
Proof.
  unfold not.
  induction 2; simpl in *; try contradiction;
    match goal with [ H : is_v_of_exp ?e1 /\  is_v_of_exp ?e2 |- _ ] => destruct H end; auto.
Qed.

Ltac cant_reduce_value := 
  match goal with [ Hv : is_v_of_exp ?v, H : reduce ?v ?e |- _ ] 
                   => contradict H; apply value_not_reduce; assumption
             end.

Ltac cant_reduce :=
  match goal with [ H : reduce ?e ?e' |- _ ] => inversion H; cant_reduce_value end.

Lemma notmatching_peb_comm : forall peb1 peb2,
  notmatching_peb peb1 peb2 ->
  notmatching_peb peb2 peb1.
Proof.
   inversion 1; constructor; auto.
Qed.
Hint Resolve notmatching_peb_comm.

Lemma find_pe_deterministic_helper : forall peb1 peb2 pes,
  find_pe peb1 pes ->
  find_pe peb2 pes ->
  (peb1 = peb2) \/  (notmatching_peb peb1 peb2).
Proof.
  induction 1; inversion 1; subst; auto.
Qed.

Lemma find_pe_deterministic : forall C ei ej pe1,
  find_pe (peb_Branch C ei) pe1 ->
  find_pe (peb_Branch C ej) pe1 ->
  ei = ej.
Proof. 
  intros C ei ej pe1 F1 F2.
  destruct (find_pe_deterministic_helper _ _ _ F1 F2).
    injection H; tauto.
    inversion H; absurd (C = C); auto.
Qed.
 
Lemma reduce_deterministic : forall e e',
  reduce e e' ->
  forall e'',
  reduce e e'' ->
  e' = e''.
Proof.
  induction 1; inversion 1; subst; 
    tauto || cant_reduce || (f_equal; auto).
      eauto using find_pe_deterministic.
Qed.

Hint Constructors mreduce.

Hint Constructors lc_pe lc_exp.
Lemma lc_find_pe : forall peb1 pe1,
  find_pe peb1 pe1 -> 
  lc_pe pe1.
Proof.
  induction 1;
    try match goal with [H: notmatching_peb ?peb1 ?peb2 |- _] => inversion H end;
    auto.
Qed.

Lemma lc_body_find_pe : forall Cj ej pe1,
  find_pe (peb_Branch Cj ej) pe1 ->
  body_exp_wrt_exp ej.
Proof.
  intros.
  remember (peb_Branch Cj ej) as peb1.
  revert Cj ej Heqpeb1.
  induction H.
   intros; subst; eauto with lngen.
   auto.
Qed.

Lemma lc_reduce : forall e e',
  reduce e e'
  -> lc_exp e /\ lc_exp e'.
Proof.
 induction 1; split; auto with lngen;
   repeat match goal with [ H: ?H1 /\ ?H2 |- _] => destruct H end;
   auto with lngen.
     eauto using lc_find_pe.
     apply lc_body_exp_wrt_exp; eauto using lc_body_find_pe.
Qed.

Lemma mreduce_lc_exp : forall e e'',
  mreduce e e'' ->  lc_exp e''.
Proof.
  induction 1; auto.
Qed.
Hint Resolve mreduce_lc_exp.

Lemma mreduce_diamond : forall e e' e'',
  mreduce e e' ->
  mreduce e e'' ->
  exists u,
    mreduce e' u /\ mreduce e'' u.
Proof.
 induction 1.
   Case "refl"; eauto.
   Case "step".
    inversion 1; subst.
      exists e''0; eauto.

      match goal with [ H1 : reduce ?e ?e',
                        H2 : reduce ?e ?e'' |- _ ] => rewrite (reduce_deterministic _ _ H1 _ H2) in * end; 
      eauto.
Qed.

Lemma mreduce_trans : forall e1 e2 e3,
  mreduce e1 e2 ->
  mreduce e2 e3 ->
  mreduce e1 e3.
Proof.
  induction 1; [ tauto | eauto].
Qed. 

Lemma lc_pure : forall w,
  pure w ->
  lc_exp w.
Proof.
  induction 1; auto.
Qed.
Lemma lc_esubst_ectx : forall u x D,
  lc_exp u ->
  lc_ectx D ->
  lc_ectx (esubst_ectx u x D).
Proof.
  induction D.
    simpl. auto.

    destruct a. simpl in *. intros H [H1 [H2 H3]]. auto.
Qed.
Hint Resolve lc_esubst_exp lc_esubst_ectx lc_pure.

(**** Lemmas about "wreduce" *****)
(** Stuff about wreduce **)

(* Sanity checking that we are defining the expected relation: *)

Hint Constructors nwreduce.
Lemma wreduce_implies_nwreduce : forall e e',
  wreduce e e' -> nwreduce e e'.
Proof.
  induction 1; auto.
Qed.

Lemma value_not_nwreduce : forall v e,
  is_v_of_exp v ->
  not (nwreduce v e).
Proof.
  unfold not.
  induction 2; simpl in *; try contradiction;
    match goal with [ H : is_v_of_exp ?e1 /\  is_v_of_exp ?e2 |- _ ] => destruct H end; auto.
Qed.

Lemma value_not_nwsteps : forall v,
  is_v_of_exp v ->
  not (nwsteps v).
Proof.
  unfold not; intros v Hv Hs. inversion Hs.
  contradict H. eauto using value_not_nwreduce.
Qed.

Lemma value_not_wreduce : forall v e,
  is_v_of_exp v ->
  not (wreduce v e).
Proof.
  unfold not.
  induction 2; simpl in *; try contradiction;
    match goal with [ H : is_v_of_exp ?e1 /\  is_v_of_exp ?e2 |- _ ] => destruct H end; auto.
Qed.


Ltac cant_wreduce_value := 
  match goal with [ Hv : is_v_of_exp ?v, H : wreduce ?v ?e |- _ ] 
                   => contradict H; apply value_not_wreduce; assumption
                | [ H1: ~(nwsteps ?e),
                    H2: wreduce ?e ?e' |- _] 
                  =>  contradict H1 ; (apply (nwsteps_intro  e e')) ; apply wreduce_implies_nwreduce; exact H2
             end.

Ltac cant_wreduce :=
  match goal with [ H : wreduce ?e ?e' |- _ ]  => inversion H; cant_wreduce_value end.

Lemma wreduce_deterministic : forall e e',
  wreduce e e' ->
  forall e'',
  wreduce e e'' ->
  e' = e''.
Proof.
  induction 1; inversion 1; subst;
    (tauto || cant_wreduce || f_equal; auto).
      eauto using find_pe_deterministic.
Qed.

Hint Constructors pure.
Lemma v_pure : forall v,
  lc_exp v ->
  is_v_of_exp v ->
  pure v.
Proof.
  induction v; intros; simpl in *; auto;  try contradiction.
     eauto.
     decompose record H0; inversion H; auto.
Qed.

Hint Resolve v_pure value_not_nwsteps.
Hint Constructors wreduce.

Lemma lc_wreduce : forall e e',
  wreduce e e'
  -> lc_exp e /\ lc_exp e'.
Proof.
 induction 1; split; auto with lngen;
   repeat match goal with [ H: ?H1 /\ ?H2 |- _] => destruct H end;
   auto with lngen.
     apply lc_e_Case; eauto using lc_find_pe.  
     apply lc_body_exp_wrt_exp; eauto using lc_body_find_pe.
Qed.

Lemma mwreduce_lc_exp : forall e e'',
  mwreduce e e'' ->  lc_exp e''.
Proof.
  induction 1; auto.
Qed.

Hint Resolve lc_wreduce mwreduce_lc_exp.
Lemma mwreduce_diamond : forall e e' e'',
  mwreduce e e' ->
  mwreduce e e'' ->
  exists u,
    mwreduce e' u /\ mwreduce e'' u.
Proof.
 induction 1.
   Case "refl"; eauto.
   Case "step".
    inversion 1; subst.
      exists e''0; eauto.

      match goal with [ H1 : wreduce ?e ?e',
                        H2 : wreduce ?e ?e'' |- _ ] => rewrite (wreduce_deterministic _ _ H1 _ H2) in * end; 
      eauto.
Qed.

Lemma pure_subst : forall w x w1,
  pure w ->
  pure w1 ->
  pure (esubst_exp w x w1).
Proof.
  induction 2; simpl; auto.
    destruct (x0 == x); auto.
    auto using esubst_exp_lc_exp_lambdabody.
    auto using esubst_exp_lc_exp_fixbody.
Qed.


(**** Evaluation contexts and plugging ****)

(* The intention is to capture this grammar (which Ott doesn't like):

evctx  :: 'EVCTX_' ::=     {{com evaluation contexts}}  {{tex E }}
 | hole	      	        ::    :: CtxHole {{ tex \Box}}
 | evctx e	        ::    :: Ctxapp1
 | v evctx	        ::    :: Ctxapp2
 | < evctx , e >	::    :: Ctxpair1
 | < v , evctx >	::    :: Ctxpair2
 | evctx . 1   	        ::    :: Ctxproj1
 | evctx . 2   	        ::    :: Ctxproj2
 | case evctx of { pe }	::    :: Ctxcase1
*)
    
Inductive evctx : Set :=
| Ctxhole  : evctx
| Ctxapp1  : evctx -> exp -> evctx
| Ctxapp2  : exp -> evctx ->  evctx
| Ctxpair1 : evctx -> exp -> evctx
| Ctxpair2 : exp -> evctx -> evctx
| Ctxproj1 : evctx -> evctx
| Ctxproj2 : evctx -> evctx
| Ctxcase1 : evctx -> pe -> evctx
| CtxCon   : contmvar -> evctx -> evctx.


Inductive lc_evctx : evctx -> Prop :=
| lc_Ctxhole  : lc_evctx Ctxhole
| lc_Ctxapp1  : forall E e2, lc_evctx E ->lc_exp e2 -> lc_evctx (Ctxapp1 E e2)
| lc_Ctxapp2  : forall v E, lc_exp v -> lc_evctx E -> lc_evctx (Ctxapp2 v E)
| lc_Ctxpair1 : forall E e2, lc_evctx E -> lc_exp e2 -> lc_evctx (Ctxpair1 E e2)
| lc_Ctxpair2 : forall v E,  lc_evctx E -> lc_exp v -> lc_evctx (Ctxpair2 v E)
| lc_Ctxproj1 : forall E, lc_evctx E -> lc_evctx (Ctxproj1 E)
| lc_Ctxproj2 : forall E, lc_evctx E -> lc_evctx (Ctxproj2 E)
| lc_Ctxcase1 : forall E pe1, lc_evctx E -> lc_pe pe1 -> lc_evctx (Ctxcase1 E pe1)
| lc_CtxCon   : forall C E, lc_evctx E -> lc_evctx (CtxCon C E).

Inductive is_cbv_evctx: evctx -> Prop := 
| is_cbv_Ctxhole  : is_cbv_evctx Ctxhole
| is_cbv_Ctxapp1  : forall E e2, is_cbv_evctx E -> is_cbv_evctx (Ctxapp1 E e2)
| is_cbv_Ctxapp2  : forall v E, is_v_of_exp v -> is_cbv_evctx E -> is_cbv_evctx (Ctxapp2 v E)
| is_cbv_Ctxpair1 : forall E e2, is_cbv_evctx E  -> is_cbv_evctx (Ctxpair1 E e2)
| is_cbv_Ctxpair2 : forall v E,  is_cbv_evctx E -> is_v_of_exp v -> is_cbv_evctx (Ctxpair2 v E)
| is_cbv_Ctxproj1 : forall E, is_cbv_evctx E -> is_cbv_evctx (Ctxproj1 E)
| is_cbv_Ctxproj2 : forall E, is_cbv_evctx E -> is_cbv_evctx (Ctxproj2 E)
| is_cbv_Ctxcase1 : forall E pe1, is_cbv_evctx E  -> is_cbv_evctx (Ctxcase1 E pe1)
| is_cbv_CtxCon   : forall C E, is_cbv_evctx E -> is_cbv_evctx (CtxCon C E).

Definition wf_evctx (E:evctx) : Prop := lc_evctx E /\ is_cbv_evctx E.

Fixpoint plug (E : evctx) (u : exp) {struct E} : exp :=
  match E  with 
  | Ctxhole  => u
  | Ctxapp1  E' e   => e_App (plug E' u) e
  | Ctxapp2  v E'  => e_App v (plug E' u)
  | Ctxpair1 E' e   => e_Sie (plug E' u) e
  | Ctxpair2 v E'  => e_Sie v (plug E' u)
  | Ctxproj1 E'       => e_Proj1 (plug E' u)
  | Ctxproj2 E'       => e_Proj2 (plug E' u)
  | Ctxcase1 E' pe1 => e_Case (plug E' u) pe1
  | CtxCon C E' => e_Con C (plug E' u)
  end.

Lemma reduce_plug : forall E e e',
  lc_evctx E->
  is_cbv_evctx E ->
  reduce e e' ->
  reduce (plug E e) (plug E e').
Proof.
  induction E; intros e1 e1' HlcE HvE Hred; 
    inversion HlcE; inversion HvE; subst;
      simpl; auto. 
Qed.

Lemma wreduce_plug : forall E e e',
  lc_evctx E ->
  is_cbv_evctx E ->
  wreduce e e' ->
  wreduce (plug E e) (plug E e').
Proof.
  induction E; intros e1 e1' HlcE HvE Hred; 
    inversion HlcE; inversion HvE; subst;
      simpl; auto. 
Qed.

(**** Lemmas about "reduce <-> mreduce for closed wellformed terms" *****)
 
Hint Constructors reduce.

Lemma wreduce_beta  : forall e e',
  reduce e e' ->
  wreduce e e'.
Proof.
  induction 1; auto.
Qed. 

Lemma nwreduce_beta  : forall e e',
  reduce e e' ->
  nwreduce e e'.
Proof.
  induction 1; auto.
Qed. 

Lemma closed_nstepped_wfw_is_v : forall w t, 
  wfe nil w t ->
  pure w -> 
  (not ( nwsteps w ))  ->
  is_v_of_exp w.
Proof.
  intros w t Wfe Pw Stuck.
  remember (@nil (asn type (exp * exp)) ) as G.
  (wfe_ind_cases (induction Wfe) Case);
    intros; subst; simpl in *; 
    try solve [inversion Pw; subst; auto | auto].
  Case "T_InCon".
    simpl in H0.
    contradict H0; auto using empty_incon.
  Case "T_Ctr".
    inversion Pw; subst.
    apply IHWfe; auto.
      intro J. apply Stuck. inversion J; subst.
        apply nwsteps_intro with (e':=e_Con C e'); auto.
  Case "T_Sigma".
    inversion Pw; subst.
    split.
      apply IHWfe1; auto.
        intro J. apply Stuck. inversion J; subst.
          apply nwsteps_intro with (e':=e_Sie e' e2); auto.
      apply IHWfe2; auto.
        intro J. apply Stuck. inversion J; subst.
          apply nwsteps_intro with (e':=e_Sie e1 e'); auto.
  Case "T_Proj1".
    assert (J:=Wfe).
    apply Progress in J.
    destruct J as [J | [e' J]].
      apply canonical_form_sie with (s1:=t1)(s2:=t2) in Wfe; auto.
        destruct Wfe as [v1 [v2 [EQ [V1 V2]]]]; subst.
        apply Stuck.
          apply nwsteps_intro with (e':=v1); auto.
            apply nwreduce_proj_beta1; auto.
              apply pure_is_v_of_exp; auto.
                inversion Pw; subst.
                inversion H0; subst; auto.
          
                inversion Pw; subst.
                inversion H0; subst; auto.
        apply ref_typeeq; simpl; auto.   

      apply Stuck. 
        apply nwsteps_intro with (e':=e_Proj1 e'); auto.
          apply nwreduce_proj_cong1.
            apply nwreduce_beta; auto.
  Case "T_Proj2".
    assert (J:=Wfe).
    apply Progress in J.
    destruct J as [J | [e' J]].
      apply canonical_form_sie with (s1:=t1)(s2:=t2) in Wfe; auto.
        destruct Wfe as [v1 [v2 [EQ [V1 V2]]]]; subst.
        apply Stuck.
          apply nwsteps_intro with (e':=v2); auto.
            apply nwreduce_proj_beta2; auto.
              apply pure_is_v_of_exp; auto.
                inversion Pw; subst.
                inversion H2; subst; auto.
          
                inversion Pw; subst.
                inversion H2; subst; auto.
        apply ref_typeeq; simpl; auto.   

      apply Stuck. 
        apply nwsteps_intro with (e':=e_Proj2 e'); auto.
          apply nwreduce_proj_cong2.
            apply nwreduce_beta; auto.
Qed.

Lemma closed_wf_wreduce_implies_reduce  : forall e t e',
  wfe nil e t ->
  wreduce e e' ->
  reduce e e'.
Proof.
  intros e t e' Wfe WReduce.
  remember (@nil (asn type (exp * exp)) ) as G.
  generalize dependent e'.
  (wfe_ind_cases (induction Wfe) Case);
    intros; subst; simpl in *; 
    try solve [inversion WReduce; subst; auto | auto].
  Case "T_InCon".
    simpl in H0.
    contradict H0; auto using empty_incon.
  Case "T_App".
    inversion WReduce; subst; auto.
    SCase "(\x.e0)e2 -> e0^e2" .
      apply closed_nstepped_wfw_is_v in Wfe2; auto.
    SCase "(fun.v1)e2 -> v1^(fun.v1)^e2" .
      apply closed_nstepped_wfw_is_v in Wfe2; auto.
    SCase "e1e2 -> e1e2'" .
      assert (J:=Wfe1).
      apply Progress in J.
      destruct J as [J | [e1' J]]; auto.
        contradict H4; auto.
          apply nwsteps_intro with (e':=e1'); auto.
            apply nwreduce_beta; auto.
  Case "T_Sigma".
    inversion WReduce; subst; auto.
    SCase "(e1,e2) -> (e1,e2')" .
      assert (J:=Wfe1).
      apply Progress in J.
      destruct J as [J | [e1' J]]; auto.
        contradict H4; auto.
          apply nwsteps_intro with (e':=e1'); auto.
            apply nwreduce_beta; auto.
  Case "T_Proj1".
    inversion WReduce; subst; auto.
    SCase "(e',w2).1 -> e'" .
      assert (J:=Wfe).
      apply Progress in J.
      destruct J as [J | [e'0 J]]; auto.
        simpl in J. destruct J as [J1 J2].
        apply proj_beta1; auto.

        inversion J; subst.
        SSCase "(e',w2)->(e1',w2)".
          contradict H1; auto.
          apply nwsteps_intro with (e':=e1'); auto.
            apply nwreduce_beta; auto. 
        SSCase "(e',w2)->(e',e2')".
          contradict H3; auto.
          apply nwsteps_intro with (e':=e2'); auto.
            apply nwreduce_beta; auto. 
  Case "T_Proj2".
    inversion WReduce; subst; auto.
    SCase "(e',w2).2 -> w2" .
      assert (J:=Wfe).
      apply Progress in J.
      destruct J as [J | [e'0 J]]; auto.
        simpl in J. destruct J as [J1 J2].
        apply proj_beta2; auto.

        inversion J; subst.
        SSCase "(w1,e')->(e1',e')".
          contradict H3; auto.
          apply nwsteps_intro with (e':=e1'); auto.
            apply nwreduce_beta; auto. 
        SSCase "(w1,e')->(w1,e2')".
          contradict H5; auto.
          apply nwsteps_intro with (e':=e2'); auto.
            apply nwreduce_beta; auto. 
  Case "T_Case".
    inversion WReduce; subst; auto.
    apply closed_nstepped_wfw_is_v in Wfe; auto.
      intro J. inversion J; subst. inversion H2; subst.
      apply H7. apply nwsteps_intro with (e':=e1'); auto.
Qed. 


