Add LoadPath "./metatheory".

Set Implicit Arguments.

Require Import Metatheory.
Require Import lang.
Require Import langExtra.
Require Import isEq.
Require Import infrastructure.
Require Import thy.
Require Import reductions.


Lemma subst_lc_ctx : forall G w x,
    lc_ctx G ->
    lc_exp w ->
    lc_ctx (esubst_G w x G).
Proof.
  intros. induction G; eauto.
  destruct a. 
  simpl in *; intuition eauto using esubst_type_lc_type.
  destruct p. 
  simpl in *; intuition eauto using esubst_exp_lc_exp.
 
Qed. 


(************** Substitution for Well-formed Judgments **************)


Inductive var_wfG: ctx -> Prop :=
 | var_wfG_Empty: var_wfG nil
 | var_wfG_Term: forall x t G,
   var_wfG G ->
   efv_type t [<=] dom G ->
   var_wfG  (AssumeList.VarAsn (exp * exp)  x   t  ::  G  ) 
| var_wfG_Eq: forall e1 e2 G,
  var_wfG G ->
  efv_exp e1 [<=] dom G ->
  efv_exp e2 [<=] dom G ->
  var_wfG (AssumeList.AltAsn type (  e1 ,  e2  ) ::  G  ).
Hint Constructors var_wfG. 

Lemma notin_exp_wfe_wfpe_mutual: 
(forall G e t,
  wfe G e t ->
 forall x, x\notin (dom G) ->
   x\notin (efv_exp e))
/\ (forall G e u pe t,
    wfpe G e u pe t ->
    forall x, x\notin (dom G) ->
      x \notin (efv_pe pe)).
Proof.
 eapply wf_epe_mutind; intros; eauto.
  Case "var".
  simpl. assert (J: x \in dom G). eapply binds_In; eauto. 
  solve_uniq. 

 Case "lam".
  pick fresh z.
  assert (J: x \notin efv_exp (open_exp_wrt_exp e (e_Var_f z))).  auto.
  simpl. 
  assert (J1: efv_exp e[<=] efv_exp (open_exp_wrt_exp e (e_Var_f z))).
  eapply efv_exp_open_exp_wrt_exp_rec_lower; eauto.
  solve_notin.
 Case "fix".
  pick fresh z.
  assert (J: x \notin efv_exp (open_exp_wrt_exp e (e_Var_f z))).  auto.
  simpl. 
  assert (J1: efv_exp e[<=] efv_exp (open_exp_wrt_exp e (e_Var_f z))).
  eapply efv_exp_open_exp_wrt_exp_rec_lower; eauto.
  solve_notin.
(*  Case "cconv". 
  assert (J:= ctxeq_dom_eq c). eapply H. solve_uniq.   
*)
 Case "pe_cons".
  simpl in *. 
  assert (J1:= H0 x H1). 
  pick fresh z. assert (J2:  x \notin efv_exp (open_exp_wrt_exp ei (e_Var_f z))).
  eauto. assert (J3: efv_exp ei [<=] efv_exp (open_exp_wrt_exp ei (e_Var_f z))). 
  eapply efv_exp_open_exp_wrt_exp_rec_lower; eauto.
  solve_notin.
Qed.

Lemma notin_exp_wfe: forall G e t x,
  wfe G e t ->
  x\notin (dom G) ->
  x\notin (efv_exp e).
Proof.
  intros.
  eapply (proj1 notin_exp_wfe_wfpe_mutual); eauto. 
Qed.

Definition fresh_wfk_prop G k (WF : wfk G k) :=
  (forall x,  x\notin (dom G) -> 
    x\notin (efv_kind k)) /\ var_wfG G.
Definition fresh_wft_prop G t k (WF : wft G t k) :=
  (forall x,  x\notin (dom G) -> 
    x\notin (efv_type t)) /\ var_wfG G.
Definition fresh_wfe_prop G e t (WF : wfe G e t) :=
  (forall x, x\notin (dom G) -> x\notin (efv_type t))
  /\ var_wfG G.
Definition fresh_wfpt_prop G e u pt k (WF : wfpt G e u pt k) :=
  (forall x, x\notin (dom G) -> x\notin (efv_pt pt))
  /\ var_wfG G.
Definition fresh_wfpe_prop G e u pe t (WF : wfpe G e u pe t) :=
 (forall x, x\notin (dom G) -> x\notin (efv_type t))
 /\ var_wfG G.
Definition fresh_wfc_prop G (WF : wfc G) :=
  (forall x x0 t, 
    x\notin (dom G) ->
    binds x0 t G ->
    x\notin (efv_type t))
  /\ var_wfG G.


Lemma fresh_wf:
  (forall G k WF, @fresh_wfk_prop G k WF) /\
  (forall G t k WF, @fresh_wft_prop G t k WF) /\
  (forall G e u pt k WF, @fresh_wfpt_prop G e u pt k WF) /\
  (forall G e t WF, @fresh_wfe_prop G e t WF) /\
  (forall G e u pe t WF, @fresh_wfpe_prop G e u pe t WF) /\
  (forall G WF, @fresh_wfc_prop G WF).
Proof.
  eapply wf_mutind with  
    (P := fresh_wfk_prop)
  (P0 := fresh_wft_prop) 
  (P1 := fresh_wfpt_prop)
  (P2 := fresh_wfe_prop)  
  (P3 := fresh_wfpe_prop);
  unfold fresh_wfk_prop, fresh_wft_prop, fresh_wfpt_prop, 
  fresh_wfe_prop, fresh_wfpe_prop, fresh_wfc_prop; intros;  
  eauto; try solve [destruct H; split; intros; auto]. 

 Case "K_Pi".
 split; intros; pick fresh z; destruct (H z); auto.
  inversion H2; subst.
  assert (J: x \notin efv_kind (open_kind_wrt_exp k (e_Var_f z))). auto.
  assert (J1: efv_kind k[<=] efv_kind (open_kind_wrt_exp k (e_Var_f z))).
  eapply efv_kind_open_kind_wrt_exp_rec_lower; eauto.
  solve_notin.

  inversion H1; auto.

 Case "Pi".
  split; intros; pick fresh z; destruct (H z); auto.
  inversion H2; subst.
  assert (J: x \notin efv_type (open_type_wrt_exp t2 (e_Var_f z))). auto.
  assert (J1: efv_type t2[<=] efv_type (open_type_wrt_exp t2 (e_Var_f z))).
  eapply efv_type_open_type_wrt_exp_rec_lower; eauto.
  solve_notin.

  inversion H1; auto.

 Case "Sigma". 
  split; intros; pick fresh z; destruct (H z); auto.
  inversion H2; subst.
  assert (J: x \notin efv_type (open_type_wrt_exp t2 (e_Var_f z))). auto.
  assert (J1: efv_type t2[<=] efv_type (open_type_wrt_exp t2 (e_Var_f z))).
  eapply efv_type_open_type_wrt_exp_rec_lower; eauto.
  solve_notin.

  inversion H1; auto.

 Case "App". 
  split; intros; destruct H; auto.
  assert (J: x\notin efv_exp e).
  eapply notin_exp_wfe; eauto.  solve_notin.

 Case "Case ". 
  split; intros; destruct H; auto.
  assert (J: x\notin efv_exp e).
  eapply notin_exp_wfe; eauto. destruct H0.
  assert (J1:= H x H1). simpl in J1. solve_notin.
(* Case "cconv". 
  split; intros; destruct H0; auto.
  assert (J:= ctxeq_dom_eq c).  destruct H. eapply H. solve_uniq.   
*)
 Case "pe_cons".
  split; intros; destruct H; auto.
  simpl in *. 
  assert (J1:= H x H1). 
  pick fresh z. assert (J2:  x \notin efv_type (open_type_wrt_exp t (e_Var_f z))).
  destruct (H0 z); auto.
  assert  (J3: efv_type t [<=] efv_type (open_type_wrt_exp t (e_Var_f z))). 
  eapply efv_type_open_type_wrt_exp_rec_lower; eauto.
  solve_notin.
 
 Case "Var".
  split; intros; destruct H; auto. eapply H; eauto.
 Case "lam".
  split; intros; pick fresh z; destruct (H z); auto.
  inversion H2; subst.
  assert (J: x \notin efv_type (open_type_wrt_exp t2 (e_Var_f z))).  auto.
  assert (J1: efv_type t2[<=] efv_type (open_type_wrt_exp t2 (e_Var_f z))).
  eapply efv_type_open_type_wrt_exp_rec_lower; eauto.
  solve_notin.
  inversion H1; auto.

 Case "fix".
  split; intros; pick fresh z; destruct (H z); auto.
  inversion H1; auto.

 Case "pair".
   split; intros; destruct H; auto.
  pick fresh z.   
  assert (J: x \notin efv_type (open_type_wrt_exp t2 (e_Var_f z))).  
  destruct (H1 z); auto.
  assert (J1: efv_type t2[<=] efv_type (open_type_wrt_exp t2 (e_Var_f z))).
  eapply efv_type_open_type_wrt_exp_rec_lower; eauto.
  solve_notin.
 Case "proj1".
   split; intros; destruct H; auto.
  simpl in *.  assert (J:= H x H0). solve_notin.

(* Case "CConv".
  split; intros; destruct H0; auto.
  assert (J:= ctxeq_dom_eq c). destruct H.   eapply H; eauto.
  solve_uniq. 
*)
 Case "C_Term".
  split; intros;  destruct H; auto.
  assert (J:(x1 = x /\ t0 = t) \/ binds x1 t0 G).
  eapply binds_cons_1; eauto.
  destruct J as [(J1, J2)| J].
  subst. destruct H0. eapply H0; eauto.
  eapply H; eauto.
  destruct H0.
  eapply var_wfG_Term; eauto.  
  assert (J: forall x, x \in efv_type t -> x \in dom G). 
  intros.  assert (J:= H0 x0).  solve_uniq.
  solve_uniq.
  
 Case "C_Eq".
  split; intros; destruct H; eauto.
  inversion H3. inversion H5. eapply H; eauto.
 
  eapply var_wfG_Eq; eauto. 
  assert (J:= @notin_exp_wfe G e1 t).  
  assert (J1: forall x, x \in efv_exp e1 -> x \in dom G). 
  intros.  assert (J2:= J x).  solve_uniq.
  solve_uniq.

  assert (J:= @notin_exp_wfe G e2 t).  
  assert (J1: forall x, x \in efv_exp e2 -> x \in dom G). 
  intros.  assert (J2:= J x).  solve_uniq.
  solve_uniq.
Qed.

Lemma notin_type_wft: 
forall G t k x,
  wft G t k->
  x\notin (dom G) ->
  x\notin efv_type t.
Proof.
 intros. 
 destruct fresh_wf as [JK [JT [JPT [JE [JPE JG]]]]].
 eapply JT; eauto.
Qed.


Lemma notin_type_wfe: 
forall G e t x,
  wfe G e t->
  x\notin (dom G) ->
  x\notin efv_type t.
Proof.
 intros. 
 destruct fresh_wf as [JK [JT [JPT [JE [JPE JG]]]]].
 eapply JE; eauto.
Qed.


Lemma notin_kind_wfk: forall G k x,
  wfk G k  ->
  x\notin (dom G) ->
  x\notin (efv_kind k).
Proof.
 intros. 
 destruct fresh_wf as [JK [JT [JPT [JE [JPE JG]]]]].
 eapply JK; eauto.
Qed.

Lemma notin_type_binds: forall (G:ctx) t x y,
  wfc G ->
  binds y t G ->
  x\notin (dom G) ->
  x\notin (efv_type t).
Proof.
 intros.
 destruct fresh_wf as [JK [JT [JPT [JE [JPE JG]]]]].
 unfold fresh_wfc_prop in JG.
 eapply JG; eauto.
Qed.
 
Lemma notin_G_var_wfG: forall G x,
  var_wfG G  ->
  x\notin (dom G) ->
  x\notin (efv_ctx G).
Proof.
 intros. generalize dependent x.
 induction H; intros; eauto; simpl in *; solve_notin.
Qed.

Lemma notin_G_wfc: forall G x,
  wfc G  ->
  x\notin (dom G) ->
  x\notin (efv_ctx G).
Proof.
 intros. 
 destruct fresh_wf as [JK [JT [JPT [JE [JPE JG]]]]].
 eapply notin_G_var_wfG; eauto.
 eapply JK; eauto.
Qed.

Lemma notin_kind_wft_mutual: 
 (forall G t k,
  wft G t k ->
  forall x,  x\notin (dom G) -> x\notin (efv_kind k))
/\ (forall G e u pt k,
  wfpt G e u pt k ->
  forall x, x\notin (dom G) -> x\notin (efv_kind k)).
Proof.
  eapply wf_tpt_mutind; eauto using notin_kind_wfk; intros.

 assert (Jt: wfk nil k).
 eapply wf_tctr_kind with (S:=Signature); eauto using wfs_sig. 
 assert (J:= @notin_kind_wfk nil k x Jt). eauto.
(*
  assert (J:= ctxeq_dom_eq c).  
  assert (J1: x \notin dom G). solve_uniq. 
  solve_notin.
*)
Qed.

Lemma notin_kind_wft: forall G t k x,
  wft G t k  ->
  x\notin (dom G) ->
  x\notin (efv_kind k).
Proof.
  intros; eapply (proj1 notin_kind_wft_mutual); eauto.
Qed.

Lemma notin_type_bind_t: forall S C t x,
  wfs S ->
  binds C (bind_t t) S ->
  x \notin efv_type t.
Proof.
 intros. 
 assert (Jt: wft nil t k_type).
  eapply wf_ctr_type; eauto. 
  eapply notin_type_wft; eauto.  
Qed. 


Lemma notin_kind_bind_k: forall S T k x,
  wfs S ->
  binds T (bind_k k) S ->
  x \notin efv_kind k.
Proof.
 intros. 
 assert (J: wfk nil k).
  eapply wf_tctr_kind; eauto. 
  eapply notin_kind_wfk; eauto.  
Qed. 
 
Lemma subset_efv_exp_domG: 
forall G e t,
  wfe G e t->
  efv_exp e [<=] dom G.
Proof.
 intros. 
 assert (J:= @notin_exp_wfe G e t).
 assert (J2: forall x, x \in efv_exp e -> x \in dom G). 
  intros.  assert (J1:= J x H).  solve_uniq.
  simpl. solve_uniq.
Qed.
 

(********* should be moved to langExtra.v *******)
Lemma esubst_G_fresh_eq :
forall G e x,
  x \notin efv_ctx G ->
  esubst_G e x G = G.
Proof. 
 intros.
 induction G; eauto.
 destruct a.
  rewrite_env (esubst_G e x (one (VarAsn (exp * exp) a t)) ++ (esubst_G e x G)).
(*  unfold esubst_G.*)
  simpl in *.  rewrite esubst_type_fresh_eq; eauto.
  rewrite IHG; eauto.
 
  destruct p. simpl in *. 
  rewrite esubst_exp_fresh_eq; eauto.
  rewrite esubst_exp_fresh_eq; eauto.
  rewrite IHG; eauto.
Qed.

Lemma efv_ctx_esubst_ctx_upper :
forall G e x,
  efv_ctx (esubst_G e x G) [<=] 
  efv_exp e \u remove x (efv_ctx G).
Proof. 
 intros. induction G; try solve [solve_uniq].
 destruct a. 
 simpl. assert (J:= efv_type_esubst_type_upper t e x).
solve_uniq.

 destruct p. simpl. 
 assert (J1:= efv_exp_esubst_exp_upper e1 e x).
 assert (J2:= efv_exp_esubst_exp_upper e0 e x).
 solve_uniq.
Qed.
 

(********* should be moved to langExtra.v *******)

Lemma dom_eq_substG: forall G e x,
dom G [=] dom (esubst_G e x G).
Proof.
 intros; induction G; auto; try solve [solve_uniq]. 
 destruct a.
 simpl_env.  simpl. solve_uniq.
 destruct p. simpl; auto.
Qed.

Lemma erase_subst_commute: forall G x w,
 erase (esubst_G w x G ) = esubst_ectx w x (erase G).
Proof.
 intros. induction G; eauto.
 destruct a. 
 simpl_env in *; auto.
 destruct p.
 simpl in *.  rewrite IHG; auto. 
Qed.

Lemma subst_binds: forall x t w y G,
  binds y t G ->
  binds y (esubst_type w x t) (esubst_G w x G).
Proof.
 intros.
 induction G; eauto.
 destruct a.
 assert (Jb1: (y= a /\ t = t0) \/ binds y t G).
    eapply binds_cons_1; eauto.
    destruct Jb1 as [(Jeq1, Jeq2) | Jb1].
   subst; auto. simpl. 
   eapply binds_cons_2; auto.
  simpl; eapply binds_cons_3; eauto.

 simpl_env. rewrite_env (esubst_G w x (one (AltAsn type p)) ++ (esubst_G w x G)).
 apply binds_app_2; eauto.
 eapply IHG; auto. inversion H; auto. inversion H0.
Qed.
 

Lemma subst_ctxeq: forall G G' G1 G2 G1' G2' t t' w x,
  ctxeq G G' ->
  G = (G2++ VarAsn (exp*exp) x t :: G1) -> 
  G' = (G2'++ VarAsn (exp*exp) x t' :: G1')  ->
  wfc G -> wfc G' ->
  pure w ->
  wfe G1 w t ->
  ctxeq (esubst_G w x G2 ++ G1) (esubst_G w x G2' ++ G1'). 
Proof.
 intros  G G' G1 G2 G1' G2' t t' w x Heq HG HG' HwfG HwfG' Hpure Hwfe.  
 generalize dependent  G1; 
 generalize dependent G2; 
 generalize dependent G1'; 
 generalize dependent G2'; 
 generalize dependent t; 
 generalize dependent t'; 
 generalize dependent x.
 induction Heq; intros; subst; auto.

 Case "nil".
  destruct (app_eq_nil G2 (VarAsn (exp * exp) x t :: G1)) as (J1, J3); eauto.
  inversion J3.

 Case "Term".
   simpl_env in *.
  assert 
    (J: (exists qs, G2' =  VarAsn (exp * exp) x t' :: qs 
                     /\ G'= qs ++ one (VarAsn (exp * exp) x0 t'0) ++ G1') \/ 
             (G2' = nil /\ one (VarAsn (exp * exp) x0 t'0) ++ G1'
               = VarAsn (exp * exp) x t' :: G')).
  eapply one_eq_app; eauto. 
  destruct J as [[G0' [J1' J2']] | [J1' J2']].
  assert 
    (J: (exists qs, G2 =  VarAsn (exp * exp) x t :: qs 
                     /\ G= qs ++ one (VarAsn (exp * exp) x0 t0) ++ G1) \/ 
             (G2 = nil /\ one (VarAsn (exp * exp) x0 t0) ++ G1
               = VarAsn (exp * exp) x t :: G)).
  eapply one_eq_app; eauto. 
  destruct J as [[G0 [J1 J2]] | [J1 J2]].
  subst. 
  unfold esubst_G; simpl. 
  inversion HwfG; inversion HwfG'; subst. 
  eapply CQ_Term; eauto.  rewrite_env (esubst_G w x0 G0). rewrite_env (esubst_G w x0 G0').
  eapply IHHeq; eauto. simpl_env; eauto.  simpl_env; eauto. 
  rewrite_env (esubst_G w x0 G0). 
  simpl_env in H. simpl in H.
  rewrite <- (@esubst_G_fresh_eq G1 w x0); auto.
  unfold esubst_G. rewrite <- map_app. rewrite_env (esubst_G w x0 (G0++G1)).
  rewrite erase_subst_commute.
  replace (erase G0 ++ erase G1) with (erase (G0 ++ G1)) in H; auto.
  eapply subst_typeeq; eauto. simpl_env; auto. 
  assert (Jwfc:= wfc_head G0 ( VarAsn (exp * exp) x0 t0::G1) H3).
  inversion Jwfc; subst; auto.
  eapply notin_G_wfc; eauto.

  subst. simpl_env in *. inversion HG; subst.  
  inversion HwfG'; subst. solve_uniq.

  assert 
    (J: (exists qs, G2 =  VarAsn (exp * exp) x t :: qs 
                     /\ G= qs ++ one (VarAsn (exp * exp) x0 t0) ++ G1) \/ 
             (G2 = nil /\ one (VarAsn (exp * exp) x0 t0) ++ G1
               = VarAsn (exp * exp) x t :: G)).
  eapply one_eq_app; eauto. 
  destruct J as [[G0 [J1 J2]] | [J1 J2]].
  subst. simpl_env in *.  inversion HG'; subst. 
   inversion HwfG; subst. solve_uniq.

  subst; simpl_env in *. inversion HG; inversion HG'; subst; eauto.

Case "Eq". 
   assert 
    (J: (exists qs, G2' =  AltAsn type (e1', e2'):: qs 
                     /\ G'= qs ++ one (VarAsn (exp * exp) x t') ++ G1') \/ 
             (G2' = nil /\ one (VarAsn (exp * exp) x t') ++ G1'
               = AltAsn type (e1', e2'):: G')).
  eapply one_eq_app; eauto. 
  destruct J as [[G0' [J1' J2']] | [J1' J2']].
   assert 
    (J: (exists qs, G2 =  AltAsn type (e1, e2):: qs 
                     /\ G= qs ++ one (VarAsn (exp * exp) x t) ++ G1) \/ 
             (G2 = nil /\ one (VarAsn (exp * exp) x t) ++ G1
               = AltAsn type (e1, e2):: G)).
  eapply one_eq_app; eauto. 
  destruct J as [[G0 [J1 J2]] | [J1 J2]].
  subst. simpl. eapply CQ_Eq; eauto.
  inversion HwfG; inversion HwfG'; eauto.
  eapply IHHeq; eauto. simpl_env; eauto. simpl_env; eauto.
  simpl_env in H. 
  rewrite <- (@esubst_G_fresh_eq G1 w x); auto.
  unfold esubst_G. rewrite <- map_app. rewrite_env (esubst_G w x (G0++G1)).
  rewrite erase_subst_commute.
  replace (erase G0 ++ erase G1) with (erase (G0 ++ G1)) in H; auto.
  eapply subst_isEq; eauto. simpl_env; auto. 
  inversion HwfG; subst. 
  assert (Jwfc:= wfc_head G0 ( VarAsn (exp * exp) x t::G1) H4).
  inversion Jwfc; subst; auto.
  eapply notin_G_wfc; eauto.

  simpl_env in H0. 
  rewrite <- (@esubst_G_fresh_eq G1 w x); auto.
  unfold esubst_G. rewrite <- map_app. rewrite_env (esubst_G w x (G0++G1)).
  rewrite erase_subst_commute.
  replace (erase G0 ++ erase G1) with (erase (G0 ++ G1)) in H0; auto.
  eapply subst_isEq; eauto. simpl_env; auto. 
  inversion HwfG; subst. 
  assert (Jwfc:= wfc_head G0 ( VarAsn (exp * exp) x t::G1) H4).
  inversion Jwfc; subst; auto.
  eapply notin_G_wfc; eauto.

  subst. simpl_env in *. inversion HG; subst.  

  assert 
    (J: (exists qs, G2 =  AltAsn type (e1, e2):: qs 
                     /\ G= qs ++ one (VarAsn (exp * exp) x t) ++ G1) \/ 
             (G2 = nil /\ one (VarAsn (exp * exp) x t) ++ G1
               = AltAsn type (e1, e2):: G)).
  eapply one_eq_app; eauto. 
  destruct J as [[G0 [J1 J2]] | [J1 J2]].
  subst. simpl_env in *.  inversion HG'.

  subst; simpl_env in *; auto.   inversion HG; inversion HG'; subst; eauto.
Qed.


Definition subst_wfk_prop G k (WF : wfk G k) :=
  forall G1 G2 x tw w, 
  G = G2++ VarAsn (exp*exp) x tw :: G1 ->
  pure w ->
  wfe G1 w tw ->
  wfk (esubst_G w x G2 ++ G1)  (esubst_kind w x k).
Definition subst_wft_prop G t k (WF : wft G t k) :=
  forall G1 G2 x tw w, 
  G = G2++ VarAsn (exp*exp) x tw :: G1 ->
  pure w ->
  wfe G1 w tw ->
  wft (esubst_G w x G2 ++ G1) (esubst_type w x t) (esubst_kind w x k).
Definition subst_wfe_prop G e t (WF : wfe G e t) :=
  forall G1 G2 x tw w, 
  G = G2++ VarAsn (exp*exp) x tw :: G1 ->
  pure w ->
  wfe G1 w tw ->
 wfe (esubst_G w x G2 ++ G1) (esubst_exp w x e) (esubst_type w x t).
Definition subst_wfpt_prop G e u pt k (WF : wfpt G e u pt k) :=
  forall G1 G2 x tw w, 
  G = G2++ VarAsn (exp*exp) x tw :: G1 ->
  pure w ->
  wfe G1 w tw ->
  wfpt (esubst_G w x G2 ++ G1) (esubst_exp w x e)
   (esubst_exp w x u) (esubst_pt w x pt)
    (esubst_kind w x k).
Definition subst_wfpe_prop G e u pe t (WF : wfpe G e u pe t) :=
  forall G1 G2 x tw w, 
  G = G2++ VarAsn (exp*exp) x tw :: G1 ->
  pure w ->
  wfe G1 w tw ->
  wfpe (esubst_G w x G2 ++ G1) (esubst_exp w x e)
   (esubst_exp w x u) (esubst_pe w x pe)
    (esubst_type w x t).
Definition subst_wfc_prop G (WF : wfc G) :=
 forall G1 G2 x tw w, 
  G = G2++ VarAsn (exp*exp) x tw :: G1 ->
  pure w ->
  wfe G1 w tw ->
  wfc (esubst_G w x G2 ++ G1).

(* 17 *)
Lemma subst_wf:
  (forall G k WF, @subst_wfk_prop G k WF) /\
  (forall G t k WF, @subst_wft_prop G t k WF) /\
  (forall G e u pt k WF, @subst_wfpt_prop G e u pt k WF) /\
  (forall G e t WF, @subst_wfe_prop G e t WF) /\
  (forall G e u pe t WF, @subst_wfpe_prop G e u pe t WF) /\
  (forall G WF, @subst_wfc_prop G WF).
Proof.
  eapply wf_mutind with  
  (P  := subst_wfk_prop) 
  (P0 := subst_wft_prop) 
  (P1 := subst_wfpt_prop)
  (P2 := subst_wfe_prop)  
  (P3 := subst_wfpe_prop)
  (P4 := subst_wfc_prop); 
  unfold subst_wfk_prop, subst_wft_prop, subst_wfpt_prop, 
  subst_wfe_prop, subst_wfpe_prop, subst_wfc_prop; intros; subst ; 
  eauto  using subst_isEq, subst_incon, subst_typeeq, subst_kindeq, subst_ctxereq, subst_ctxeq, subst_pure. 

Case "WFK_Pi". 
 eapply WFK_Pi with (L:=L \u {{x}}); eauto. 
 intros. fold esubst_kind. 
 rewrite_env  (esubst_G w0 x (VarAsn (exp * exp) x0 t :: G2) ++ G1).
 replace (e_Var_f x0) with (esubst_exp w0 x (e_Var_f x0)); auto. 
 rewrite <- esubst_kind_open_kind_wrt_exp. 
 eapply (H x0); eauto.  simpl_env; auto. 
 eapply lc_wfe; eauto.    rewrite esubst_exp_fresh_eq; eauto. 

Case "K_Incon".
 apply K_InCon; auto.  eapply H; eauto.
 simpl_env in i. assert (J:= subst_incon x H2 i).
 simpl_env.
 rewrite <- (@esubst_G_fresh_eq G1 w1 x); auto.
 repeat rewrite erase_subst_commute; auto.
 unfold esubst_ectx in *; simpl_env in *; auto.
 rewrite <- List.map_app; auto.
 assert (Jw:= wfc_head G2 (VarAsn (exp * exp) x tw :: G1) w).
 inversion Jw; subst; auto. eapply notin_G_wfc; eauto.
  
 assert (J:=efv_type_esubst_type_upper t w1 x).
 assert (J1: efv_exp w1 [<=] dom G1). 
 eapply subset_efv_exp_domG; eauto. 
 assert (J2:= dom_eq_substG G2 w1 x).  solve_uniq.

 eapply esubst_type_lc_type; eauto. 
 eapply H0; eauto.

Case "K_Const". 
 rewrite esubst_kind_fresh_eq; auto.
 eapply K_Const; eauto.
 eapply notin_kind_bind_k with (S:=Signature); eauto using wfs_sig.

Case "K_Pi ". 
 eapply K_Pi with (L:=L \u {{x}}); eauto. 
 intros. fold esubst_type. 
 rewrite_env  (esubst_G w0 x (VarAsn (exp * exp) x0 t1 :: G2) ++ G1).
 replace (e_Var_f x0) with (esubst_exp w0 x (e_Var_f x0)); auto. 
 rewrite <- esubst_type_open_type_wrt_exp. 
 eapply (H x0); eauto.  simpl_env; auto. 
 eapply lc_wfe; eauto.    rewrite esubst_exp_fresh_eq; eauto. 

Case "K_Sigma".
 eapply K_Sigma with (L:=L \u {{x}}); eauto. 
 intros. fold esubst_type. 
 rewrite_env  (esubst_G w0 x (VarAsn (exp * exp) x0 t1 :: G2) ++ G1).
 replace (e_Var_f x0) with (esubst_exp w0 x (e_Var_f x0)); auto. 
 rewrite <- esubst_type_open_type_wrt_exp. 
 eapply (H x0); eauto.  simpl_env; auto. 
 eapply lc_wfe; eauto.    rewrite esubst_exp_fresh_eq; eauto. 

Case "K_App". 
eapply K_App with (t1:= esubst_type w2 x t1) (k1:=esubst_kind w2 x k1 ) (L:=L\u{{x}}); eauto.
intros.
 replace (e_Var_f x0) with (esubst_exp w2 x (e_Var_f x0)); auto. 
 rewrite <- esubst_kind_open_kind_wrt_exp. 
 replace 
     ((esubst_exp w2 x (e_Var_f x0), esubst_exp w2 x e)
      :: erase (esubst_G w2 x G2 ++ G1))
   with (esubst_ectx w2 x 
           (((e_Var_f x0), e) :: erase (G2 ++ G1))).

 eapply subst_kindeq; eauto. simpl_env in *. 
 rewrite_env ((e_Var_f x0, e) :: erase G2 ++ erase G1).
 eapply k0; eauto.  

  replace (esubst_G w2 x G2 ++ G1) with (esubst_G w2 x G2 ++ esubst_G w2 x G1).
  simpl_env.
  repeat rewrite erase_subst_commute; auto. 
  unfold esubst_ectx; simpl_env; auto.
  repeat rewrite List.map_app; auto.

  rewrite (@esubst_G_fresh_eq G1 w2 x); simpl_env; auto.
  assert (Jwf: wfc  (G2 ++ VarAsn (exp * exp) x tw :: G1)).          
   eapply reg_wfe_c; eauto.
   assert (J:= wfc_head G2 (VarAsn (exp * exp) x tw :: G1) Jwf).
   inversion J; subst. 
  eapply notin_G_wfc; eauto.  eapply lc_wfe; eauto.
 rewrite esubst_exp_fresh_eq; eauto. 

Case "K_Case".
 simpl. eapply K_Case; eauto.

Case "K_KConv".
 eapply K_KConv with (k:=  (esubst_kind w1 x k)); eauto.
 rewrite <- (@esubst_G_fresh_eq G1 w1 x); auto.
 simpl_env.  repeat rewrite erase_subst_commute; auto.
 replace  (esubst_ectx w1 x (erase G2) ++ esubst_ectx w1 x (erase G1))
 with (esubst_ectx w1 x (erase G2 ++ erase G1)).
 eapply subst_kindeq; eauto.
 simpl_env in k0. eapply k0; eauto.
 
 unfold esubst_ectx; simpl_env; auto. 
 rewrite List.map_app; auto.

 assert (Jwf: wfc  (G2 ++ VarAsn (exp * exp) x tw :: G1)).          
   eapply reg_wft_c; eauto.
   assert (J:= wfc_head G2 (VarAsn (exp * exp) x tw :: G1) Jwf).
   inversion J; subst. 
  eapply notin_G_wfc; eauto.  

(*
Case "K_CConv". 
 destruct (ctxeq_concat G2 (VarAsn (exp * exp) x tw :: G1) c) 
   as [G2' [G' (Jeq, Jctxeq2)]].
 inversion Jctxeq2; subst. 
 assert (Jwfc: wfc (G2' ++ VarAsn (exp * exp) x t0 :: G0) ).
 simpl_env in *; eapply reg_wft_c; eauto.
 assert (Jwfw: wfe G0 w1 t0).
 eapply T_TConv with (t:= tw); eauto using sym_typeeq.
 eapply T_CConv with (G:=G1); eauto using sym_ctxeq.
 eapply wfc_head with (G1:=G2' ++ one (VarAsn (exp * exp) x t0)); simpl_env; eauto.
  assert (J:= wfc_head G2' (VarAsn (exp * exp) x t0 :: G0) Jwfc).
  inversion J; subst; auto.

  eapply K_CConv with (G:= esubst_G w1 x G2' ++ G0) ; eauto.
  eapply subst_ctxeq with (t:=t0) (t':=tw); eauto. 
*)
(*Case "PK_Empty". 
 simpl.  eapply PK_Empty;eauto. 
 eapply esubst_exp_lc_exp; eauto.
 eapply esubst_exp_lc_exp; eauto. 
*)
Case "PK_Cons". 
  assert (J2 : x `notin` efv_type (t_Pi s (t_App (t_Con T) ui))).
  eapply notin_type_bind_t with (S:=Signature); eauto using wfs_sig. 
  
  simpl. eapply PK_Cons with (L:=L\u{{x}}); eauto.
  intros. replace (e_Var_f x0) with (esubst_exp w1 x (e_Var_f x0)); auto. 
  rewrite <- esubst_type_open_type_wrt_exp. 
  replace s with (esubst_type w1 x s). 
  replace (open_exp_wrt_exp ui (esubst_exp w1 x (e_Var_f x0)))
    with (esubst_exp w1 x (open_exp_wrt_exp ui (e_Var_f x0))).
  rewrite_env 
  ((esubst_G w1 x 
    (AltAsn type (e, e_Con C (e_Var_f x0))
      :: AltAsn type (u, open_exp_wrt_exp ui (e_Var_f x0))
      :: VarAsn (exp * exp) x0 s ::G2))  
  ++G1).
  eapply H0; eauto. auto.
  rewrite esubst_exp_fresh_eq; eauto. 
  rewrite esubst_exp_fresh_eq; eauto.
  assert (J1: x \notin (efv_exp ui)). simpl in *; solve_notin.
  assert (J3: efv_exp (open_exp_wrt_exp ui (e_Var_f x0) )
    [<=]efv_exp (e_Var_f x0) \u efv_exp ui).
  eapply efv_exp_open_exp_wrt_exp_rec_upper; eauto.
  simpl in *.  solve_uniq. 
  rewrite esubst_type_fresh_eq; eauto.
  simpl in *.  solve_uniq. 
  
  eapply lc_wfe; eauto.
  rewrite esubst_exp_fresh_eq; eauto.

Case "T_Incon". 
 apply T_InCon; auto. eapply H; eauto.
 simpl_env in i. assert (J:= subst_incon x H2 i).
 simpl_env.
 rewrite <- (@esubst_G_fresh_eq G1 w1 x); auto.
 repeat rewrite erase_subst_commute; auto.
 unfold esubst_ectx in *; simpl_env in *; auto.
 rewrite <- List.map_app; auto.
 assert (Jw:= wfc_head G2 (VarAsn (exp * exp) x tw :: G1) w).
 inversion Jw; subst; auto. eapply notin_G_wfc; eauto.
  
 assert (J:=efv_exp_esubst_exp_upper e w1 x).
 assert (J1: efv_exp w1 [<=] dom G1). 
 eapply subset_efv_exp_domG; eauto. 
 assert (J2:= dom_eq_substG G2 w1 x).  solve_uniq.
 
 eapply H0; eauto.

Case "T_Var".
  assert (Juniq: uniq (G2 ++ VarAsn (exp * exp) x0 tw :: G1)). 
   eapply uniq_wfc; eauto.
  assert (Jb:binds x t G2 /\ x \notin dom (VarAsn (exp * exp) x0 tw :: G1)  
             \/ (binds x t (VarAsn (exp * exp) x0 tw :: G1) 
                /\ x \notin dom G2)).
  eapply binds_app_uniq; eauto.
  destruct Jb as [(Jb, Jnotin)|(Jb, Jnotin)].
  SCase "x in G2".
    simpl. destruct (x==x0). 
    subst.  solve_uniq.
    eapply T_Var; eauto.
    eapply binds_app_1. eapply subst_binds; eauto.
  SCase "x \notin G2".
    assert (Jb1: (x = x0 /\ t = tw) \/ binds x t G1).
    eapply binds_cons_1; eauto.
    destruct Jb1 as [(Jeq1, Jeq2) | Jb1].
    simpl.  destruct (x == x0).
    subst.
    rewrite (@esubst_type_fresh_eq tw); auto. 
    rewrite_env (nil++esubst_G w0 x0 G2 ++ G1).
    replace G1 with (nil++G1) in H2; auto.
    eapply weakening_wfe; eauto.
    simpl_env; eapply H; eauto.
    eapply notin_type_wfe; eauto.
    solve_uniq. contradiction.

  SSCase "x \in G1".
    simpl.  destruct (x==x0).
    subst.  assert (Jin: x0 \in dom G1).  eapply binds_In; eauto. 
    solve_uniq.

    eapply T_Var; eauto.
    eapply binds_app_2; auto. rewrite (esubst_type_fresh_eq ); auto.
    eapply notin_type_binds with (G:=G1); eauto. 
    eapply reg_wfe_c; eauto. solve_uniq.

Case "T_Ctr". 
 assert (J:=@notin_type_bind_t Signature C (t_Pi s (t_App (t_Con T) u)) x (wfs_sig) b).
 simpl in J. 
 eapply T_Ctr with (s:= esubst_type w1 x s) (u:= u) (T:=T) (L:=L\u{{x}}); eauto.
 rewrite esubst_type_fresh_eq; auto.
 fold esubst_exp. 
 intros. 
 replace (e_Var_f x0) with (esubst_exp w1 x (e_Var_f x0)); auto. 
 replace (t_App (t_Con T) (open_exp_wrt_exp u (esubst_exp w1 x (e_Var_f x0))))
   with (esubst_type w1 x (t_App (t_Con T) 
      (open_exp_wrt_exp u (e_Var_f x0)))).
 replace 
     ((esubst_exp w1 x (e_Var_f x0), esubst_exp w1 x e)
      :: erase (esubst_G w1 x G2 ++ G1))
   with (esubst_ectx w1 x 
           (((e_Var_f x0), e) :: erase (G2 ++ G1))).

 eapply subst_typeeq; eauto. simpl_env in *. 
 rewrite_env ((e_Var_f x0, e) :: erase G2 ++ erase G1).
 eapply t0; eauto.

  replace (esubst_G w1 x G2 ++ G1) with (esubst_G w1 x G2 ++ esubst_G w1 x G1).
  simpl_env.
  repeat rewrite erase_subst_commute; auto. 
  unfold esubst_ectx; simpl_env; auto.
  repeat rewrite List.map_app; auto.

  rewrite (@esubst_G_fresh_eq G1 w1 x); simpl_env; auto.
  assert (Jwf: wfc  (G2 ++ VarAsn (exp * exp) x tw :: G1)).          
   eapply reg_wfe_c; eauto.
   assert (J1:= wfc_head G2 (VarAsn (exp * exp) x tw :: G1) Jwf).
   inversion J1; subst. 
  eapply notin_G_wfc; eauto.

 replace (open_exp_wrt_exp u (esubst_exp w1 x (e_Var_f x0)))
   with (open_exp_wrt_exp (esubst_exp w1 x u) (esubst_exp w1 x (e_Var_f x0))).
 rewrite <- esubst_exp_open_exp_wrt_exp; simpl; auto. 
  assert (Ju: esubst_exp w1 x u = u).
  rewrite esubst_exp_fresh_eq; eauto.
  rewrite Ju; auto.
 rewrite esubst_exp_fresh_eq; eauto. 

Case "T_Fun".
 eapply T_Fun with (L:=L \u {{x}}); eauto. 
 intros. fold esubst_exp. fold esubst_type.
 rewrite_env  (esubst_G w0 x (VarAsn (exp * exp) x0 t1 :: G2) ++ G1).
 replace (e_Var_f x0) with (esubst_exp w0 x (e_Var_f x0)); auto. 
 rewrite <- esubst_type_open_type_wrt_exp. 
 rewrite <- esubst_exp_open_exp_wrt_exp. 
 eapply (H x0); eauto.  simpl_env; auto. 
 eapply lc_wfe; eauto.  eapply lc_wfe; eauto.    
 rewrite esubst_exp_fresh_eq; eauto. 

Case "T_App". 
 eapply T_App with (t1:= esubst_type w2 x t1) (t2:=esubst_type w2 x t2 ) (L:=L\u{{x}}); eauto.
 fold esubst_exp. intros.
 replace (e_Var_f x0) with (esubst_exp w2 x (e_Var_f x0)); auto. 
 rewrite <- esubst_type_open_type_wrt_exp. 
 replace 
     ((esubst_exp w2 x (e_Var_f x0), esubst_exp w2 x e2)
      :: erase (esubst_G w2 x G2 ++ G1))
   with (esubst_ectx w2 x 
           (((e_Var_f x0), e2) :: erase (G2 ++ G1))).

 eapply subst_typeeq; eauto. simpl_env in *. 
 rewrite_env ((e_Var_f x0, e2) :: erase G2 ++ erase G1).
 eapply t0; eauto. 

  replace (esubst_G w2 x G2 ++ G1) with (esubst_G w2 x G2 ++ esubst_G w2 x G1).
  simpl_env.
  repeat rewrite erase_subst_commute; auto. 
  unfold esubst_ectx; simpl_env; auto.
  repeat rewrite List.map_app; auto.

  rewrite (@esubst_G_fresh_eq G1 w2 x); simpl_env; auto.
  assert (Jwf: wfc  (G2 ++ VarAsn (exp * exp) x tw :: G1)).          
   eapply reg_wfe_c; eauto.
   assert (J:= wfc_head G2 (VarAsn (exp * exp) x tw :: G1) Jwf).
   inversion J; subst. 
  eapply notin_G_wfc; eauto.

  eapply lc_wfe; eauto.
 rewrite esubst_exp_fresh_eq; eauto. 

Case "T_Fix". 
 eapply T_Fix with (L:=L \u {{x}}); eauto. 
 intros. fold esubst_exp.
 rewrite_env  (esubst_G w0 x (VarAsn (exp * exp) f (t_Pi t1 t2) :: G2) ++ G1).
 replace (e_Var_f f) with (esubst_exp w0 x (e_Var_f f)); auto. 
 rewrite <- esubst_exp_open_exp_wrt_exp. 
 eapply (H f); eauto.  simpl_env; auto. 
 eapply lc_wfe; eauto.  
 rewrite esubst_exp_fresh_eq; eauto.
 fold esubst_exp. 
 intros. 
 rewrite <- (@esubst_exp_fresh_eq (e_Var_f f) w0 x); auto.
 rewrite <- esubst_exp_open_exp_wrt_exp. 
 eapply is_v_of_exp_esubst_exp ;eauto.
 eapply lc_wfe; eauto.

Case "T_Sigma". 
 simpl. eapply T_Sigma with (t2':= esubst_type w2 x t2') (L:=L\u{{x}}); eauto.
 intros.  
 replace (e_Var_f x0) with (esubst_exp w2 x (e_Var_f x0)); auto. 
 rewrite <- esubst_type_open_type_wrt_exp. 
 replace 
     ((esubst_exp w2 x (e_Var_f x0), esubst_exp w2 x e1)
      :: erase (esubst_G w2 x G2 ++ G1))
   with (esubst_ectx w2 x 
           (((e_Var_f x0), e1) :: erase (G2 ++ G1))).

 eapply subst_typeeq; eauto. simpl_env in *. 
 rewrite_env ((e_Var_f x0, e1) :: erase G2 ++ erase G1).
 eapply t; eauto.

  replace (esubst_G w2 x G2 ++ G1) with (esubst_G w2 x G2 ++ esubst_G w2 x G1).
  simpl_env.
  repeat rewrite erase_subst_commute; auto. 
  unfold esubst_ectx; simpl_env; auto.
  repeat rewrite List.map_app; auto.

  rewrite (@esubst_G_fresh_eq G1 w2 x); simpl_env; auto.
  assert (Jwf: wfc  (G2 ++ VarAsn (exp * exp) x tw :: G1)).          
   eapply reg_wfe_c; eauto.
   assert (J:= wfc_head G2 (VarAsn (exp * exp) x tw :: G1) Jwf).
   inversion J; subst. 
  eapply notin_G_wfc; eauto.

  eapply lc_wfe; eauto.
 rewrite esubst_exp_fresh_eq; eauto. 

 intros. 
 rewrite_env  (esubst_G w2 x (VarAsn (exp * exp) x0 t1 :: G2) ++ G1).
 replace (e_Var_f x0) with (esubst_exp w2 x (e_Var_f x0)); auto. 
 rewrite <- esubst_type_open_type_wrt_exp. simpl in H1.
 eapply (H1 x0); eauto.  simpl_env; auto. 
 eapply lc_wfe; eauto.  
 rewrite esubst_exp_fresh_eq; eauto. 

Case "T_Proj1". 
 simpl. eapply T_Proj1 with (t2:= esubst_type w0 x t2); eauto.

Case "T_Proj2". 
 simpl. eapply T_Proj2 with (t2:= esubst_type w1 x t2) 
 (t1:= esubst_type w1 x t1) (L:=L\u{{x}}); eauto.
 intros.
 replace (e_Var_f x0) with (esubst_exp w1 x (e_Var_f x0)); auto. 
 rewrite <- esubst_type_open_type_wrt_exp. 
 replace 
     ((esubst_exp w1 x (e_Var_f x0), e_Proj1 (esubst_exp w1 x e))
      :: erase (esubst_G w1 x G2 ++ G1))
   with (esubst_ectx w1 x 
           (((e_Var_f x0), e_Proj1 e) :: erase (G2 ++ G1))).

 eapply subst_typeeq; eauto. simpl_env in *. 
 rewrite_env ((e_Var_f x0, e_Proj1 e) :: erase G2 ++ erase G1).
 eapply t0; eauto.

  replace (esubst_G w1 x G2 ++ G1) with (esubst_G w1 x G2 ++ esubst_G w1 x G1).
  simpl_env.
  repeat rewrite erase_subst_commute; auto. 
  unfold esubst_ectx; simpl_env; auto.
  repeat rewrite List.map_app; auto.

  rewrite (@esubst_G_fresh_eq G1 w1 x); simpl_env; auto.
  assert (Jwf: wfc  (G2 ++ VarAsn (exp * exp) x tw :: G1)).          
   eapply reg_wfe_c; eauto.
   assert (J:= wfc_head G2 (VarAsn (exp * exp) x tw :: G1) Jwf).
   inversion J; subst. 
  eapply notin_G_wfc; eauto.

  eapply lc_wfe; eauto.
 rewrite esubst_exp_fresh_eq; eauto. 

Case "T_Case ".  
 simpl. eapply T_Case; eauto. simpl in H0; eapply H0; eauto.
 intros. destruct (e0 Ci si ui); eauto.
 exists  (esubst_exp w2 x x0). eapply find_pe_subst'; eauto.

Case "T_TConv".
 eapply T_TConv with (t:=  (esubst_type w1 x t)); eauto.
 rewrite <- (@esubst_G_fresh_eq G1 w1 x); auto.
 simpl_env.  repeat rewrite erase_subst_commute; auto.
 replace  (esubst_ectx w1 x (erase G2) ++ esubst_ectx w1 x (erase G1))
 with (esubst_ectx w1 x (erase G2 ++ erase G1)).
 eapply subst_typeeq; eauto.
 simpl_env in t0. eapply t0; eauto.
 
 unfold esubst_ectx; simpl_env; auto. 
 rewrite List.map_app; auto.

 assert (Jwf: wfc  (G2 ++ VarAsn (exp * exp) x tw :: G1)).          
   eapply reg_wft_c; eauto.
   assert (J:= wfc_head G2 (VarAsn (exp * exp) x tw :: G1) Jwf).
   inversion J; subst. 
  eapply notin_G_wfc; eauto.  

(*
Case "T_CConv". 
 destruct (ctxeq_concat G2 (VarAsn (exp * exp) x tw :: G1) c) 
   as [G2' [G' (Jeq, Jctxeq2)]].
 inversion Jctxeq2; subst. 
 assert (Jwfc: wfc (G2' ++ VarAsn (exp * exp) x t0 :: G0) ).
 simpl_env in *; eapply reg_wfe_c; eauto.
 assert (Jwfw: wfe G0 w1 t0).
 eapply T_TConv with (t:= tw); eauto using sym_typeeq.
 eapply T_CConv with (G:=G1); eauto using sym_ctxeq.
 eapply wfc_head with (G1:=G2' ++ one (VarAsn (exp * exp) x t0)); simpl_env; eauto.
  assert (J:= wfc_head G2' (VarAsn (exp * exp) x t0 :: G0) Jwfc).
  inversion J; subst; auto.

  eapply T_CConv with (G:= esubst_G w1 x G2' ++ G0) ; eauto.

  eapply subst_ctxeq with (t:=t0) (t':=tw); eauto. 
*)
(*
Case "P_Done".
 simpl.  eapply P_Done;eauto. 
 eapply esubst_exp_lc_exp; eauto.
 eapply esubst_exp_lc_exp; eauto.
 *)
Case  "AP_Cons". 
 assert (J2 : x `notin` efv_type (t_Pi si (t_App (t_Con T) ui))).
 eapply notin_type_bind_t with (S:=Signature); eauto using wfs_sig. 

 simpl. eapply AP_Cons with (L:=L\u{{x}}); eauto.
 intros. replace (e_Var_f xi) with (esubst_exp w1 x (e_Var_f xi)); auto. 
 rewrite <- esubst_exp_open_exp_wrt_exp. 
 replace si with (esubst_type w1 x si). 
 replace (open_exp_wrt_exp ui (esubst_exp w1 x (e_Var_f xi)))
   with (esubst_exp w1 x (open_exp_wrt_exp ui (e_Var_f xi))).
 rewrite_env 
 ((esubst_G w1 x 
   (AltAsn type (e, e_Con Ci (e_Var_f xi))
   :: AltAsn type (u, open_exp_wrt_exp ui (e_Var_f xi))
         :: VarAsn (exp * exp) xi si ::G2))  
   ++G1).
 eapply H; eauto. auto.
 rewrite esubst_exp_fresh_eq; eauto. 
 rewrite esubst_exp_fresh_eq; eauto.
 assert (J1: x \notin (efv_exp ui)). simpl in *; solve_notin.
 assert (J3: efv_exp (open_exp_wrt_exp ui (e_Var_f xi) )
     [<=]efv_exp (e_Var_f xi) \u efv_exp ui).
 eapply efv_exp_open_exp_wrt_exp_rec_upper; eauto.
  simpl in *.  solve_uniq. 
 rewrite esubst_type_fresh_eq; eauto.
 simpl in *.  solve_uniq. 

 eapply lc_wfe; eauto.
 rewrite esubst_exp_fresh_eq; eauto.

Case "C_Empty". 
    assert (J:False). eapply app_cons_not_nil; eauto. destruct J.
Case "C_Term". 
 assert (J:  (exists qs, G2 = VarAsn (exp * exp) x t :: qs 
         /\ G = qs ++  VarAsn (exp * exp) x0 tw :: G1)
      \/ (G2 = nil /\ 
         VarAsn (exp * exp) x0 tw :: G1 = VarAsn (exp * exp) x t :: G)).
  eapply one_eq_app; eauto. 
  destruct J as [[G2' [J1 J2]] | [J1 J2]].
  SCase "G2 = x:t ::G2'".
   subst. 
  rewrite_env ((VarAsn (exp * exp) x (esubst_type w1 x0 t)) 
     :: (esubst_G w1 x0 G2')).
  eapply C_Term; eauto. 
  rewrite_env ((esubst_G w1 x0 G2') ++ G1). 
  assert (J: x \notin dom(esubst_G w1 x0 G2')).
  eapply notin_substG; eauto. eapply notin_exp_wfe; eauto. 
  simpl_env in n. simpl_env.  solve_notin.
  
  SCase "G2 = nil".
  subst.  rewrite app_nil_1 in H1. inversion H1; subst. 
  simpl_env; eauto.

Case "C_Eq". 
 assert (J:  (exists qs, G2 =AltAsn type (e1, e2) :: qs 
         /\ G = qs ++  VarAsn (exp * exp) x tw :: G1)
      \/ (G2 = nil /\ 
         VarAsn (exp * exp) x tw :: G1 = AltAsn type (e1, e2)  :: G)).
  eapply one_eq_app; eauto. 
  destruct J as [[G2' [J1 J2]] | [J1 J2]].
  SCase "G2 = e1~=e2 ::G2'".
   subst. 
  rewrite_env( AltAsn type (esubst_exp w2 x e1, esubst_exp w2 x e2 ) 
     :: (esubst_G w2 x G2' ++ G1)).
  eapply C_Eq; eauto. 
  
  SCase "G2 = nil".
  subst.  rewrite app_nil_1 in H2. inversion H2. 
Qed.

Lemma subst_wfk : forall G1 G2 k x t w,
  wfk (G2++ VarAsn (exp*exp) x t :: G1) k ->
  pure w ->
  wfe G1 w t ->
  wfk (esubst_G w x G2 ++ G1)  (esubst_kind w x k).
Proof.
  pose (J := subst_wf).
  destruct J as [JK [JT [JPT [JE [JPE JG]]]]]. 
  intros.  eapply JK; eauto.
Qed.


Lemma subst_wft : forall G1 G2 t k x tw w ,
  wft (G2++ VarAsn (exp*exp) x tw :: G1) t k ->
  pure w ->
  wfe G1 w tw ->
  wft (esubst_G w x G2 ++ G1) (esubst_type w x t) (esubst_kind w x k).
Proof.
  pose (J := subst_wf).
  destruct J as [JK [JT [JPT [JE [JPE JG]]]]]. 
  intros; eapply JT; eauto.
Qed.


Lemma subst_wfe : forall G1 G2 e t x tw w ,
  wfe (G2++ VarAsn (exp*exp) x tw :: G1) e t ->
  pure w ->
  wfe G1 w tw ->
  wfe (esubst_G w x G2 ++ G1) (esubst_exp w x e) (esubst_type w x t) .
Proof.
  pose (J := subst_wf).
  destruct J as [JK [JT [JPT [JE [JPE JG]]]]]. 
  intros; eapply JE; eauto.
Qed.


Lemma subst_wfc : forall G1 G2 x t w ,
  wfc (G2++ VarAsn (exp*exp) x t :: G1)->
  pure w ->
  wfe G1 w t ->
  wfc (esubst_G w x G2 ++ G1).
Proof.
  pose (J := subst_wf).
  destruct J as [JK [JT [JPT [JE [JPE JG]]]]]. 
  intros; eapply JG; eauto.
Qed.


Lemma subst_wfk_simpl : forall G k x t w,
  wfk (VarAsn (exp*exp) x t :: G) k ->
  pure w ->
  wfe G w t ->
  wfk G (esubst_kind w x k).
Proof.
  intros.  
  rewrite_env (esubst_G w x nil ++ G).
  replace  (VarAsn (exp * exp) x t :: G) with  (nil ++ VarAsn (exp * exp) x t :: G) in H; auto. 
  eapply subst_wfk; eauto.
Qed.


Lemma subst_wft_simpl : forall G t k x tw w ,
  wft (VarAsn (exp*exp) x tw :: G) t k ->
  pure w ->
  wfe G w tw ->
  wft G (esubst_type w x t) (esubst_kind w x k).
Proof.
  intros.  
  rewrite_env (esubst_G w x nil ++ G).
  replace  (VarAsn (exp * exp) x t :: G) with  (nil ++ VarAsn (exp * exp) x t :: G) in H; auto. 
  eapply subst_wft; eauto.
Qed.


Lemma subst_wfe_simpl : forall G e t x tw w ,
  wfe (VarAsn (exp*exp) x tw :: G) e t ->
  pure w ->
  wfe G w tw ->
  wfe G (esubst_exp w x e) (esubst_type w x t) .
Proof.
  intros.  
  rewrite_env (esubst_G w x nil ++ G).
  replace  (VarAsn (exp * exp) x t :: G) with  (nil ++ VarAsn (exp * exp) x t :: G) in H; auto. 
  eapply subst_wfe; eauto.
Qed.


(*************** renaming ************************)

Lemma renaming_wfk: forall G k x y t,
 wfk (VarAsn (exp * exp) x t ::G) (open_kind_wrt_exp k (e_Var_f x)) ->
 x \notin dom G \u efv_kind k \u efv_type t ->
 y \notin dom G \u efv_kind k \u efv_type t ->
 wfk (VarAsn (exp * exp) y t::G) (open_kind_wrt_exp k (e_Var_f y)).
Proof.
  intros. 
  destruct (x==y); subst; auto.
  assert (Jwfcx: wfc (VarAsn (exp * exp) x t ::G) ).
  eapply reg_wfk_c; eauto.
  inversion Jwfcx; subst. 
  assert (Jwfcy: wfc (VarAsn (exp*exp) y t:: G)).
  eapply C_Term; eauto.
  rewrite (@esubst_kind_intro x k (e_Var_f y) ).
  eapply subst_wfk_simpl with (t:=t); eauto.
  simpl_env. eapply weakening_wfk; eauto.
  rewrite_env (VarAsn (exp * exp) x t:: one (VarAsn (exp * exp) y t) ++ G).
  eapply C_Term; eauto.
  rewrite_env (nil++one (VarAsn (exp * exp) y t) ++ G).
  replace G with (nil++G) in H7; auto.
  eapply weakening_wft;eauto.
  eapply T_Var; eauto.
  eapply binds_cons_2; eauto.
  simpl; auto.
Qed.

Lemma renaming_wft: forall G k t x y t0,
 wft (VarAsn (exp * exp) x t0 ::G) (open_type_wrt_exp t (e_Var_f x)) k ->
 x \notin dom G \u efv_type t \u efv_kind k \u efv_type t0 ->
 y \notin dom G \u efv_type t \u efv_kind k \u efv_type t ->
 wft (VarAsn (exp * exp) y t0::G) (open_type_wrt_exp t (e_Var_f y))  
  (esubst_kind (e_Var_f y) x k).
Proof.
  intros. 
  destruct (x==y); subst; auto. 
  rewrite <- (@open_kind_wrt_exp_lc_kind k (e_Var_f y)).
  rewrite <- (@esubst_kind_intro y k (e_Var_f y) ); auto.
  rewrite open_kind_wrt_exp_lc_kind; auto.
  eapply lc_wft; eauto. 

  assert (Jwfcx: wfc (VarAsn (exp * exp) x t0 ::G) ).
  eapply reg_wft_c; eauto.
  inversion Jwfcx; subst. 
  assert (Jwfcy: wfc (VarAsn (exp*exp) y t0:: G)).
  eapply C_Term; eauto.
  rewrite (@esubst_type_intro x t (e_Var_f y) ); auto.
(*  rewrite (@esubst_kind_intro x k (e_Var_f y) ); auto.*)
  eapply subst_wft_simpl with (tw:=t0); eauto.
  simpl_env. eapply weakening_wft; eauto.
  rewrite_env (VarAsn (exp * exp) x t0:: one (VarAsn (exp * exp) y t0) ++ G).
  eapply C_Term; eauto.
  rewrite_env (nil++one (VarAsn (exp * exp) y t0) ++ G).
  replace G with (nil++G) in H7; auto.
  eapply weakening_wft;eauto.
  eapply T_Var; eauto.
  eapply binds_cons_2; eauto.
Qed.

Lemma renaming_wfe: forall G e t x y t0,
 wfe (VarAsn (exp * exp) x t0 ::G) (open_exp_wrt_exp e (e_Var_f x)) t ->
 x \notin dom G \u efv_type t \u efv_exp e \u efv_type t0 ->
 y \notin dom G \u efv_type t \u efv_exp e \u efv_type t ->
 wfe (VarAsn (exp * exp) y t0::G) (open_exp_wrt_exp e (e_Var_f y))
  (esubst_type (e_Var_f y) x t).
Proof.
  intros. 
  destruct (x==y); subst; auto.

  rewrite <- (@open_type_wrt_exp_lc_type t (e_Var_f y)).
  rewrite <- (@esubst_type_intro y t (e_Var_f y) ); auto.
  rewrite open_type_wrt_exp_lc_type; auto.
  eapply lc_wfe; eauto.

  assert (Jwfcx: wfc (VarAsn (exp * exp) x t0 ::G) ).
  eapply reg_wfe_c; eauto.
  inversion Jwfcx; subst. 
  assert (Jwfcy: wfc (VarAsn (exp*exp) y t0:: G)).
  eapply C_Term; eauto.
(*  rewrite (@esubst_type_intro x t (e_Var_f y) ); auto.*)
  rewrite (@esubst_exp_intro x e (e_Var_f y) ); auto.
  eapply subst_wfe_simpl with (tw:=t0); eauto.
  simpl_env. eapply weakening_wfe; eauto.
  rewrite_env (VarAsn (exp * exp) x t0:: one (VarAsn (exp * exp) y t0) ++ G).
  eapply C_Term; eauto.
  rewrite_env (nil++one (VarAsn (exp * exp) y t0) ++ G).
  replace G with (nil++G) in H7; auto.
  eapply weakening_wft;eauto.
  eapply T_Var; eauto.
  eapply binds_cons_2; eauto.
Qed.

(****************inversions typing ******************)
(*
Lemma lc_ctx_erase : forall G,
 lc_ctx G ->
 lc_ectx (erase G).
Proof.
  intros. induction G; eauto.
  destruct a.
  inversion H; subst;  simpl; eauto.
  destruct p. simpl in *. 
  destruct H as (J1, (J2, J3)). repeat split; eauto.
Qed. 
*)
Lemma inversion_elam: forall D G e0 e t t1 t2 L,
  D = erase G ->
 ~ (incon D) ->
 wfe G e0 t ->
 e0 = (e_Lam e) ->
 typeeq D t (t_Pi t1 t2) ->
 wft G  (t_Pi t1 t2) k_type ->
 exists x,   
   x \notin  L  /\
   wfe  (AssumeList.VarAsn (exp * exp)  x   t1  ::  G )  
   ( open_exp_wrt_exp e (e_Var_f x) )  
   ( open_type_wrt_exp t2 (e_Var_f x) ).
Proof.
 intros. generalize dependent D.
 induction H1; intros; subst; try solve [inversion H2]; eauto.

 Case "InCon".  contradiction.
 Case "Lam".  
  clear H0.
  assert (Jeq1 : typeeq (erase G) t0 t1). eapply inversion_typeeq_pi1; eauto.
  assert (Jeq2: exists x,
    x \notin  (L\u L0\u dom G \u efv_type t1 \u efv_type t2)  /\
   typeeq (erase G)  ( open_type_wrt_exp t3 (e_Var_f x) )  
   ( open_type_wrt_exp t2 (e_Var_f x) )).
  eapply inversion_typeeq_pi2 with (t:= t_Pi t0 t3) (s:=t_Pi t1 t2); eauto. 
  destruct Jeq2 as [x (Jin, Jeq2)].
  exists x; intros; inversion H2; subst; split; eauto.
  eapply T_TConv with (t:= (open_type_wrt_exp t3 (e_Var_f x)));eauto.
  eapply ctx_conv_wfe with (G:= VarAsn (exp * exp) x t0 :: G); eauto.
  eapply C_Term; eauto. eapply reg_wft_c;eauto.  eapply inversion_wft_pi1; eauto. 
  eapply CQ_Term; eauto. eapply ref_ctxeq; eauto. 
  assert (J: exists x,
    x \notin  (L\u L0\u dom G\u efv_type t1 \u efv_type t2)  /\
   wft (VarAsn (exp*exp) x t1 ::G) ( open_type_wrt_exp t2 (e_Var_f x) ) k_type).
  eapply inversion_wft_pi2;eauto.
  destruct J as [y (Jx, Jwf)].
  rewrite <- (@esubst_kind_fresh_eq k_type (e_Var_f x) y); auto.
  eapply renaming_wft; eauto.
  
 Case "TConv".
 eapply IHwfe; eauto. eapply trans_typeeq; eauto.

(*
 Case "CConv".  
  destruct IHwfe with (D:= erase G); eauto.
  eapply K_CConv; eauto using sym_ctxeq.  eapply reg_wfe_c; eauto.
  unfold not; intro; unfold not in H5; apply H5. 
  apply ctx_conv_incon with (D:=erase G); auto.
  eapply ctxeq_erase ; eauto.

  eapply ctx_conv_typeeq; eauto. eapply ctxeq_erase; eauto using sym_ctxeq.

  destruct H2; exists x; split; auto. 
  eapply T_CConv; eauto.
  eapply CQ_Term; eauto. eapply ref_typeeq; eauto.
  destruct (lc_ctxeq H) as (J1, J2); eauto. eapply lc_ctx_erase; eauto. 
  destruct (lc_wft H4) as (J1, (J2, J3)).  inversion J2; auto.
  eapply C_Term; eauto.  
  assert (J: wfc (VarAsn (exp * exp) x t1 :: G)).
  eapply reg_wfe_c; eauto.
  assert (J1: dom G [=] dom G'). eapply ctxeq_dom_eq; eauto. 
  inversion J; subst.  solve_uniq.
  eapply inversion_wft_pi1; eauto.
*)
Qed.
 
 

Lemma inversion_Fix: forall D G e0 e t L,
  D = erase G ->
 ~ (incon D) ->
 wfe G e0 t ->
 e0 = (e_Fix e) ->
 wft G  t k_type ->
 (exists f,   
   f \notin  L  /\
   wfe  (AssumeList.VarAsn (exp * exp)  f  t  ::  G )  
   ( open_exp_wrt_exp e (e_Var_f f) )  t 
  /\(is_v_of_exp   ( open_exp_wrt_exp e (e_Var_f f) )  ) ).
Proof.
 intros. generalize dependent D.
 induction H1; intros; subst; try solve [inversion H2]; eauto.

 Case "InCon".  contradiction.
 Case "Fix".  
 clear H0.
  inversion H2; subst. 
  pick fresh z; exists z; split; eauto.
 
 Case "TConv".
 assert (Jwft: wft G t k_type). eapply wft_wfe; eauto.
 destruct IHwfe with (D:=erase G); eauto.
 destruct H2. exists x; split; auto.
 assert (Jlct: lc_type t). eapply lc_wft; eauto.
 assert (Jlct': lc_type t'). eapply lc_wft; eauto.
 destruct H4.
 assert (Jwfc:= reg_wfe_c H4). 
 assert (Jx: x \notin dom G). inversion Jwfc; auto.
 split; auto.
 eapply T_TConv; eauto. eapply ctx_conv_wfe; eauto.
 eapply C_Term; eauto. eapply reg_wft_c; eauto. 
 eapply CQ_Term; eauto using sym_typeeq. eapply ref_ctxeq; eauto. 
 rewrite_env (nil++one(VarAsn (exp * exp) x t')++ G).
 replace G with (nil ++G) in H3; auto.
 eapply weakening_wft; eauto.
 rewrite_env (VarAsn (exp * exp) x t':: G); eapply C_Term; eauto.
 eapply reg_wfe_c;eauto. 
(*
 Case "CConv".  
  assert (Jwft: wft G t k_type). eapply wft_wfe; eauto.
  destruct IHwfe with (D:= erase G); eauto.
  intro; apply H5; auto. eapply ctx_conv_incon; eauto.
  eapply ctxeq_erase ; eauto.
  destruct H2; exists x; split; auto. 
  destruct H4; split; auto.
  eapply T_CConv; eauto.
  eapply CQ_Term; eauto. eapply ref_typeeq; eauto. 
  eapply lc_ctx_erase; eauto. eapply lc_wft; eauto.
  eapply lc_wfe; eauto.

  eapply C_Term; eauto.  
  assert (J: wfc (VarAsn (exp * exp) x t:: G)).
  eapply reg_wfe_c; eauto.
  assert (J1: dom G [=] dom G'). eapply ctxeq_dom_eq; eauto. 
  inversion J; subst.  solve_uniq.
*)
Qed.
 
 

Lemma inversion_Pair: forall D G e0 e1 e2 t t1 t2 L,
  D = erase G ->
 ~ (incon D) ->
 wfe G e0 t ->
 e0 = (e_Sie e1 e2) ->
 typeeq D t (t_Sigma t1 t2) ->
 wft G  (t_Sigma t1 t2) k_type ->
( wfe G e1 t1 
 /\ (exists t2', wfe G e2 t2' 
    /\ exists x, x \notin  L  /\
   typeeq  ((e_Var_f x, e1):: D)  t2'
   ( open_type_wrt_exp t2 (e_Var_f x) ))).
Proof.
 intros. generalize dependent D.
 induction H1; intros; subst; try solve [inversion H2]; eauto.

 Case "InCon".  contradiction.
 Case "Sigma".   
  clear IHwfe1 IHwfe2.
  assert (Jeq1 : typeeq (erase G) t0 t1). eapply inversion_typeeq_sigma1; eauto.
  assert (Jeq2: exists x,
    x \notin  (L\u L0\u dom G \u efv_type t1 \u efv_type t2)  /\
   typeeq (erase G)  ( open_type_wrt_exp t3 (e_Var_f x) )  
   ( open_type_wrt_exp t2 (e_Var_f x) )).
  eapply inversion_typeeq_sigma2 with (t:= t_Sigma t0 t3) (s:=t_Sigma t1 t2); eauto. 
  destruct Jeq2 as [x (Jin, Jeq2)].
  intros; inversion H2; subst; split; eauto.
  eapply T_TConv; eauto. eapply inversion_wft_sigma1;eauto.
  exists t2'; split; auto.
  exists x; split; auto.
  eapply trans_typeeq with (t2:=  (open_type_wrt_exp t3 (e_Var_f x))); eauto.
  rewrite_env (nil ++ one(e_Var_f x, e1) ++ erase G).
  replace (erase G) with (nil ++ erase G) in Jeq2; auto.
  eapply weakening_typeeq; eauto.
  simpl; split; auto.

 Case "TConv".
 eapply IHwfe; eauto. eapply trans_typeeq; eauto.
(*
 Case "CConv".  
  destruct IHwfe with (D:= erase G); eauto.
  eapply K_CConv; eauto using sym_ctxeq.  eapply reg_wfe_c; eauto.
  unfold not; intro; unfold not in H5; apply H5. 
  apply ctx_conv_incon with (D:=erase G); auto.
  eapply ctxeq_erase ; eauto.

  eapply ctx_conv_typeeq; eauto. eapply ctxeq_erase; eauto using sym_ctxeq.

  destruct H3 as [t2' (Jwfe2, [x  (Jx, Jtypeeq)])].
  split; auto.
  eapply T_CConv; eauto.

  exists t2'; split; auto.   eapply T_CConv; eauto.
  exists x; split; auto. eapply ctx_conv_typeeq; eauto.
  eapply CEQ_Eq; eauto using ref_isEq. eapply ctxeq_erase;eauto.
*)
Qed.
 
 

Lemma inversion_T_Ctr: forall D G e0 v u t C T,
  D = erase G ->
 ~ (incon D) ->
 wfe G e0 t ->
 e0 = (e_Con C v) ->
 typeeq D t (t_App(t_Con T) u) ->
 wft G  (t_App(t_Con T) u) k_type ->
 pure v ->
( exists s, exists ui, exists T0,
  binds C (bind_t (t_Pi s (t_App(t_Con T0) ui))) Signature/\
  wfe G v s /\ isEq D u (open_exp_wrt_exp ui v) /\ T = T0).
Proof.
 intros. generalize dependent D.
 induction H1; intros; subst; try solve [inversion H2]; eauto.

 Case "InCon".  contradiction.
 Case "T_Ctr".  
  clear IHwfe.
  inversion H2; subst. 
  exists s; exists u0; exists T0. (*; repeat split; eauto.*)
 
  pick fresh z.
  assert (J: typeeq  ((e_Var_f z, v) :: erase G)
     (t_App (t_Con T0) (open_exp_wrt_exp u0 (e_Var_f z)))
      (t_App (t_Con T) u)).
  eapply trans_typeeq with (t2:= t); eauto.
  rewrite_env (nil++ one (e_Var_f z, v) ++ erase G).
  replace (erase G) with (nil++erase G) in H8; auto.
  eapply weakening_typeeq; eauto. simpl; split;auto.
  assert (J1:= subst_typeeq z J H5).
  simpl in J1.
  destruct  (@eq_dec expvar _ z z) in J1.
  rewrite esubst_exp_fresh_eq in J1; auto. 
  assert (J2: typeeq (nil++esubst_ectx v z (erase G))
         (t_App (t_Con T0) (esubst_exp v z (open_exp_wrt_exp u0 (e_Var_f z))))
         (t_App (t_Con T) (esubst_exp v z u))).
  eapply cut_typeeq with (e1:= v) (e2:=v); eauto.
  eapply ref_isEq; eauto. 

  simpl_env in J2.
  rewrite (@esubst_exp_fresh_eq u) in J2; auto.
  rewrite <- esubst_exp_intro in J2; auto.
  rewrite <-  erase_subst_commute in J2; auto.
  rewrite esubst_G_fresh_eq in J2; auto.

  repeat split; eauto.
  apply sym_isEq; auto.
  eapply inversion_typeeq_app2 
    with (s:= t_App (t_Con T) u) 
       (t:= (t_App (t_Con T0) (open_exp_wrt_exp u0 v))); eauto.

  assert (J3: typeeq (erase G) (t_Con T0) (t_Con T)).   
  eapply inversion_typeeq_app1; eauto.
  symmetry. 
  eapply inversion_typeeq_T with (t:= t_Con T0) (s:=t_Con T) 
     (D:=erase G); eauto.
  
  eapply  notin_G_wfc; eauto. eapply reg_wft_c; eauto.
  destruct n; auto.

 Case "TConv".
 eapply IHwfe; eauto. eapply trans_typeeq; eauto.
(*
 Case "CConv".  
  destruct IHwfe with (D:= erase G); eauto.
  eapply K_CConv; eauto using sym_ctxeq.  eapply reg_wfe_c; eauto.
  unfold not; intro; unfold not in H6; apply H6. 
  apply ctx_conv_incon with (D:=erase G); auto.
  eapply ctxeq_erase ; eauto.

  eapply ctx_conv_typeeq; eauto. eapply ctxeq_erase; eauto using sym_ctxeq.
 destruct H2 as [u0[T0 (Jbind, (Jwfv, (J, J'))) ]].
 exists x; exists u0; exists T0; repeat split; eauto.
  eapply ctx_conv_isEq with (D1:= erase G); auto.  
  eapply ctxeq_erase; eauto using sym_ctxeq.*)
Qed.

Lemma fv_subst_find_pe: forall pb p,
  find_pe pb p -> efv_peb pb [<=] efv_pe p.
Proof.
  intros pb p Hfind.
  induction Hfind; simpl; solve_uniq.
Qed.

Lemma fv_subset_reduce: forall e e',
 reduce e e' ->
 efv_exp e' [<=] efv_exp e.
Proof.
intros e e' Hred.
 generalize dependent e'.
 induction e; intros; try solve [inversion Hred; eauto]; eauto.
 Case "Con".
 inversion Hred; subst; eauto.
 assert (J:= IHe e1' H2).  solve_uniq. 
 Case "App".
  inversion Hred; subst. 
  SCase "beta".
   assert (J:= efv_exp_open_exp_wrt_exp_upper e12 e2).
    simpl in *; solve_uniq.
  SCase "funbeta".
  assert (J:= efv_exp_open_exp_wrt_exp_upper v1 (e_Fix v1)).
    simpl in *; solve_uniq.
  SCase "cong1".
  assert (J1:= IHe1 e1' H3). simpl; solve_uniq.
  SCase "cong2".
  assert (J1:= IHe2 e2' H5). simpl; solve_uniq.
 Case "Case ".
  inversion Hred; subst.
  assert (J:= IHe e'0 H3).  simpl; solve_uniq. 
  assert (J:= efv_exp_open_exp_wrt_exp_upper ej v5).
  assert (J':=fv_subst_find_pe H5). 
  simpl in *; solve_uniq.
 Case "pair".
  inversion Hred; subst. 
  SCase "cong1".
  assert (J1:= IHe1 e1' H3). simpl; solve_uniq.
  SCase "cong2".
  assert (J1:= IHe2 e2' H5). simpl; solve_uniq.
 Case "Proj1".
  inversion Hred; subst. 
  simpl; solve_uniq.
  assert (J:= IHe e'0 H0).  solve_uniq. 
 Case "Proj2".
  inversion Hred; subst. 
  simpl; solve_uniq.
  assert (J:= IHe e'0 H0).  solve_uniq. 
Qed.

Lemma fv_subset_wfe: forall G e t,
 wfe G e t ->
 efv_exp e [<=] dom G.
Proof. 
  intros. 
  assert (J:= @notin_exp_wfe G e t).  
  assert (J1: forall x, x \in efv_exp e -> x \in dom G). 
  intros.  assert (J2:= J x).  solve_uniq.
  solve_uniq.
Qed. 

Lemma ctx_conv_wfpe: forall G e e' u pt t t1,
  wfpe G e u pt t ->
  isEq (erase G) e e' ->
  (forall G t, wfe G e t -> wfe G e' t) ->
  wfe G e t1 ->
  wfpe G e' u pt t.
Proof.
  intros.
  induction H; eauto.
  
  eapply AP_Cons with (L:=L\u efv_exp e \u efv_exp e' \u efv_type t \u dom G); eauto.
  intros.  eapply ctx_conv_wfe; eauto.  
  assert (Jwfe: wfe
         (AltAsn type (e, e_Con Ci (e_Var_f xi))
          :: AltAsn type (u, open_exp_wrt_exp ui (e_Var_f xi))
             :: VarAsn (exp * exp) xi si :: G)
         (open_exp_wrt_exp ei (e_Var_f xi)) t).
   eauto. assert (Jwfc:= reg_wfe_c Jwfe). 
   inversion Jwfc; subst. 
   eapply C_Eq; eauto.

  eapply CQ_Eq; eauto.   eapply CQ_Eq; eauto. 
  eapply CQ_Term; eauto.  eapply ref_ctxeq; eauto.
  eapply ref_typeeq; eauto. 
  assert (J:= wf_ctr_type Ci (t_Pi si (t_App (t_Con T) ui)) wfs_sig H). 
  destruct (lc_wft J) as (J1, (J2, J3)). 
  inversion J2; eauto.  eapply ref_isEq; eauto. simpl;  auto.
  eapply ref_isEq; auto. simpl; auto.
  eapply  lc_open_ui; eauto.
   rewrite_env (nil++erase (one (AltAsn type (u, open_exp_wrt_exp ui (e_Var_f xi)))) ++ erase G).
   eapply weakening_isEq; eauto. simpl; split; auto.  split; auto. 
   eapply lc_open_ui; eauto. 
   eapply ref_isEq; eauto. simpl; auto. repeat split; auto. 
   eapply lc_open_ui; eauto. 
Qed.

Lemma wfpe_find_pe: forall L G Cj v ej u pe t e peb,
  wfpe G e u pe t ->
  e = (e_Con Cj v) ->
  find_pe peb pe -> 
  peb = peb_Branch Cj ej ->
  (exists sj, exists T, exists uj, 
    binds  Cj  (bind_t   (t_Pi sj (t_App (t_Con T) uj))  ) Signature
    /\ ( exists xj , xj \notin  L  /\ 
      wfe  (AltAsn type (e_Con Cj v , (e_Con Cj (e_Var_f xj)) ) 
        :: (AltAsn type (u , ( open_exp_wrt_exp uj (e_Var_f xj) )   )
          :: (VarAsn (exp * exp) xj sj ::G )))    
      ( open_exp_wrt_exp ej (e_Var_f xj) ) t)).
Proof.
  intros L G Cj v ej u pe t e peb Hwfpe Heqe Hfind Heqpeb.
(*  generalize dependent ej.*)
  induction Hfind; intros; subst; eauto.
  inversion Hwfpe; subst; eauto.
  exists si; exists T; exists ui; split; eauto.
  pick fresh z.
  exists z; split; eauto.
  
 inversion Hwfpe; subst; eapply IHHfind; eauto.
Qed.

Lemma Preservation: forall G e e' t,
  wfe G e t -> reduce e e'-> wfe G e' t.
Proof.
 intros G e e' t Hwfe Hred.
 generalize dependent t. generalize dependent G.
 induction Hred; subst; intros; try solve[ inversion Hwfe]. 
Case "beta".
 destruct (incon_decidable (erase G) ) as [Jcon | Jincon].
 apply T_InCon; auto. eapply reg_wfe_c; eauto.
 assert (J:= efv_exp_open_exp_wrt_exp_upper e12 v2).
 assert (J':= subset_efv_exp_domG Hwfe).
 simpl in *;  solve_uniq. apply lc_body_exp_wrt_exp; auto.
 inversion H1; auto. eapply wft_wfe; eauto.
 remember (e_App (e_Lam e12) v2) as E. 
 induction Hwfe; subst; try solve [contradiction | inversion HeqE; inversion HeqG].
   SCase "app".
    inversion HeqE; subst.
    assert (J:exists x,   
      x \notin  L \u dom G \u efv_ctx G \u efv_type t \u efv_exp v2 \u efv_exp e12 \u efv_type t1 \u efv_type t2 
      /\ wfe  (AssumeList.VarAsn (exp * exp)  x   t1  ::  G )  
      ( open_exp_wrt_exp e12 (e_Var_f x) )  
      ( open_type_wrt_exp t2 (e_Var_f x) )).
   eapply inversion_elam; eauto.
   eapply ref_typeeq; eauto.
   eapply lc_ctx_erase; eauto. 
   eapply wft_wfe; eauto.
   destruct J as [x (Jin, Jwfe12)].
   eapply T_TConv with (t:= open_type_wrt_exp t2 v2); eauto.
   rewrite <- (@esubst_G_fresh_eq G v2 x); auto.
   rewrite (@esubst_exp_intro x e12 v2); auto.
   rewrite (@esubst_type_intro x t2 v2); auto.
   eapply subst_wfe_simpl with (tw:=t1); eauto.
   rewrite esubst_G_fresh_eq; auto.
   rewrite esubst_G_fresh_eq; auto.

   rewrite <- (@esubst_G_fresh_eq G v2 x); auto.
   rewrite (@esubst_type_intro x t2 v2); auto.
   rewrite <- (esubst_type_fresh_eq t v2 x); auto.
   rewrite  erase_subst_commute; eauto.
   rewrite_env (nil++esubst_ectx v2 x (erase G)).
   eapply cut_typeeq with (e1:= v2) (e2:= v2); eauto using ref_isEq.
   replace (v2, v2) with (esubst_exp v2 x (e_Var_f x), esubst_exp v2 x v2).
   rewrite_env (esubst_ectx v2 x ((e_Var_f x, v2)::(erase G))).
   eapply subst_typeeq; eauto.
   rewrite (@esubst_exp_fresh_eq v2); auto.
   simpl. destruct (@eq_dec expvar _ x x); eauto.
   destruct n; auto. 
   eapply ref_isEq; auto.
   eapply subst_lc_ectx; eauto. eapply lc_ctx_erase; eauto. 
   
   SCase "TConv". 
    eapply T_TConv with (t:=t); eauto.
(*   SCase "CConv".
   eapply T_CConv with (G:=G); eauto. eapply IHHwfe; eauto.
   intro; apply Jincon.
   apply ctx_conv_incon with (D:=erase G); auto.
   eapply ctxeq_erase ; eauto.*)

Case "funbeta". 
 destruct (incon_decidable (erase G) ) as [Jcon | Jincon].
 apply T_InCon; auto. eapply reg_wfe_c; eauto.
 assert (J:= efv_exp_open_exp_wrt_exp_upper v1 (e_Fix v1)).
 assert (J':= subset_efv_exp_domG Hwfe).
 simpl in *;  solve_uniq. 
 apply lc_e_App; auto. apply lc_body_exp_wrt_exp; auto.  
 inversion H3; auto.
 eapply wft_wfe; eauto.
 remember (e_App (e_Fix v1) v2) as E. 
 induction Hwfe; subst; try solve [contradiction | inversion HeqE; inversion HeqG].
   SCase "app".
    inversion HeqE; subst.
    eapply T_App; eauto.
    assert (J: (exists f,   
        f \notin  L \u dom G \u efv_ctx G \u efv_type t1 \u efv_type t2 \u efv_exp v1 /\
      wfe  (AssumeList.VarAsn (exp * exp)  f  (t_Pi t1 t2)  ::  G )  
      ( open_exp_wrt_exp v1 (e_Var_f f) )  (t_Pi t1 t2)
     /\(is_v_of_exp   ( open_exp_wrt_exp v1 (e_Var_f f) )  ))).
    eapply inversion_Fix; eauto. eapply wft_wfe; eauto.
    destruct J as [f (Jf, (Jwf, Jv))].
    rewrite <- (@esubst_G_fresh_eq G (e_Fix v1) f); auto.
    rewrite <- (@esubst_type_fresh_eq (t_Pi t1 t2) (e_Fix v1)  f ); auto.
    rewrite (esubst_exp_intro  f v1 (e_Fix v1) ); auto.
    eapply subst_wfe_simpl with (tw:= (t_Pi t1 t2)); eauto.
     rewrite esubst_G_fresh_eq; auto. 
    rewrite esubst_G_fresh_eq; auto.
   SCase "TConv". 
    eapply T_TConv with (t:=t); eauto.
 (*  SCase "CConv".
   eapply T_CConv with (G:=G); eauto. eapply IHHwfe; eauto.
   intro; apply Jincon.
   apply ctx_conv_incon with (D:=erase G); auto.
   eapply ctxeq_erase ; eauto.*)

Case "cong1". 
  remember (e_App e1 e2) as E. (* remember (nil:ctx) as G.*)
  induction Hwfe; subst; try solve [contradiction | inversion HeqE; inversion HeqG].
   SCase "incon".
    apply T_InCon; auto. 
    assert (J:= fv_subset_reduce Hred). 
    simpl in *;  solve_uniq. apply lc_e_App; auto.
    eapply lc_reduce ; eauto.
   SCase "app".
    inversion HeqE; subst.  eapply T_App; eauto.
   SCase "TConv". 
    eapply T_TConv with (t:=t); eauto.
(*   SCase "CConv".
   eapply T_CConv with (G:=G); eauto.*)
Case "cong2". 
  remember (e_App v1 e2) as E. (* remember (nil:ctx) as G.*)
  induction Hwfe; subst; try solve [contradiction | inversion HeqE; inversion HeqG].
   SCase "incon".
    apply T_InCon; auto. 
    assert (J:= fv_subset_reduce Hred). 
    simpl in *;  solve_uniq. apply lc_e_App; auto. 
    eapply lc_reduce; eauto.
   SCase "app".
    inversion HeqE; subst.
    eapply T_App with (L:=L); eauto.
    intros.  eapply ctx_conv_typeeq; eauto.
    eapply CEQ_Eq; eauto. eapply ref_ctxereq; eauto. 
    eapply lc_ctx_erase;eauto.
    eapply ref_isEq; eauto.     eapply lc_ctx_erase;eauto.
    rewrite_env (nil++erase G ++ nil).
    eapply weakening_isEq; eauto.  eapply lc_ctx_erase;eauto.
    simpl_env.
    eapply beta_isEq; eauto.
    SCase "TConv". 
    eapply T_TConv with (t:=t); eauto.
(*   SCase "CConv".
   eapply T_CConv with (G:=G); eauto.*)
Case "Ctr". 
  remember (e_Con C e1) as E. 
  induction Hwfe; subst; try solve [contradiction | inversion HeqE; inversion HeqG].
   SCase "incon".
    apply T_InCon; auto. 
    assert (J:= fv_subset_reduce Hred). 
    simpl in *;  solve_uniq. apply lc_e_Con; auto.
    eapply lc_reduce; eauto.
   SCase "ctr".
    inversion HeqE; subst.
    eapply T_Ctr with (L:=L); eauto.
    intros.  eapply ctx_conv_typeeq; eauto.
    eapply CEQ_Eq; eauto. eapply ref_ctxereq; eauto.
     eapply lc_ctx_erase;eauto.
    eapply ref_isEq; eauto.  eapply lc_ctx_erase;eauto.
    rewrite_env (nil++erase G ++ nil).
    eapply weakening_isEq; eauto.  eapply lc_ctx_erase;eauto. simpl_env.
    eapply beta_isEq; eauto.
   SCase "TConv". 
    eapply T_TConv with (t:=t); eauto.
(*   SCase "CConv".
   eapply T_CConv with (G:=G); eauto.*)
Case "Pair cong1".
  remember (e_Sie e1 e2) as E. (* remember (nil:ctx) as G.*)
  induction Hwfe; subst; try solve [contradiction | inversion HeqE; inversion HeqG].
   SCase "incon".
    apply T_InCon; auto. 
    assert (J:= fv_subset_reduce Hred). 
    simpl in *;  solve_uniq. apply lc_e_Sie; auto.
    eapply lc_reduce; eauto.
   SCase "pair".
    inversion HeqE; subst.
    eapply T_Sigma; eauto.
    intros.  eapply ctx_conv_typeeq; eauto.
   eapply CEQ_Eq; eauto. eapply ref_ctxereq; eauto.
    eapply lc_ctx_erase;eauto. 
   eapply ref_isEq; eauto.  eapply lc_ctx_erase;eauto.
   rewrite_env (nil++erase G ++ nil).
   eapply weakening_isEq; eauto.  eapply lc_ctx_erase;eauto. simpl_env.
   eapply beta_isEq; eauto.
   SCase "TConv". 
    eapply T_TConv with (t:=t); eauto.
(*   SCase "CConv".
   eapply T_CConv with (G:=G); eauto. *)
Case "Pair cong2".
  remember (e_Sie v5 e2) as E. (* remember (nil:ctx) as G.*)
  induction Hwfe; subst; try solve [contradiction | inversion HeqE; inversion HeqG].
   SCase "incon".
    apply T_InCon; auto. 
    assert (J:= fv_subset_reduce Hred). 
    simpl in *;  solve_uniq. apply lc_e_Sie; auto.
    eapply lc_reduce; eauto.
   SCase "pair".
    inversion HeqE; subst.
    eapply T_Sigma; eauto.
   SCase "TConv". 
    eapply T_TConv with (t:=t); eauto.
(*   SCase "CConv".
   eapply T_CConv with (G:=G); eauto.
*)
 
Case "proj1".
  destruct (incon_decidable (erase G) ) as [Jcon | Jincon].
  apply T_InCon; auto. eapply reg_wfe_c; eauto.
  assert (J:= subset_efv_exp_domG Hwfe). simpl in J; solve_uniq.
  eapply wft_wfe;eauto.

  remember (e_Proj1 (e_Sie v1 v2)) as E.
  induction Hwfe; subst; try solve [contradiction | inversion HeqE; inversion HeqG].
  SCase "proj1".
    inversion HeqE; subst.
    assert (J: ( wfe G v1 t1 
 /\ (exists t2', wfe G v2 t2' 
    /\ exists x, x \notin {}  /\
   typeeq  ((e_Var_f x, v1):: erase G)  t2'
   ( open_type_wrt_exp t2 (e_Var_f x) )))).
    eapply inversion_Pair with (e2:=v2); eauto.
    eapply ref_typeeq; eauto. eapply lc_ctx_erase; eauto. eapply wft_wfe; eauto.
    destruct J; auto.
   SCase "TConv". 
    eapply T_TConv with (t:=t); eauto.
(*   SCase "CConv".
   eapply T_CConv with (G:=G); eauto. eapply IHHwfe; eauto.
   intro; apply Jincon. 
   apply ctx_conv_incon with (D:=erase G); auto.
   eapply ctxeq_erase ; eauto.
*)
Case "proj2".
  destruct (incon_decidable (erase G) ) as [Jcon | Jincon].
  apply T_InCon; auto. eapply reg_wfe_c; eauto.
  assert (J:= subset_efv_exp_domG Hwfe). simpl in J; solve_uniq.
  eapply wft_wfe; eauto.

  remember (e_Proj2 (e_Sie v1 v2)) as E. (* remember (nil:ctx) as G.*)
  induction Hwfe; subst; try solve [contradiction | inversion HeqE; inversion HeqG].
   SCase "proj2".
    inversion HeqE; subst.
      assert (J: ( wfe G v1 t1 
     /\ (exists t2', wfe G v2 t2' 
       /\ exists x, x \notin  L\u efv_type t \u efv_exp v1 \u efv_ctx G \u dom G  /\
         typeeq  ((e_Var_f x, v1):: (erase G))  t2'
         ( open_type_wrt_exp t2 (e_Var_f x) )))). 
   eapply inversion_Pair; eauto. 
   eapply ref_typeeq. eapply lc_ctx_erase; eauto.  
   eapply lc_wfe; eauto. eapply wft_wfe; eauto.
   destruct J as (Jwft1, [t2' (Jwfe', [x (Jx, Jeq)])]).
   eapply T_TConv; eauto.
   rewrite <- (@esubst_G_fresh_eq G v1 x); auto.
   rewrite erase_subst_commute; eauto.
   rewrite <- (@esubst_type_fresh_eq t2' v1 x); auto.
   rewrite <- (@esubst_type_fresh_eq t v1 x); auto.
   rewrite_env  (nil++esubst_ectx v1 x (erase G)).
   eapply cut_typeeq with (e1:=esubst_exp v1 x (e_Var_f x)) (e2:=esubst_exp v1 x v1); eauto.
   rewrite_env  (esubst_ectx v1 x ((e_Var_f x, v1)::(erase G))).
   eapply subst_typeeq; eauto.
   eapply trans_typeeq with (t2:=  (open_type_wrt_exp t2 (e_Var_f x))); eauto.
   eapply ctx_conv_typeeq; eauto.
   eapply CEQ_Eq; eauto. eapply ref_ctxereq.
   eapply lc_ctx_erase; eauto.
   eapply ref_isEq; auto.  eapply lc_ctx_erase; eauto.
   rewrite_env (nil++erase G ++ nil).
   eapply weakening_isEq; eauto.  eapply lc_ctx_erase; eauto. simpl_env.
   eapply beta_isEq; eauto.
   rewrite (@esubst_exp_fresh_eq v1); auto.
   simpl. destruct (@eq_dec expvar _ x x).
   eapply ref_isEq. rewrite <- erase_subst_commute. 
    eapply lc_ctx_erase; eauto.
    eapply subst_lc_ctx; eauto. 
    eapply lc_wfe; eauto.
    destruct n; auto.
   eapply notin_type_wfe; eauto.

   SCase "TConv". 
    eapply T_TConv with (t:=t); eauto.
(*   SCase "CConv".
   eapply T_CConv with (G:=G); eauto.
   eapply IHHwfe; eauto.
   intro; apply Jincon. 
   apply ctx_conv_incon with (D:=erase G); auto.
   eapply ctxeq_erase ; eauto.
*)
Case "proj1 cong".
  remember (e_Proj1 e) as E. (* remember (nil:ctx) as G.*)
  induction Hwfe; subst; try solve [contradiction | inversion HeqE; inversion HeqG].
   SCase "incon".
    apply T_InCon; auto. 
    assert (J:= fv_subset_reduce Hred). 
    simpl in *;  solve_uniq. apply lc_e_Proj1; auto.
    eapply lc_reduce; eauto.
   SCase "proj1".
    inversion HeqE; subst. eapply T_Proj1 with (t1:=t1) (t2:=t2); eauto.
   SCase "TConv". 
    eapply T_TConv with (t:=t); eauto.
(*   SCase "CConv".
   eapply T_CConv with (G:=G); eauto.
*)
Case "proj2 cong".
  remember (e_Proj2 e) as E. (* remember (nil:ctx) as G.*)
  induction Hwfe; subst; try solve [contradiction | inversion HeqE; inversion HeqG].
   SCase "incon".
    apply T_InCon; auto. 
    assert (J:= fv_subset_reduce Hred). 
    simpl in *;  solve_uniq.
    apply lc_e_Proj2; auto.
    eapply lc_reduce; eauto.
   SCase "proj2".
    inversion HeqE; subst. eapply T_Proj2 with (t1:=t1) (t2:=t2) (L:=L); eauto.
    intros. eapply ctx_conv_typeeq; eauto.
    eapply CEQ_Eq; eauto. eapply ref_ctxereq; eauto.
    eapply lc_ctx_erase;eauto.
    eapply ref_isEq; eauto.     eapply lc_ctx_erase;eauto.
    rewrite_env (nil++erase G ++ nil).
    eapply weakening_isEq; eauto.     eapply lc_ctx_erase;eauto. simpl_env.
    eapply beta_isEq; eauto.
   SCase "TConv". 
    eapply T_TConv with (t:=t); eauto.
(*   SCase "CConv".
   eapply T_CConv with (G:=G); eauto.*)

Case "case cong". 
  remember (e_Case e pe5) as E. (* remember (nil:ctx) as G.*)
  induction Hwfe; subst; try solve [contradiction | inversion HeqE; inversion HeqG].
   SCase "incon".
    apply T_InCon; auto. 
    assert (J:= fv_subset_reduce Hred). 
    simpl in *;  solve_uniq. apply lc_e_Case ;auto.
    eapply lc_reduce; eauto.
   SCase "case".
    inversion HeqE; subst. 
    eapply T_Case ;eauto. eapply  ctx_conv_wfpe; eauto.
    rewrite_env (nil++erase G ++ nil).
    eapply weakening_isEq; eauto. eapply lc_ctx_erase;eauto. simpl_env.
    eapply beta_isEq; eauto.
   SCase "TConv". 
    eapply T_TConv with (t:=t); eauto.
(*   SCase "CConv".
   eapply T_CConv with (G:=G); eauto.
*)
Case "case beta".
  destruct (incon_decidable (erase G) ) as [Jcon | Jincon].
  apply T_InCon; auto. eapply reg_wfe_c; eauto.
  assert (J:= fv_subst_find_pe H2). 
  assert (J':=  efv_exp_open_exp_wrt_exp_upper ej v5).
  assert (J'':=  subset_efv_exp_domG Hwfe ).
  simpl in *. solve_uniq. apply lc_body_exp_wrt_exp; auto.
  assert (J:= lc_find_pe (peb_Branch Cj ej) pe5 H2).
  eapply  lc_body_find_pe; eauto. eapply wft_wfe; eauto.

  remember (e_Case (e_Con Cj v5) pe5) as E. 
  induction Hwfe; subst; try solve [contradiction | inversion HeqE; inversion HeqG].
  SCase "case".
    inversion HeqE; subst. 

    assert (J: exists sj, exists T, exists uj, 
    binds  Cj  (bind_t   (t_Pi sj (t_App (t_Con T) uj))  ) Signature
    /\ ( exists xj , xj \notin  dom G\u efv_type t \u efv_exp v5 \u efv_exp u \u efv_exp ej/\ 
      wfe  (AltAsn type (e_Con Cj v5 , (e_Con Cj (e_Var_f xj)) ) 
        :: (AltAsn type (u , ( open_exp_wrt_exp uj (e_Var_f xj) )   )
          :: (VarAsn (exp * exp) xj sj ::G )))    
      ( open_exp_wrt_exp ej (e_Var_f xj) ) t)). 
   eapply wfpe_find_pe; eauto.
   destruct J as [sj [T0 [uj (Jbind, [xj (Jin, Jwfej)])]]].
   assert (J:  exists s, exists ui, exists T0,
  binds Cj (bind_t (t_Pi s (t_App(t_Con T0) ui))) Signature/\
  wfe G v5 s /\ isEq (erase G) u (open_exp_wrt_exp ui v5) /\ T = T0).
  eapply inversion_T_Ctr; eauto.
  eapply ref_typeeq; eauto.
  eapply lc_ctx_erase; eauto. 
  eapply wft_wfe;eauto.
  destruct J as [s [ui [T1 (Jbind1, (Jwfv5, (JisEq, JT)))]]].
  assert (J: s = sj /\ T1 = T0 /\ ui = uj).
  eapply datacon_bind_uniq; eauto. 
  destruct J as (Jeq1, (Jeq2, Jeq3)).  subst.
  
   rewrite (@esubst_exp_intro xj ej v5); auto.
   rewrite <- (@esubst_type_fresh_eq t v5 xj); auto.
   rewrite_env (nil++ G).
   eapply cut_wfe with (e1:= esubst_exp v5 xj u) 
     (e2:= esubst_exp v5 xj (open_exp_wrt_exp uj (e_Var_f xj))); eauto.
   eapply cut_wfe with (e1:= esubst_exp v5 xj (e_Con Cj v5)) 
     (e2:= esubst_exp v5 xj (e_Con Cj (e_Var_f xj))); eauto.
   rewrite_env  (
     esubst_G v5 xj
     (AltAsn type
        (e_Con Cj v5,
        e_Con Cj (e_Var_f xj))
      :: one (AltAsn type
           (u, (open_exp_wrt_exp uj (e_Var_f xj))) )) 
         ++ G).
    eapply subst_wfe with (tw:=sj); eauto.
   simpl. destruct (@eq_dec expvar _ xj xj); eauto.
   rewrite (@esubst_exp_fresh_eq v5); auto.
   eapply ref_isEq; eauto.  simpl;repeat split; auto.
   eapply lc_esubst_exp; auto. eapply lc_open_ui; eauto.
   
   destruct n; auto.
   rewrite <- esubst_exp_intro.
   rewrite esubst_exp_fresh_eq; eauto.
   assert (J: wft nil (t_Pi sj (t_App (t_Con T0) uj)) k_type).
   eapply wf_ctr_type with (S:= Signature); eauto using wfs_sig.
   assert (J1: xj \notin efv_type  (t_Pi sj (t_App (t_Con T0) uj)) ).
   eapply notin_type_wft; eauto. simpl in J1; eauto.

   SCase "TConv". 
    eapply T_TConv with (t:=t); eauto.
(*   SCase "CConv".
   eapply T_CConv with (G:=G); eauto.
   eapply IHHwfe; eauto.
   intro; apply Jincon. 
   apply ctx_conv_incon with (D:=erase G); auto.
   eapply ctxeq_erase ; eauto.*)
Qed.
 
